/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_chkpt_req.c,v 1.150.12.8 2008/01/08 05:58:21 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/time.h>
#include <asm/uaccess.h>


// Interval for the watchdog to wakeup HZ = 1s
#define CR_CHKPT_WD_INTERVAL	(HZ)

// release_request()
//
// Release a request structure, free if reference count drops to zero.
//
// Returns non-zero if the request has actually been freed.
//
// No locks needed
static int release_request(cr_chkpt_req_t *req)
{
	int result = atomic_dec_and_test(&req->ref_count);

	if (result) {
		struct list_head *elem, *next;
		list_for_each_safe(elem, next, &req->procs) {
			cr_chkpt_proc_req_t *proc_req = list_entry(elem, cr_chkpt_proc_req_t, list);
			if (proc_req->shanon_tbl) {
			    vfree(proc_req->shanon_tbl);
			}
			kfree(proc_req);
		}
		cr_dest_free(&req->dest);
		cr_release_objectmap(req->map);
		kfree(req);
		CR_MODULE_PUT();
		CR_KTRACE_REFCNT("Free cr_chkpt_req_t %p", req);
	}

	return result;
}


// __delete_from_req(req, cr_task)
//
// Remove a given task from the request list and try to free resources.
// Returns non-zero if the request list becomes empty.
//
// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
static int __delete_from_req(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	int empty;

	list_del_init(&cr_task->req_list);
	list_del_init(&cr_task->proc_req_list);
	empty = list_empty(&req->tasks);
	if (empty) {
		wake_up(&req->wait);
	}
	cr_task->chkpt_req = NULL;
	__cr_task_put(cr_task);

	return empty;
}

// delete_from_req(req, cr_task)
//
// Remove a given task from the request list and try to free resources.
// Returns non-zero if the request list becomes empty.
//
// Call w/o holding locks
static int delete_from_req(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	int empty;

	write_lock(&cr_task_lock);
	write_lock(&req->lock);
	empty = __delete_from_req(req, cr_task);
	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

	return empty;
}

// delete_dead_task(req, cr_task)
//
// Remove a dead (or dying) task from the request
//
// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
//
// We don't (yet) raise any error flags here because it is legal
// for us to lose a race against an exiting process/thread.
// The reap path will catch the case in which ALL tasks died.
static void delete_dead_task(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	cr_chkpt_proc_req_t *proc_req = cr_task->chkpt_proc_req;

	// We make *some* effort here not to underflow if the task has
	// actually completed some of the barriers.  However, we still
	// can't know if this task signalled some that are not complete.
	if (!atomic_read(&proc_req->postsig_barrier.count)) goto done_postsig;
	if (!atomic_read(&proc_req->presig_barrier.count)) goto done_presig;
	if (!atomic_read(&req->postdump_barrier.count)) goto done_postdump;
	if (!atomic_read(&proc_req->vmadump_barrier.count)) goto done_vmadump;
	if (!atomic_read(&proc_req->predump_barrier.count)) goto done_predump;
	if ((cr_task->phase == CR_PHASE1) && !atomic_read(&proc_req->phase_barrier.count)) goto done_phase;
	if (!atomic_read(&req->preshared_barrier.count)) goto done_preshared;

	cr_barrier_notify(&req->preshared_barrier);
done_preshared:
	if (cr_task->phase == CR_PHASE1) {
		cr_barrier_notify(&proc_req->phase_barrier);
                if (cr_barrier_test(&proc_req->phase_barrier)) {
                        read_lock(&tasklist_lock);
                        __cr_trigger_phase2(proc_req);
                        read_unlock(&tasklist_lock);
                }
        }
done_phase:
	cr_barrier_notify(&proc_req->predump_barrier);
done_predump:
	cr_barrier_notify(&proc_req->vmadump_barrier);
done_vmadump:
	cr_barrier_notify(&req->postdump_barrier);
done_postdump:
	cr_barrier_notify(&proc_req->presig_barrier);
done_presig:
	cr_barrier_notify(&proc_req->postsig_barrier);
done_postsig:
	__delete_from_req(req, cr_task);
	(void)release_request(req);
}

// check_done(req)
// 
// Returns non-zero if the request can be reaped.
//
// Call w/o holding the locks
static int check_done(cr_chkpt_req_t *req)
{
	int retval;

	write_lock(&req->lock);
	retval = list_empty(&req->tasks);
	write_unlock(&req->lock);

	return retval;
}

#if !HAVE_STRUCT_WORK_STRUCT
// watchdog_timer(req)
//
// Timer which schedules a watchdog to run in process context
static void watchdog_timer(unsigned long arg)
{
	schedule_task((struct tq_struct *)arg);
}

static void schedule_delayed_work(cr_delayed_work_t *work, unsigned long delay)
{
	work->timer.expires = jiffies + delay;
	add_timer(&work->timer);
}
#endif

// add_watchdog(req)
//
// Add the timer for the watchdog, taking care not to schedule past
// the request's expiration time.
// XXX: no expiration yet!
//
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
static void add_watchdog(cr_chkpt_req_t *req)
{
	unsigned long delay;

	delay = CR_CHKPT_WD_INTERVAL;
	if (req->has_expiration && time_after(jiffies + delay, req->expiration)) {
		delay = req->expiration - jiffies;
	}
	schedule_delayed_work(&req->watchdog, delay);
}

// watchdog_task(req)
//
// Called periodically to cleanup after zombies and to enforce
// the time limit on a checkpoint request.
//
// XXX: no time limit yet
#if HAVE_STRUCT_DELAYED_WORK
static void watchdog_task(struct work_struct *work)
{
	cr_chkpt_req_t *req = container_of(work, cr_chkpt_req_t, watchdog.work);
#else
static void watchdog_task(void *arg)
{
	cr_chkpt_req_t *req = arg;
#endif
	struct list_head *elem, *next;
	int empty;

	write_lock(&cr_task_lock);
	write_lock(&req->lock);

	list_for_each_safe(elem, next, &req->tasks) {
		cr_task_t *cr_task = list_entry(elem, cr_task_t, req_list);
		struct task_struct *task = cr_task->task;
		if (cri_task_dead(task)) {
			int signo = task->exit_code & 0x7f;
			CR_WARN("%s: process %d exited with code %d "
				"and signal %d during checkpoint",
				__FUNCTION__,
				task->pid,
				(task->exit_code & 0xff00) >> 8,
				signo);
			// Raise an error if we killed it
			// XXX: this is now unlikely, given that we check
			// for the signal handler at request time.  Remove?
			if (signo == CR_SIGNUM) {
				req->result = -CR_ENOSUPPORT;
			}
			// Note that we hold one reference, thus we are
			// certain that this call won't destroy req.
			delete_dead_task(req, cr_task);
#if CRI_DEBUG
		} else {
			CR_SIGNAL_LOCK(task);
			if (sigismember(&task->blocked, CR_SIGNUM) && 
			    sigismember(&task->pending.signal, CR_SIGNUM)) {
				CR_ERR("Task %d has CR_SIGNUM blocked", task->pid);
			}
			// XXX: Need 1-arg recalc_sigpending() or equivalent
			//sigdelset(&task->blocked, CR_SIGNUM);
			//cr_recalc_sigpending_tsk(task)
			CR_SIGNAL_UNLOCK(task);
#endif
		}
	}
	empty = list_empty(&req->tasks);

	if (!empty) {
		if (req->has_expiration &&
		    time_after_eq(jiffies, req->expiration)) {
			CR_INFO("%s: timeout detected", __FUNCTION__);
			// XXX: handle timeout case here
#if 1 // XXX: temporary alternative
			req->has_expiration = 0;
			add_watchdog(req);
#endif
		} else {
			add_watchdog(req);
		}
	}

	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

	if (empty) {
		release_request(req);
	}
}

// alloc_request()
//
// Allocate and initialize a request structure
//
// No locks needed
static cr_chkpt_req_t * alloc_request(void)
{
	cr_chkpt_req_t *req = NULL;
	cr_objectmap_t map;

	if (!CR_MODULE_GET()) {
		CR_ERR("Checkpoint request after rmmod!");
		req = ERR_PTR(-EINVAL);
		goto out_rmmod;
	}

        map = cr_alloc_objectmap();
	if (!map) {
                goto out_modput;
	}

	req = kmalloc(sizeof(*req), GFP_ATOMIC);
	if (req) {
		CR_KTRACE_REFCNT("Alloc cr_chkpt_req_t %p", req);
		memset(req, 0, sizeof(*req));
		atomic_set(&req->ref_count, 1);
		atomic_set(&req->completed, 0);
		INIT_LIST_HEAD(&req->tasks);
		INIT_LIST_HEAD(&req->procs);
		init_waitqueue_head(&req->wait);
		req->requester = current->tgid;
		req->lock = RW_LOCK_UNLOCKED;
		req->result = 0;
		req->has_expiration = 0;
		req->map = map;
		req->die = 0;
	        init_MUTEX(&req->serial_mutex);
		CR_INIT_WORK(&req->watchdog, &watchdog_task, req, &watchdog_timer);
		cr_barrier_init(&req->preshared_barrier, 0);
		cr_barrier_init(&req->postdump_barrier, 0);
	} else {
                goto out_freemap;
	}

	return req;

out_freemap:
        cr_release_objectmap(map);
out_modput:
	CR_MODULE_PUT();
out_rmmod:
	return req;
}

// "lookup" a proc (thread group) by task->mm
// req->lock required unless otherwise serialized
static cr_chkpt_proc_req_t *
lookup_proc_req(cr_chkpt_req_t *req, struct task_struct *task)
{
	cr_chkpt_proc_req_t *result = NULL;
	struct list_head *elem;
	const struct mm_struct *mm = task->mm;

	list_for_each(elem, &req->procs) {
	    cr_chkpt_proc_req_t *tmp = list_entry(elem, cr_chkpt_proc_req_t, list);
	    if (tmp->mm == mm) {
		result = tmp;
		break;
	    }
	}
	if (!result) {
	    result = kmalloc(sizeof(*result), GFP_ATOMIC);
	    memset(result, 0, sizeof(*result));
	    result->mm = mm;
	    result->req = req;
	    INIT_LIST_HEAD(&result->tasks);
	    cr_barrier_init(&result->phase_barrier, 0);
	    cr_barrier_init(&result->predump_barrier, 0);
	    cr_barrier_init(&result->vmadump_barrier, 0);
	    cr_barrier_init(&result->presig_barrier, 0);
	    cr_barrier_init(&result->postsig_barrier, 0);
	    init_MUTEX(&result->serial_mutex);
	    list_add_tail(&result->list, &req->procs);
	    init_waitqueue_head(&result->wait);
	    result->saved_sa.sa.sa_handler = SIG_ERR;
	    result->forced_sa.sa.sa_handler = SIG_ERR;
	}

	return result;
}

// cr_is_dumpable(task)
// The 2.4 kernel kept changing its mind about how to check this
#if HAVE_IS_DUMPABLE
  #define cr_is_dumpable(t) (((t)->mm) && is_dumpable(t))
#else
  #if HAVE_TASK_TASK_DUMPABLE
    #define __cr_task_is_dumpable(t) ((t)->task_dumpable)
  #elif HAVE_TASK_DUMPABLE
    #define __cr_task_is_dumpable(t) ((t)->dumpable)
  #else
    #define __cr_task_is_dumpable(t) (1)
  #endif

  #if HAVE_MM_DUMPABLE
    #define __cr_mm_is_dumpable(mm) ((mm)->dumpable)
  #elif defined(CR_KCODE_get_dumpable)
    #define __cr_mm_is_dumpable(mm) get_dumpable(mm)
  #else
    #define __cr_mm_is_dumpable(mm) (1)
  #endif

  #define cr_is_dumpable(t) \
	(((t)->mm) && __cr_task_is_dumpable(t) && __cr_mm_is_dumpable((t)->mm))
#endif

// bad_perms(task)
//
// Verify that the 'current' task is permitted to checkpoint the
// indicated task.  The checks are based on ensuring that the
// invoking user "owns" the task and will not expose sensitive data.
//
// Must call w/ tasklist_lock held to prevent task from disappearing.
//
// Returns 0 if OK.
// Returns non-zero error code if permission should be denied.
//
static int bad_perms(struct task_struct *task)
{
	int result = 0;

	// We are always supposed to check capable/suser last.
	task_lock(task);
	if ((!cr_is_dumpable(task) ||
	     ((current->euid != task->suid) && (current->euid != task->uid) &&
	      (current->uid  != task->suid) && (current->uid  != task->uid)))
	    && !cr_capable(CAP_KILL)) {
		result = -EPERM;
	}
	task_unlock(task);

	return result;
}

// add_task(req, task)
//
// This function first verfies permissions to checkpoint the task.
// This function next records the req->task and task->chkpt_req associations.
// The req->task association in kept by the task_list field of the req.
// The task->chkpt_req association is kept by searching the list of requests,
// but could be put in a hash later.
//
// req->lock is not needed here because we are sole user.
// cr_task_lock must be held for the lookups.
//
// Returns 0 in the normal case.
// Returns non-zero error code on failure, such as out-of-memory.
//
static int add_task(cr_chkpt_req_t *req, struct task_struct *task)
{
	cr_chkpt_proc_req_t *proc_req = lookup_proc_req(req, task);
	cr_task_t *cr_task;
	int result;

	CR_KTRACE_FUNC_ENTRY("task=%p (%d)", task, task->pid);

	result = 0;

	/* XXX: Now that we check in add_proc(), these checks might be redundant */
	if (cri_task_zombie(task)) {
		CR_WARN("Skipped zombie task %d - a post-restart wait() will not find this task", task->pid);
		goto out;
	} else if (cri_task_dead(task)) {
		goto out;
	}

	result = bad_perms(task);
	if (result) {
		CR_WARN("Bad permissions to checkpoint task %d", task->pid);
		// Bad permissions
		goto out;
	}

	cr_task = __cr_task_get(task, 1);

	result = -ENOMEM;
	if (!cr_task) {
		// No memory
		goto out;
	}

        if (cr_task->chkpt_req == req) {
                if (proc_req->duplicate_flag != 1) {
                        /*
                         * This should NEVER happen
                         */
                        CR_WARN("Tried to add_task %d twice!", task->pid);
		        __cr_task_put(cr_task);
                        result = -EINVAL;
                        goto out;
                }
		CR_KTRACE_LOW_LVL("Confirmed duplicate task %d", task->pid);
        } else if (cr_task->chkpt_req || cr_task->rstrt_proc_req) {
	        result = -EBUSY;
		// Overlapping request
		CR_KTRACE_LOW_LVL("Request overlap when checkpointing task %d", task->pid);
		__cr_task_put(cr_task);
		goto out;
	}

	/* XXX: This is a good spot to check the task for proper
	 * checkpointing support (i.e. the signal handler).
	 * However, when we move away from the signal this will
	 * need to go away. */
	if (!cr_task->handler) {
	    __sighandler_t handler;
	    CR_SIGNAL_LOCK(task);
	    handler = CR_SIGNAL_HAND(task, CR_SIGNUM);
	    CR_SIGNAL_UNLOCK(task);
	    if ((handler == SIG_DFL) || (handler == SIG_IGN)) {
		result = -CR_ENOSUPPORT;
		__cr_task_put(cr_task);
		goto out;
	    }
	} else {
	    // Identify our saved handler, in case user has overwritten.
	    proc_req->forced_sa = cr_task->handler_sa;
	}

	result = 0;
        if (proc_req->duplicate_flag) {
		// These next two checks are somewhat unnecessary.  They're
		// just there to warn us if we're about to do something
		// completely wrong.  This can only happen if the
		// any of the other code that modifies the req_list or
		// proc_req_list has done something horrible. 
		//
		// assert(cr_task->chkpt_req == req);
		// assert(cr_task->chkpt_proc_req == proc_req);
		//
                if (cr_task->chkpt_req != req) {
	                CR_KTRACE_LOW_LVL("AAHH!!  req mismatch!");
                }

                if (cr_task->chkpt_proc_req != proc_req) {
	                CR_KTRACE_LOW_LVL("AAHH!!  proc_req mismatch!");
                }

	        list_del_init(&cr_task->req_list);
	        list_add_tail(&cr_task->req_list, &req->tasks);

                /*
		 * Ok, this looks stupid, but it's the right thing to do.
		 * This is a duplicate proc_req, which means that our 
		 * cr_task_get above has to be undone.  We know that there's
		 * already a reference to the task around, otherwise it would
		 * never have been marked as a duplicate in the first place.
                 */
		__cr_task_put(cr_task);
        } else {
	        if (cr_task->phase == CR_PHASE1) {
		        atomic_inc(&proc_req->phase_barrier.count);
	        }
	        ++proc_req->thread_count;
	        atomic_inc(&proc_req->predump_barrier.count);
	        atomic_inc(&proc_req->vmadump_barrier.count);
	        atomic_inc(&proc_req->presig_barrier.count);
	        atomic_inc(&proc_req->postsig_barrier.count);
		atomic_inc(&req->preshared_barrier.count);
		atomic_inc(&req->postdump_barrier.count);
	        atomic_inc(&req->ref_count);
	        list_add_tail(&cr_task->proc_req_list, &proc_req->tasks);
	        cr_task->chkpt_req = req;
	        cr_task->chkpt_proc_req = proc_req;
	        list_add_tail(&cr_task->req_list, &req->tasks);
	}

out:
	return result;
}

// add_proc(req, task)
//
// loop over add_task for all threads in the proc
//
// Returns 0 in the normal case.
// Returns non-zero error code on failure, such as out-of-memory.
//
static int add_proc(cr_chkpt_req_t *req, struct task_struct *proc)
{
	struct task_struct *task;
	int result = 0;

	CR_KTRACE_LOW_LVL("Add proc pid=%d", proc->pid);

	if (cri_task_zombie(proc)) {
		/* XXX: W/ NPTL the non-main threads could still be live */
		CR_WARN("Skipped zombie process %d - a post-restart wait() will not find this process", proc->tgid);
		goto out;
	} else if (cri_task_dead(proc)) {
		goto out;
	}

#if HAVE_2_4_TASK_LINKAGE
	CR_DO_EACH_TASK_PROC(proc, task) {
		result = add_task(req, task);
		if (result) {
			break;
		}
	} CR_WHILE_EACH_TASK_PROC(proc, task);
#elif HAVE_2_6_TASK_LINKAGE
	if (!proc->mm || !proc->signal) {
		result = -EPERM;	/* tried to checkpoint a kernel thread or something ?? */
       	} else if (atomic_read(&proc->signal->count) == atomic_read(&proc->mm->mm_users)) {
		/* NEW (nptl) pthreads (or single threaded) */
		CR_DO_EACH_TASK_TGID(proc->tgid, task) {
			result = add_task(req, task);
			if (result) {
				break;
			}
		} CR_WHILE_EACH_TASK_TGID(proc->tgid, task);
	} else {
		/* OLD (linuthreads) pthreads */
		CR_DO_EACH_TASK_PROC(proc, task) {
			result = add_task(req, task);
			if (result) {
				break;
			}
		} CR_WHILE_EACH_TASK_PROC(proc, task);
	}
#else
    #error
#endif

out:
	return result;
}

// build_req_proc(req, target)
//
// Locates the tasks which comprise a potentially multi-threaded
// process and verifies that the current task has permissions
// necessary to checkpoint them.
// A multi-threaded process may be named by the pid of any thread.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_proc(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *proc;
	int result = 0;

	read_lock(&tasklist_lock);
	proc = target ? find_task_by_pid(target)
		      : current; // Default target is self
	if (proc) {
		/* Get all the tasks in the process */
		result = add_proc(req, proc);
	}
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_PROC;
	return result;
}

// build_req_pgrp(req, target)
//
// Locates the tasks with a given process group id and verifies that the
// current task has permissions necessary to checkpoint them.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_pgrp(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *proc;
	int result = 0;

	if (!target) {
		// Default target is own pgrp
		target = cr_task_pgrp(current);
	}

	read_lock(&tasklist_lock);
	CR_DO_EACH_TASK_PGID(target, proc) {
		result = add_proc(req, proc);
		if (result) {
			break;
		}
	} CR_WHILE_EACH_TASK_PGID(target, proc);
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_PGRP;

	return result;
}

// build_req_sess(req, target)
//
// Locates the tasks with a given session id and verifies that the
// current task has permissions necessary to checkpoint them.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_sess(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *proc;
	int result = 0;

	if (!target) {
		// Default target is own session
		target = cr_task_session(current);
	}

	read_lock(&tasklist_lock);
	CR_DO_EACH_TASK_SID(target, proc) {
		result = add_proc(req, proc);
		if (result) {
			break;
		}
	} CR_WHILE_EACH_TASK_SID(target, proc);
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_SESS;

	return result;
}

// build_req_tree(req, target)
//
// Locates the tasks which comprise a process tree (a process and
// all of its decendents).
// Note that if a parent-grandchild relation exists, but the child
// has exited, then this grandchild is excluded.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
// Implements a breadth-first-search "in-place" using the req's own
// task list.
//
// Note: Depending on LinuxThreads vs NPTL the members of the root
// task's thread group may either be its children or its siblings.
// So, we unconditionally add the entire thread group of the root.
// All other tasks are found via the lists of children, taking care
// not to revisit the members of the root task's thread group.
//
static int build_req_tree(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *task;
	int result = 0;

        CR_KTRACE_LOW_LVL("in build_req_tree");

	read_lock(&tasklist_lock);
	task = target ? find_task_by_pid(target)
		      : current; // Default target is self
	if (task) {
		struct list_head *elem;
		struct mm_struct *root_mm = task->mm;

		/* Add "root" and all of its thread group */
		result = add_proc(req, task);
		if (result) {
			goto out_fail;
		}

                CR_KTRACE_LOW_LVL("scanning children");
		list_for_each(elem, &req->tasks) {
			struct task_struct *child;
			cr_task_t *cr_task = list_entry(elem, cr_task_t, req_list);
			task = cr_task->task;

                	CR_KTRACE_LOW_LVL("found child %d", task->pid);

			/* Add all the children - we get their children in the list_for_each() */
			CR_DO_EACH_CHILD(child, task) {

				if (child->mm == root_mm) continue;	/* Already visited */
				result = add_task(req, child);
				if (result) {
					goto out_fail;
				}
			} CR_WHILE_EACH_CHILD(child, task);
		}
	}
out_fail:
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_TREE;

	return result;	
}

// cr_chkpt_req(user_req)
//
// Processes a checkpoint request received from user space.
// Dispatches to build_req_*() to build a list of tasks.
// Calls cr_trigger_phase1() to trigger the checkpoint.
//
// Returns 0 on success, or a negative error code on failure.
int cr_chkpt_req(struct file *filp, void *arg)
{
	struct cr_chkpt_req	 ureq;
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	int result = 0;
	pid_t target;
	char * scope_msg = NULL;
	int did_retry = 0;

	CR_KTRACE_FUNC_ENTRY();

	// Check that we don't have a request on this fd already
	priv = filp->private_data;
	if (priv->chkpt_req) {
		return -EAGAIN;
	}

retry:
	// Allocate the kernel's structure to track the request
	req = alloc_request();
	if (!req) {
		return -ENOMEM;
	} else if (IS_ERR(req)) {
		return PTR_ERR(req);
	}

	// Copy the user's request
	if (copy_from_user(&ureq, arg, sizeof(ureq))) {
		release_request(req);
		return -EFAULT;
	}

	// XXX: Also copy the options if not NULL
	req->dump_format = ureq.dump_format;
	req->signal = ureq.signal;
	req->flags = ureq.flags;
	target = req->target = ureq.cr_target;

	/* write simple syslog message */
	switch (ureq.cr_scope) {
	case CR_SCOPE_PROC:
		if (!target) {
			target = current->tgid;
			scope_msg = "its own process";
		} else {
			scope_msg = "process";
		}
		break;
	case CR_SCOPE_PGRP:
		if (!target) { 
			target = cr_task_pgrp(current);
			scope_msg = "its own process group";
		} else {
			scope_msg = "process group";
		}
		break;
	case CR_SCOPE_SESS:
		if (!target) { 
			target = cr_task_session(current);
			scope_msg = "its own session";
		} else {
			scope_msg = "session";
		}
		break;
	case CR_SCOPE_TREE:
		if (!target) { 
			target = current->tgid;
			scope_msg = "its own process tree";
		} else {
			scope_msg = "process tree";
		}
		break;

	default:
		CR_KTRACE_BADPARM("invalid scope value: %d", ureq.cr_scope);
		release_request(req);
		return -EINVAL;
	}

	if (ureq.signal) {
		CR_KTRACE_HIGH_LVL("process %d checkpointing %s %d with signal %d",
			current->pid, scope_msg, target, ureq.signal);
	} else {
		CR_KTRACE_HIGH_LVL("process %d checkpointing %s %d",
			current->pid, scope_msg, target);
	}

	CR_KTRACE_LOW_LVL(" checkpoint params:  secs= %u, opts=%p, fd=%d",
			  ureq.cr_secs, ureq.cr_opts, ureq.cr_fd);

	// Validate the destination file descriptor
	result = cr_dest_init(&req->dest, ureq.cr_fd, filp);

	// Hold the lock needed to ensure the request is constructed
	// atomically w.r.t. registration of Phase[12] checkpoint tasks
	// and other checkpoint requests.
	write_lock(&cr_task_lock);

	// Build the list of tasks
	if (!result) {
		switch(ureq.cr_scope) {
		case CR_SCOPE_PROC:
			result = build_req_proc(req, ureq.cr_target);
			break;

		case CR_SCOPE_PGRP:
			result = build_req_pgrp(req, ureq.cr_target);
			break;

		case CR_SCOPE_SESS:
			result = build_req_sess(req, ureq.cr_target);
			break;

		case CR_SCOPE_TREE:
			result = build_req_tree(req, ureq.cr_target);
			break;

		default:
			result = -EINVAL;
		}
	}

	if (!result) {
		// Validate the task list
		if (list_empty(&req->tasks)) {
			// Fail if list is empty
			result = -ESRCH;
		} else if (req->flags & CR_CHKPT_PROHIBIT_SELF) {
			struct list_head *elem;
			cr_task_t *cr_task;

			// If the request says to prohibit "self", then scan
			// the list for "current", returning EDEADLK if found.
			//
			// req->lock not needed because we are sole user of req
			list_for_each(elem, &req->tasks) {
				cr_task = list_entry(elem, cr_task_t, req_list);
				if (cr_task->task == current) {
					result = -EDEADLK;
					break;
				}
			}
		}
	}

	// Either cleanup or finalize
	if (result) {
		struct list_head *elem, *next;
		cr_task_t *cr_task;

		// XXX: is this enough cleanup?
		//
		// req->lock not needed because we are sole user of req
		list_for_each_safe(elem, next, &req->tasks) {
			cr_task = list_entry(elem, cr_task_t, req_list);
			__delete_from_req(req, cr_task);
			// Note that we hold one reference, thus we are
			// certain release_request() won't destroy req.
			(void)release_request(req);
		}

		write_unlock(&cr_task_lock);

		release_request(req);

		// Retry *ONCE* on result = -CR_ENOSUPPORT
		// We sleep 0.5s and then try again in the hope that the
		// CR_ENOSUPPORT was due to a "new born" process that
		// had not yet registered the signal handler.
		if ((result == -CR_ENOSUPPORT) && !did_retry) {
			DECLARE_WAIT_QUEUE_HEAD(dummy);
			CR_WARN("Retry on -CR_ENOSUPPORT");
			(void)wait_event_interruptible_timeout(dummy, 0, HZ/2);
			did_retry = 1;
			goto retry;
		}
	} else {
		// Send the triggers
		cr_trigger_phase1(req);

		// Release the lock
		write_unlock(&cr_task_lock);

		// add watchdog timer, which constitutes one reference:
		if (ureq.cr_secs
#if BITS_PER_LONG == 32
			       	&& (ureq.cr_secs < (MAX_JIFFY_OFFSET / HZ))
#endif
				                                           ) {
			req->expiration = jiffies + HZ * ureq.cr_secs;
			req->has_expiration = 1;
		} else {
			// Either the requester specified no expiration or
			// the requested expiration is too large to handle.
			req->has_expiration = 0;
		}
		add_watchdog(req);
		// watchdog gets its own reference
		atomic_inc(&req->ref_count);

		priv->chkpt_req = req;
	}


	return result;
}

/*
 * I'm almost certain that this is in list.h already...
 */
static void cr_list_save(struct list_head *to, struct list_head *from)
{
        INIT_LIST_HEAD(to);
        list_add(to, from);
        list_del(from);
        INIT_LIST_HEAD(from);

}

static void
duplicate_set(cr_chkpt_req_t *req, int val)
{
        struct list_head *elem;

        list_for_each(elem, &req->procs) {
                cr_chkpt_proc_req_t *proc_req = list_entry(elem, cr_chkpt_proc_req_t, list);

                proc_req->duplicate_flag = val;
        }
}

// cr_chkpt_fwd(user_req)
//
// Propagates a checkpoint request to a different target.
int cr_chkpt_fwd(struct file *filp, void *arg)
{
	cr_task_t *my_cr_task;
	struct cr_fwd_req	 ureq;
	cr_chkpt_req_t		*req = NULL;
	cr_chkpt_proc_req_t	*proc_req = NULL;
	int result = 0;
	pid_t target;
	char * scope_msg = NULL;
        struct list_head old_req_tasks;

	CR_KTRACE_FUNC_ENTRY();

        // Lookup this task and find the matching request
        my_cr_task = cr_task_get(current);
	if (my_cr_task) {
		req = my_cr_task->chkpt_req;
		proc_req = my_cr_task->chkpt_proc_req;
	}
        cr_task_put(my_cr_task);
        if (!my_cr_task || !req || !proc_req) {
        	result = -ESRCH;
		goto out;
        }

	// Copy the user's request
	if (copy_from_user(&ureq, arg, sizeof(ureq))) {
		kfree(req);
		result = -EFAULT;
                goto out;
	}

	/* write simple syslog message */
        /* XXX: Move this into a common subroutine? */
	target = ureq.cr_target;
	switch (ureq.cr_scope) {
	case CR_SCOPE_PROC:
		if (!target) {
			target = current->tgid;
			scope_msg = "its own process";
		} else {
			scope_msg = "process";
		}
		break;
	case CR_SCOPE_PGRP:
		if (!target) { 
			target = cr_task_pgrp(current);
			scope_msg = "its own process group";
		} else {
			scope_msg = "process group";
		}
		break;
	case CR_SCOPE_SESS:
		if (!target) { 
			target = cr_task_session(current);
			scope_msg = "its own session";
		} else {
			scope_msg = "session";
		}
		break;
	case CR_SCOPE_TREE:
		if (!target) { 
			target = current->tgid;
			scope_msg = "its own process tree";
		} else {
			scope_msg = "process tree";
		}
		break;

	default:
		CR_KTRACE_BADPARM("invalid scope value: %d", ureq.cr_scope);
		result = -EINVAL;
                goto out;
	}

	
	CR_KTRACE_HIGH_LVL("process %d forwarding to %s %d",
			current->pid, scope_msg, target);

	// Hold the lock needed to ensure the request is constructed
	// atomically w.r.t. registration of Phase[12] checkpoint tasks
	// and other checkpoint requests.
	write_lock(&cr_task_lock);
        write_lock(&req->lock);

        cr_list_save(&old_req_tasks, &req->tasks);

        /*
         * Mark all proc_req's with the duplicate flag
         */
        duplicate_set(req, 1);

	// Build the list of tasks
	if (!result) {
		switch(ureq.cr_scope) {
		case CR_SCOPE_PROC:
			result = build_req_proc(req, ureq.cr_target);
			break;

		case CR_SCOPE_PGRP:
			result = build_req_pgrp(req, ureq.cr_target);
			break;

		case CR_SCOPE_SESS:
			result = build_req_sess(req, ureq.cr_target);
			break;

		case CR_SCOPE_TREE:
			result = build_req_tree(req, ureq.cr_target);
			break;

		default:
			result = -EINVAL;
		}
	}
        CR_KTRACE_LOW_LVL("result = %d", result);

	if (!result) {
		// Validate the task list
		if (list_empty(&req->tasks)) {
			// Fail if list is empty
                        CR_KTRACE_LOW_LVL("Forwarding list is empty.");
			result = -ESRCH;
		} 
        }  

        if (1) {
	    struct list_head *old_elem, *new_elem;
	    cr_task_t *new_cr_task;
	    cr_task_t *old_cr_task;

            /* dump the tasks out */
            list_for_each(old_elem, &old_req_tasks) {
                old_cr_task = list_entry(old_elem, cr_task_t, req_list);
	        CR_KTRACE_LOW_LVL("old task: %d", old_cr_task->task->pid);
            }

            list_for_each(new_elem, &req->tasks) {
		cr_chkpt_proc_req_t *proc_req;
                new_cr_task = list_entry(new_elem, cr_task_t, req_list);
                proc_req = new_cr_task->chkpt_proc_req;
                if (proc_req->duplicate_flag) {
		        CR_KTRACE_LOW_LVL("dup task: %d", new_cr_task->task->pid);
                } else {
	                CR_KTRACE_LOW_LVL("new task: %d", new_cr_task->task->pid);
                }
            }

#if 0
            /* check for duplicate tasks in the task list.  Ok, this check
	     * may not do what you think.  We check for duplicate tasks on
	     * the same request in the build_req_* calls above.  This check
	     * goes off and looks for duplicates that were missed by 
	     * build_req_*.  There SHOULD NEVER BE ANY OF THESE.
             */
            list_for_each(old_elem, &old_req_tasks) {
                old_cr_task = list_entry(old_elem, cr_task_t, req_list);
                list_for_each_safe(new_elem, next, &req->tasks) {
                    new_cr_task = list_entry(new_elem, cr_task_t, req_list);
                    if (new_cr_task == old_cr_task) {
                        result = -EBUSY;
		        CR_KTRACE_LOW_LVL("Request overlap when forwarding to task %d", new_cr_task->task->pid);
                    }
                } 
            }
#endif
	}

	// Send the triggers or clean up
	if (result) {
		struct list_head *elem, *next;
		cr_task_t *cr_task;

                /* clean up code.  We aborted (possibly part way) while adding
                 * processes to the request.  We now want to undo everything
                 * that we may have set up earlier, and restore the req to
                 * it's state before any of this forwarding ever occurred.
                 *
                 * 1.  Remove references to useless cr_task_t's in the req
                 *     Remove the tasks from req->tasks
                 *     Remove the tasks from req->procs
                 *     (__delete_from_req)
                 * 2.  Get rid of the associated proc_req's (?)
                 */

		list_for_each_safe(elem, next, &req->tasks) {
	                cr_chkpt_proc_req_t *proc_req;

			cr_task = list_entry(elem, cr_task_t, req_list);

	                CR_KTRACE_LOW_LVL("cleaning up task: %d", cr_task->task->pid);
                        proc_req = cr_task->chkpt_proc_req;

                        /* put the duplicates back on the original list */
                        if (proc_req->duplicate_flag) {
                               list_del_init(&cr_task->req_list);
                               list_add_tail(&cr_task->req_list, &old_req_tasks);
                                continue;
                        }

			cr_barrier_notify(&req->preshared_barrier);
			if (cr_task->phase == CR_PHASE1) {
				cr_barrier_notify(&proc_req->phase_barrier);
			}
			cr_barrier_notify(&proc_req->predump_barrier);
			cr_barrier_notify(&proc_req->vmadump_barrier);
			cr_barrier_notify(&req->postdump_barrier);
			cr_barrier_notify(&proc_req->presig_barrier);
			cr_barrier_notify(&proc_req->postsig_barrier);

			__delete_from_req(req, cr_task);

			// Note that the requestor and each checkpointing task
                        // hold a reference, so release_request() can't 
                        // destroy req.
			(void)release_request(req);
		}
	} else {
                struct list_head *elem;

                read_lock(&tasklist_lock);
                list_for_each(elem, &req->procs) {
                        cr_chkpt_proc_req_t *proc_req = list_entry(elem, cr_chkpt_proc_req_t, list);

                        /* For duplicates:  They're already going to be
                         * triggered, or have already been triggered, so we
                         * just need to clear the duplicate flag later.
                         */
                        if (!proc_req->duplicate_flag) {
				__cr_trigger_phase1(proc_req);
			}
                }
                read_unlock(&tasklist_lock);
	}

        /* clear the duplicate flag */
        duplicate_set(req, 0);

        /* Merge the new task list back with the original. */
        list_splice(&old_req_tasks, &req->tasks);
        INIT_LIST_HEAD(&old_req_tasks);

#if CRI_DEBUG
        if (1) {
	    struct list_head *old_elem, *new_elem;
	    cr_task_t *new_cr_task;
	    cr_task_t *old_cr_task;

            /* dump the tasks out */
            list_for_each(old_elem, &old_req_tasks) {
                old_cr_task = list_entry(old_elem, cr_task_t, req_list);
	        CR_KTRACE_LOW_LVL("old_req_tasks %d", old_cr_task->task->pid);
            }

            list_for_each(new_elem, &req->tasks) {
                new_cr_task = list_entry(new_elem, cr_task_t, req_list);
	        CR_KTRACE_LOW_LVL("req->tasks %d", new_cr_task->task->pid);
            }
	}
#endif

	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);
out:
	return result;
}

/**
 * cr_abort_checkpoint
 * @filp: The file on which the ioctl was received.
 * @flags:  abort type
 *
 * Returns:   0 if checkpoint successfully aborted.
 */
int cr_abort_checkpoint(struct file *filp, unsigned long flags)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;
	int result = 0;
	int block = 0;

	CR_KTRACE_FUNC_ENTRY("flags=0x%lx", flags);

	/* Lookup this task, and get request from it */
	cr_task = cr_task_get(current);
	if (!cr_task) {
		return -ESRCH;
	}
	req = cr_task->chkpt_req;
	if(!req) {
		cr_task_put(cr_task);
		return -ESRCH;
	}

	write_lock(&cr_task_lock);
	write_lock(&req->lock);
	
	/* Set status flag of checkpoint to aborted: causes other tasks in
	 * checkpoint (either globally or locally) to abort by end of predump barrier. */
	switch (flags) {
	    case CR_CHECKPOINT_PERM_FAILURE:
		result = req->result = -CR_EPERMFAIL;
		req->die = 1;	/* tell peers to kill themselves */
		send_sig_info(SIGKILL, NULL, current); /* kill self */
		break;

	    case CR_CHECKPOINT_TEMP_FAILURE:
		result = req->result = -CR_ETEMPFAIL;
		break;

	    case CR_CHECKPOINT_OMIT: {
		cr_chkpt_proc_req_t *proc_req = cr_task->chkpt_proc_req;
		proc_req->omit = 1; /* tell same-proc peers not to checkpoint */
		result = -CR_EOMITTED;

		/* We need to block for checkpoint to complete.
		 * Otherwise shared objects (pipes, shmem, etc) might become corrupted.
		 * However, need to do so *after* dropping the locks.
		 */
		block = 1;
		atomic_inc(&req->ref_count); /* So delete_dead_task() can't destroy req */
		break;
	    }

	    case 0:
		/* Do nothing */
		CR_WARN("%s: called w/ pointless zero argument\n", __FUNCTION__);
		goto out_no_delete;
		break;

	    default:
		CR_ERR("%s: called w/ invalid argument\n", __FUNCTION__);
		result = -EINVAL;
		goto out_no_delete;
	}

	delete_dead_task(req, cr_task);
out_no_delete:
	write_unlock(&req->lock);
	__cr_task_put(cr_task);
	write_unlock(&cr_task_lock);

	if (block) {
		CR_KTRACE_LOW_LVL("blocking for checkpoint to complete (count=%d)", atomic_read(&req->postdump_barrier.count));
		cr_barrier_wait(&req->postdump_barrier);
		(void)release_request(req);
	}

	CR_KTRACE_FUNC_EXIT("Returning %d", result);
	return result;
}



// cr_chkpt_task_complete(cr_task, completed)
// 
// Routine to release a given cr_ckpt_req_t.
// Called when a task has completed checkpointing itself (or aborts).
void cr_chkpt_task_complete(cr_task_t *cr_task)
{
	cr_chkpt_req_t *req = cr_task->chkpt_req;
	cr_chkpt_proc_req_t *proc_req = cr_task->chkpt_proc_req;
	struct task_struct *task = cr_task->task;
	int signal = 0;

	delete_from_req(req, cr_task);
	if (req->signal && !proc_req->omit) {
		signal = req->signal;
	} else if (cr_task->stopped) {
		signal = SIGSTOP;
	}
	if (signal) {
		cr_barrier_enter(&proc_req->presig_barrier);
#if HAVE_2_6_SIGNAL_STRUCT
		if ((atomic_read(&task->signal->count) == 1) ||
		    !test_and_set_bit(0, &proc_req->done_sig)) {
			kill_proc(task->tgid, signal, 0);
		}
#else
		send_sig_info(signal, NULL, task);
#endif     
		cr_barrier_enter(&proc_req->postsig_barrier);
	}
	release_request(req);
}

// cr_chkpt_req_release(filp, priv)
//
// Routine to release checkpoint request structures
// when a task closes the file (or exits).
//
// Must be called BEFORE cr_phase[12]_release
void cr_chkpt_req_release(struct file *filp, cr_pdata_t *priv)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;

	// If we are a "target" of a request with registered handlers,
	// then cleanup from what appears to be an unexpected exit/crash.
	// Note that we can't do this for filp's which don't have handlers
	// because cr_ctrl_fd() can validly close() an fd with no handlers.
	// Note: It is possible to reach here for a successful checkpoint
	// with a fatal req->signal because CR_OP_CHKPT_CONT may not run.
	cr_task = cr_task_get(current);
	if (cr_task) {
		req = cr_task->chkpt_req;
		if (req && (cr_task->phase != CR_NO_PHASE)) {
			write_lock(&cr_task_lock);
			write_lock(&req->lock);
			delete_dead_task(cr_task->chkpt_req, cr_task);
			write_unlock(&req->lock);
			write_unlock(&cr_task_lock);
		}
		cr_task_put(cr_task);
	}

	// If we are a requester, release the outstanding request.
	req = priv->chkpt_req;
	if (req) {
		(void)release_request(req);
	}
}

// cr_chkpt_done(timeval)
//
// Test/wait for a checkpoint request to complete.
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// Note that we write the remaining time back to timeval.
//
// XXX: Note that we can get stuck if a task rudely exits/dies while we
// are waiting.  We should consider waking up to reap such tasks.  That
// should be taken care of when a checkpoint time limit is added.
//
// XXX: consider ERESTARTSYS/ERESTARTNOHAND in place of EINTR?
int cr_chkpt_done(struct file *filp, void *arg)
{
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	struct timeval		timeval;
		
	unsigned long timeout;
	int retval = 0;

	CR_KTRACE_FUNC_ENTRY();

	priv = filp->private_data;
	if (!priv) {
		return -EINVAL;
	}
	req = priv->chkpt_req;
	if (!req) {
		return -EINVAL;
	}

	// Fetch/decode the optional timeout from userspace
	timeout = MAX_SCHEDULE_TIMEOUT;
	if (arg) {
		retval = copy_from_user(&timeval, arg, sizeof(timeval));
		if (retval) {
			return retval;
		}

		timeout = timeval_to_jiffies(&timeval);
	} 

	retval = 0;
	timeout = wait_event_interruptible_timeout(req->wait, (retval = check_done(req)), timeout);

	// Copy remaining time back to userspace if applicable
	if (arg) {
		struct timeval *tvp = (struct timeval *)arg;
		jiffies_to_timeval(timeout, &timeval);
		put_user(timeval.tv_sec, &tvp->tv_sec);
		put_user(timeval.tv_usec, &tvp->tv_usec);
		// like select() we fail silently if user timeval is R/O
	}

	// Check if we were interrupted
	if (!retval && timeout) {
		retval = -EINTR;
	}

	CR_KTRACE_FUNC_EXIT("Returning %d", retval);

	return retval;
}

// cr_chkpt_reap()
//
// Reap a completed checkpoint.
// Returns:
// 	req->result on success
// 	-EINVAL if no un-reaped checkpoint request is associated w/ this fd
// 	-EAGAIN if the request is not completed
//	-ESRCH if no task was checkpointed (possibly because they all exited on us)
int cr_chkpt_reap(struct file *filp)
{
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	int			retval;	

	retval = -EINVAL;
	priv = filp->private_data;
	if (priv) {
		req = priv->chkpt_req;
		if (req) {
			retval = -EAGAIN;
			if (check_done(req)) {
				priv->chkpt_req = NULL;
				retval = req->result;
				// req->result might be zero because nobody was actually checkpointed
				if (!retval && !atomic_read(&req->completed)) {
					retval = -ESRCH;
				}
				(void)release_request(req);
			}
		}
	}

	return retval;
}

int cr_chkpt_info(struct file *filp, struct cr_chkpt_info *uinfo)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;
	char *buf;
	char *path = NULL;
	int retval;

	CR_KTRACE_FUNC_ENTRY();

	buf = __getname();
	if (!buf) {
		return -ENOMEM;
	}

	retval = -ESRCH;
	cr_task = cr_task_get(current);
	if (!cr_task) {
		goto out;
	}
	req = cr_task->chkpt_req;
	cr_task_put(cr_task);
	if (!req) {
		goto out;
	}

	if (put_user(req->requester, &uinfo->requester) ||
	    put_user(req->target, &uinfo->target) ||
	    put_user(req->checkpoint_scope, &uinfo->scope) ||
	    put_user(req->signal, &uinfo->signal)) {
		retval = -EFAULT;
		goto out;
	}

	path = cr_location2path(&(req->dest), buf, PATH_MAX);
	if (IS_ERR(path)) {
		retval = PTR_ERR(path);
	} else if (path) {
		char *ubuf;
		if (get_user(ubuf, &(uinfo->dest))) {
			retval = -EFAULT;
		} else {
			retval = copy_to_user(ubuf, path, strlen(path));
		}
	} else {
		retval = -EBADF;
	}

out:
	__putname(buf);

	return retval;
}
