/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_io.c,v 1.37.26.1 2007/12/07 03:58:53 phargrov Exp $
 */

/*
 * Implements basic read and write calls for kernel space
 */

#include "cr_module.h"

#include <asm/uaccess.h>
#include <linux/errno.h>
#include <linux/dnotify.h>

/* Thin wrappers for args added over time: */

#if HAVE_4_ARG_DO_GENERIC_FILE_READ
  #define cr_generic_file_read(_f,_p,_d,_a) do_generic_file_read((_f),(_p),(_d),(_a))
#elif HAVE_5_ARG_DO_GENERIC_FILE_READ
  /* RH9 and RHEL add a non-block flag */
  #define cr_generic_file_read(_f,_p,_d,_a) do_generic_file_read((_f),(_p),(_d),(_a),0)
#else
  #error "Don't know how to call do_generic_file_read()"
#endif

#if HAVE_3_ARG_VFS_CREATE
  #define cr_vfs_create(_i,_d,_m,_v) vfs_create((_i),(_d),(_m))
#elif HAVE_4_ARG_VFS_CREATE
  #define cr_vfs_create(_i,_d,_m,_v) vfs_create((_i),(_d),(_m),(_v))
#else
  #error "Don't know how to call vfs_create()"
#endif

#if HAVE_4_ARG_VFS_MKNOD
  #define cr_vfs_mknod(_i,_d,_v,_m,_dev) vfs_mknod((_i),(_d),(_m),(_dev))
#elif HAVE_5_ARG_VFS_MKNOD
  #define cr_vfs_mknod(_i,_d,_v,_m,_dev) vfs_mknod((_i),(_d),(_v),(_m),(_dev))
#else
  #error "Don't know how to call vfs_mknod()"
#endif

#if HAVE_2_ARG_VFS_UNLINK
  #define cr_vfs_unlink(_i,_d,_v) vfs_unlink((_i),(_d))
#elif HAVE_3_ARG_VFS_UNLINK
  #define cr_vfs_unlink(_i,_d,_v) vfs_unlink((_i),(_d),(_v))
#else
  #error "Don't know how to call vfs_unlink()"
#endif

#if HAVE_2_ARG_NOTIFY_CHANGE
  #define cr_notify_change(_d,_v,_a) notify_change((_d),(_a))
#elif HAVE_3_ARG_NOTIFY_CHANGE
  #define cr_notify_change(_d,_v,_a) notify_change((_d),(_v),(_a))
#else
  #error "Don't know how to call notify_change()"
#endif

#ifndef CR_KCODE_vfs_read
  /* If you've ever seen the code for sys_read, this might look familiar.  */
  static ssize_t 
  vfs_read(struct file *file, char * buf, size_t count, loff_t *pos)
  {
    ssize_t ret;

    ret = -EBADF;
    if (file) {
        if (file->f_mode & FMODE_READ) {
  #if HAVE_LOCKS_VERIFY_AREA
            ret = locks_verify_area(FLOCK_VERIFY_READ, file->f_dentry->d_inode,
                        file, file->f_pos, count);
  #else
	    ret = rw_verify_area(READ, file, &file->f_pos, count);
  #endif
            if (!ret) {
                ssize_t (*read)(struct file *, char *, size_t, loff_t *);
                ret = -EINVAL;
                if (file->f_op && (read = file->f_op->read) != NULL)
                    ret = read(file, buf, count, pos);
            }
        }
	/* We skip inode_dir_notify, since it's not exported by the kernel */
    }
    return ret;
  }
#endif

#ifndef CR_KCODE_vfs_write
  static ssize_t 
  vfs_write(struct file *file, const char * buf, size_t count, loff_t *pos)
  {
    ssize_t ret;

    ret = -EBADF;
    if (file) {
        if (file->f_mode & FMODE_WRITE) {
  #if HAVE_LOCKS_VERIFY_AREA
            struct inode *inode = file->f_dentry->d_inode;
            ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
                file->f_pos, count);
  #else
	    ret = rw_verify_area(WRITE, file, &file->f_pos, count);
  #endif
            if (!ret) {
                ssize_t (*write)(struct file *, const char *, size_t, loff_t *);
                ret = -EINVAL;
                if (file->f_op && (write = file->f_op->write) != NULL)
                    ret = write(file, buf, count, pos);
            }
        }
	/* We skip inode_dir_notify, since it's not exported by the kernel */
    }
    return ret;
  }
#endif

/* Basically ripped off from vmadump */

ssize_t
cr_uread(struct file *file, void *buf, size_t count)
{
    ssize_t retval;
    ssize_t bytes_read;
    ssize_t bytes_left;

    bytes_left = count;

    while (bytes_left) {
       bytes_read = vfs_read(file, buf, bytes_left, &file->f_pos);
       if (bytes_read < 0) {
	   CR_ERR("vfs_read returned %ld", (long int)bytes_read);
	   retval = bytes_read;
	   goto out;
       } 
       if (bytes_read == 0) {
	   retval = count - bytes_left;
	   goto out;
       }
       bytes_left -= bytes_read;
       buf += bytes_read;
    }
    retval = count;
out:
    return retval; 
}

ssize_t
cr_uwrite(struct file *file, const void *buf, size_t count)
{
    ssize_t retval;
    ssize_t bytes_written;
    ssize_t bytes_left;
    const char *p = buf;

    bytes_left = count;

    while (bytes_left) {
       bytes_written = vfs_write(file, p, bytes_left, &file->f_pos);
       if (bytes_written < 0) {
	   retval = bytes_written;
	   goto out;
       } 
       if (bytes_written == 0) {
	   retval = count - bytes_left;
	   goto out;
       }
       bytes_left -= bytes_written;
       p += bytes_written;
    }
    retval = count;
out:
    return retval; 
}

ssize_t
cr_kread(struct file *file, void *buf, size_t count)
{
    ssize_t retval;
    mm_segment_t oldfs = get_fs();
    set_fs(KERNEL_DS);
    retval = cr_uread(file, buf, count);
    set_fs(oldfs);
    return retval; 
}

ssize_t
cr_kwrite(struct file *file, const void *buf, size_t count)
{
    ssize_t retval;
    mm_segment_t oldfs = get_fs();
    set_fs(KERNEL_DS);
    retval = cr_uwrite(file, buf, count);
    set_fs(oldfs);
    return retval; 
}


/* Skip unused data
 * XXX: Could/should we just seek when possible?
 */
int cr_skip(struct file *filp, loff_t len)
{
    ssize_t r;
    void *buf;
    mm_segment_t oldfs;

    r = -ENOMEM;
    buf = (void *) __get_free_page(GFP_KERNEL);
    if (!buf) goto out;

    oldfs = get_fs();
    set_fs(KERNEL_DS);
    while (len) {
	r = vfs_read(filp, buf, (len > PAGE_SIZE) ? PAGE_SIZE : len, &filp->f_pos);
	if (r <= 0) break;
	len -= r;
    }
    set_fs(oldfs);
    free_page((unsigned long)buf);

    if (r == 0) r = -EIO; /* EOF */
out:
    return (int)r;
}

/*
 * cr_fgets
 *
 * Sort of like fgets.
 *   Returns length (including '\0') on success, or <0 on error.
 *   0 return means string is NULL, 1 means empty string.
 */
int
cr_fgets(char *buf, int size, struct file *filp)
{
    int ret;
    int len = 0;

    /* Argument checking */
    ret = -EINVAL;
    if (!buf) {
        CR_ERR("cr_fgets: NULL buffer pass as argument!");
        goto out;
    }
    if (size <= 0) {
        CR_ERR("cr_fgets: Bad buffer size %d.", size);
        goto out;
    }

    ret = cr_kread(filp, &len, sizeof(len));
    if (ret < 0) {
        CR_ERR("cr_fgets: Bad file read.");
        goto out;
    } else if (ret != sizeof(len)) {
        ret = -EIO;
        CR_ERR("cr_fgets: Short read.");
        goto out;
    }
    if ((len < 0) || (len > size)) {
        ret = -EINVAL;
        CR_ERR("cr_fgets: Bad string length %d in file.", len);
        goto out;
    }

    if (len == 0) {
        /* old string was NULL.  i.e. never initialized rather than a
	 * pointer to an empty string */
        goto out;
    }

    ret = cr_kread(filp, buf, len);
    if (ret < 0) {
        CR_ERR("cr_fgets: Bad file read.");
        goto out;
    } else if (ret != len) {
        ret = -EIO;
        CR_ERR("cr_fgets: Short read.");
        goto out;
    }

out:
    return (ret < 0) ? ret : len;
}

/*
 * cr_fputs
 * 
 * Slightly different than fputs, since it writes the '\0' also, and refuses to
 * write anything larger than a PATH_MAX
 * 
 * Returns number of bytes written on success.
 */
int cr_fputs(const char *buf, struct file *filp)
{
    int ret;
    int wrote=0;
    int len;

    ret = -EIO;

    /* we actually want to know when someone wrote NULL to an area, so we
     * can restore the NULL on restart. */
    if (!buf) {
        /* magic length of 0 for NULL buf */
        len = 0;
    } else {
        /* we write out the length WITH the '\0' termination. */
        len = strlen(buf)+1;
    }

    /* compare l+1 to avoid type issue wrt promotion of len+1 to unsigned
     * long in comparison against PATH_MAX */
    if ((len < 0) || (len+1 > PATH_MAX+1)) {
        CR_ERR("cr_fputs:  String length (%d) out of bounds.", len);
        ret = -EINVAL;
        goto out;
    }

    ret = cr_kwrite(filp, &len, sizeof(len));
    wrote += ret;
    if (ret != sizeof(len)) {
        CR_ERR("cr_fputs: Bad file write");
        ret = -1;
        goto out;
    }
  
    if (len) {
        ret = cr_kwrite(filp, buf, len);
        wrote += ret;
        if (ret != len) {
            CR_ERR("cr_fputs: Bad file write");
            ret = -1;
            goto out;
        }
    }

    ret = wrote;

out:
    return ret;
}

/* Read in a filename from the context file, obtaining memory from __getname */
char *cr_getname(struct file *filp, int null_ok)
{
    char *name;
    int err;

    name = __getname();
    if (!name) {
        CR_ERR("Couldn't allocate buffer for file name.");
        err = -ENOMEM;
        goto out;
    }

    /* now read out the name */
    err = cr_fgets(name, PATH_MAX, filp);
    if (err < 0) {
        CR_ERR("Bad read of filename.");
        err = -EIO;
        goto out_free;
    } else if (err == 0) {
	if (!null_ok) err = -EIO;
	goto out_free;
    }

    return name;

out_free:
    __putname(name);
out:
    return (err < 0) ? ERR_PTR(err) : NULL;
}

/*
 * Copy the given number of bytes from one file to another.
 * Uses naive approach that should work for all types.
 * This is only used when a source file lacks a readpage method
 * (e.g. restart from a pipes or sockets).
 * XXX: Could we use non-blocking writes and double buffering?
 *
 * Note: Caller is responsible for checking count==0 or {dst,src}_ppos==NULL.
 */
static loff_t
cr_copyfile_buffered(struct file *dst_filp, loff_t *dst_ppos, struct file *src_filp, loff_t *src_ppos, loff_t count)
{
    const size_t maxsz = (4 << PAGE_SHIFT);
    mm_segment_t oldfs;
    loff_t retval = 0;
    loff_t bytes_left = count;
    char *buf;

    buf = vmalloc((count < maxsz) ? count : maxsz);
    retval = -ENOMEM;
    if (!buf) goto out_nobuf;

    retval = -EIO;
    oldfs = get_fs();
    set_fs(KERNEL_DS);

    while (bytes_left) {
        ssize_t buffered;
	char *p = buf;

	/* Read as much as we can in a single call */
	buffered = vfs_read(src_filp, p, (bytes_left < maxsz) ? bytes_left : maxsz, src_ppos);
        if (!buffered) goto out_eof;
        if (buffered < 0) {
	    CR_ERR("vfs_read returned %ld", (long int)buffered);
	    retval = buffered;
	    goto out_err;
        } 

	/* Write as much as we read */
        while (buffered) {
            ssize_t w = vfs_write(dst_filp, p, buffered, dst_ppos);
            if (!w) goto out_eof;
            if (w < 0) {
	        retval = w;
	        goto out_err;
            } 
            bytes_left -= w;
            buffered -= w;
            p += w;
	}
    }
out_eof:
    retval = count - bytes_left;
out_err:
    set_fs(oldfs);
    vfree(buf);
out_nobuf:
    return retval;
}

struct cr_copyfile_actor_data {
  struct file *dst_filp;
  loff_t *dst_ppos;
};

#if HAVE_READ_DESCRIPTOR_T_BUF
  #define CR_ACTOR_DATA(_desc) ((_desc)->buf)
#elif HAVE_READ_DESCRIPTOR_T_ARG_DATA
  #define CR_ACTOR_DATA(_desc) ((_desc)->arg.data)
#else
  #error
#endif

static int
cr_copyfile_actor(read_descriptor_t *desc, struct page *page, unsigned long offset, unsigned long size)
{ 
    const struct cr_copyfile_actor_data *data = (struct cr_copyfile_actor_data *)CR_ACTOR_DATA(desc);
    unsigned long bytes_left;
    char *addr, *p;
    mm_segment_t oldfs;
    ssize_t w = 0;

    if (size > desc->count) {
	size = desc->count;
    }
    bytes_left = size;

    oldfs = get_fs();
    set_fs(KERNEL_DS);

    addr = kmap(page);
    p = addr + offset;
    while (bytes_left) {
	w = vfs_write(data->dst_filp, p, bytes_left, data->dst_ppos);
        if (w <= 0) break;
	p += w;
	bytes_left -= w;
    }
    kunmap(page);

    set_fs(oldfs);

    if (w < 0) {
	desc->error = w;
	w = 0;
    } else {
	w = size - bytes_left;
	desc->count -= w;
	desc->written += w;
    }

    return (int)w;
}

/* Copy the given number of bytes from one file to another.
 * Uses provided src_ppos for src_file (or src_file->f_pos if NULL).
 * Uses provided dsc_ppos for dst_file (or dst_file->f_pos if NULL).
 * "zerocopy" reads are used when the source has a readpage method.
 */
loff_t
cr_copyfile(struct file *dst_filp, loff_t *dst_ppos, struct file *src_filp, loff_t *src_ppos, loff_t count)
{
    struct address_space *mapping;
    struct cr_copyfile_actor_data data;
    read_descriptor_t desc;
    loff_t retval = 0;

    if (!count) return 0;

    if (!dst_ppos) dst_ppos = &dst_filp->f_pos;
    if (!src_ppos) src_ppos = &src_filp->f_pos;

    mapping = src_filp->f_dentry->d_inode->i_mapping;
    if (!mapping || !mapping->a_ops || !mapping->a_ops->readpage) {
	/* Can't use page cache, so do our own buffering */
	return cr_copyfile_buffered(dst_filp, dst_ppos, src_filp, src_ppos, count);
    }

    data.dst_filp = dst_filp;
    data.dst_ppos = dst_ppos;
    CR_ACTOR_DATA(&desc) = (char *)&data; /* Yes, (char*) cast is needed */

#if SIZEOF_VOID_P == 4
    while (count) {
	/* Max 1GB chunks to avoid overflow (negative values) in desc.written */
	size_t chunk = ((count > 0x40000000) ? 0x40000000 : count);
        desc.count = chunk;
	desc.written = 0;
	desc.error = 0;
	cr_generic_file_read(src_filp, src_ppos, &desc, cr_copyfile_actor);
	if (!desc.written) return desc.error;
	if (desc.written != chunk) return -EIO;
	retval += chunk;
	count -= chunk;
    }
#else
    desc.count = count;
    desc.written = 0;
    desc.error = 0;
    cr_generic_file_read(src_filp, src_ppos, &desc, cr_copyfile_actor);
    retval = desc.written;
    if (!retval) {
	retval = desc.error;
    } else if (retval != count) {
	retval = -EIO;
    }
#endif

    return retval;
}

static char *
cr_getpath(struct dentry *dentry, struct vfsmount *mnt, char *buf, int size)
{
    struct dentry *d;
    struct vfsmount *m;
    char *path;

    path = NULL;

    if (dentry==NULL) {
        CR_WARN("dentry is NULL!");
	goto out;
    }
    if (mnt==NULL) {
        CR_WARN("mnt is NULL!");
	goto out;
    }

    d = dget(dentry);
    m = mntget(mnt);

    if (d==NULL) {
        CR_WARN("d dentry is NULL!");
	goto out;
    }
    if (m==NULL) {
        CR_WARN("m vfsmount is NULL!");
	goto out;
    }

    path = d_path(d, m, buf, size);
    dput(d);
    mntput(m); 

out:
    return path;
}

char *
cr_location2path(cr_location_t *loc, char *buf, int size)
{
	char *path = NULL;

	CR_KTRACE_FUNC_ENTRY();

	if (loc->filp) {
		/* It's a file */
    		path = cr_getpath(loc->filp->f_dentry, loc->filp->f_vfsmnt,
				  buf, size);
	} else if (loc->fs) {
		/* It's a directory */
		/* We can get away w/o locking since we never change it */
    		path = cr_getpath(loc->fs->pwd, loc->fs->pwdmnt,
				  buf, size - 2);

		if (path) {
			int len = strlen(buf);
			buf[len+0] = '/';
			buf[len+1] = '.';
			buf[len+2] = '\0';
		}
	} else {
		/* I don't know what it is */
		path = "";
	}

	if (!path) {
		path = ERR_PTR(-EBADF);
	}

	return path;
}

/* Saves pathname, returning bytes written (or <0 on error),
 * NULL dentry yields saved string value of NULL (distinct from empty string).
 * Uses supplied buf, if any, or will alloc/free otherwise.
 */
int cr_save_pathname(struct file *cr_filp, struct dentry *dentry, struct vfsmount *mnt, char *orig_buf, int size)
{
    int retval;
    const char *name = NULL;
    char *buf = orig_buf;

    /* Short cut on NULL dentry */
    if (!dentry) {
	goto write;
    }

    /* Allocate buf if none was supplied */
    if (!buf) {
        retval = -ENOMEM;
        buf = __getname();
        if (!buf) {
            goto out;
        }
	size = PATH_MAX;
    }

    /* find the file name */
    name = cr_getpath(dentry, mnt, buf, size);
    if (name == NULL) {
        CR_ERR("Bad or non/existant name!");
        retval = -EBADF;
        goto out_bad;
    }

    /* now write out the name */
write:
    retval = cr_fputs(name, cr_filp);
    if (retval < 0) {
        CR_ERR("cr_save_pathname - Bad file write! (cr_fputs returned %d)", retval);
        goto out_bad;
    }

out_bad:
    if (buf && !orig_buf) {
      __putname(buf);
    }
out:
    return retval;
}

int cr_save_filename(struct file *cr_filp, struct file *filp, char *buf, int size) {
    return cr_save_pathname(cr_filp, filp?filp->f_dentry:NULL, filp?filp->f_vfsmnt:NULL, buf, size);
}


/* based on linux/fs/namei.c:lookup_create */
/* Differs in that the balancing up() takes place on error */
static struct dentry *cr_lookup_create(struct nameidata *nd, int is_dir)
{
    struct dentry *dentry;

    dentry = lookup_create(nd, is_dir);
    if (IS_ERR(dentry)) {
      cr_inode_unlock(nd->dentry->d_inode);
    }
    return dentry;
}

/*
 * cr_anonymous_rename
 *
 * Rewrite a filename to an anonymous value
 *
 * len is strlen and size is buffer size
 */
static char *
cr_anonymous_rename(const char *in_buf, unsigned long id)
{
    size_t len;
    char *out_buf, *p;

    len = strlen(in_buf);
    if (len >= PATH_MAX) {
	/* XXX: probably not what we want to do */
	len = PATH_MAX - 1;
    }

    /* strdup() */
    out_buf = __getname();
    if (out_buf == NULL) {
	goto out;
    }
    memcpy(out_buf, in_buf, len+1);

    /* dirname() */
    p = out_buf + len;
    while ((p != out_buf) && (*p != '/')) {
	--p;
    }
    ++p;
    len = p - out_buf;

#if BITS_PER_LONG == 32
    if (len > PATH_MAX - 20) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%08lx", (unsigned int)current->pid, id);
#elif BITS_PER_LONG == 64
    if (len > PATH_MAX - 24) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%016lx", (unsigned int)current->pid, id);
#else
    #error "No value for BITS_PER_LONG"
#endif

    return out_buf;

out_free:
    __putname(out_buf);
out:
    return NULL;
}

/* Call permision() and dentry_open().
 * Caller should dget() and mntget() just as they would for dentry_open(). */
struct file *
cr_dentry_open(struct dentry *dentry, struct vfsmount *mnt, int flags)
{
    struct file *filp;
    int acc_mask = ("\000\004\002\006"[(flags)&O_ACCMODE]); /* ICK (from linux/fs/namei.c) */
    int err;

    err = cr_permission(dentry->d_inode, acc_mask, NULL);
    filp = ERR_PTR(err);
    if (!IS_ERR(filp)) filp = dentry_open(dentry, mnt, flags);

    return filp;
}

/* cr_mknod - based on linux/fs/namei.c:sys_mknod
 *
 * Creates regular files or fifos (no devices) making them anonymous (unlinked)
 * if desired.
 * Returns a dentry for the resulting filesystem objects, and the corresponding
 * vfsmnt can be obtained in nd->mnt.  Together these two can be passed
 * to dentry_open() or cr_dentry_open(), even for an unlinked inode.
 * In the event of an error, no dput() or path_release() is required,
 * otherwise they are.
 *
 * In the event that an object exists with the given name, it will be
 * check for the proper mode prior to return, yielding -EEXIST on conflict.
 */
struct dentry *
cr_mknod(struct nameidata *nd, const char *name, int mode, unsigned long unlinked_id)
{
    struct dentry * dentry;
    int err;

    if (unlinked_id) {
	/* Generate a replacement name which we will use instead of the original one. */
	name = cr_anonymous_rename(name, unlinked_id);
	if (!name) {
	    CR_ERR("cr_mknod - failed to rename unlinked object");
	    err = -ENOMEM;
	    goto out;
	}
    }

    err = cr_path_lookup(name, LOOKUP_PARENT, nd);
    if (err) {
	CR_ERR("Couldn't path_lookup for mknod %s.  err=%d.", name, err);
	goto out_free;
    }

    dentry = cr_lookup_create(nd, 0);
    if (IS_ERR(dentry)) {
	err = PTR_ERR(dentry);
	CR_ERR("Couldn't lookup_create for mknod %s.  err=%d.", name, err);
	goto out_release;
    }

    switch (mode & S_IFMT) {
    case S_IFREG:
	err = cr_vfs_create(nd->dentry->d_inode, dentry, mode, nd);
	break;
    case S_IFIFO:
	err = cr_vfs_mknod(nd->dentry->d_inode, dentry, nd->mnt, mode, 0 /* ignored */);
	break;
    default:
	CR_ERR("Unknown/invalid type %d passed to cr_mknod %s.", (mode&S_IFMT), name);
	err = -EINVAL;
    }
    if (unlinked_id && !err) { /* Note that we don't unlink if we failed to create */
	dget(dentry);	/* ensure unlink doesn't destroy the dentry */
	/* Note possibility of silent failure here: */
	(void)cr_vfs_unlink(nd->dentry->d_inode, dentry, nd->mnt);
	dput(dentry);
    }
    cr_inode_unlock(nd->dentry->d_inode);
    if ((err == -EEXIST) && !((dentry->d_inode->i_mode ^ mode) & S_IFMT)) {
	/* We fall through and return the dentry */
    } else if (err) {
	CR_ERR("Couldn't cr_mknod %s.  err=%d.", name, err);
	goto out_put;
    }

    return dentry;

out_put:
    dput(dentry);
out_release:
    path_release(nd);
out_free:
    if (unlinked_id) {
	__putname(name);
    }
out:
    return (struct dentry *)ERR_PTR(err);
}

/* Calls cr_mknod and then opens with the given flags, returning a (struct file *) */
struct file *
cr_filp_mknod(const char *name, int mode, int flags, unsigned long unlinked_id) {
    struct nameidata nd;
    struct dentry * dentry;
    struct file *filp;

    /* mknod */
    dentry = cr_mknod(&nd, name, mode, unlinked_id);
    if (IS_ERR(dentry)) {
	CR_ERR("Failed to recreate %sfilesystem object %s, err=%d.",
			unlinked_id?"unlinked ":"", name, (int)PTR_ERR(dentry));
	filp = (struct file *)dentry;
	goto out;
    }

    /* now open it */
    filp = cr_dentry_open(dget(dentry), mntget(nd.mnt), flags);
    if (IS_ERR(filp)) {
	CR_ERR("Failed to reopen %sfilesystem object %s, err=%d.",
			unlinked_id?"unlinked ":"", name, (int)PTR_ERR(dentry));
        goto out_dput;
    }

    /* check that we actually got the expected type */
    if ((mode ^ filp->f_dentry->d_inode->i_mode) & S_IFMT) {
	CR_ERR("Type conflict when recreating %sfilesystem object %s.",
			unlinked_id?"unlinked ":"", name);
	fput(filp);
	filp = ERR_PTR(-EEXIST);
	goto out_dput;
    }

out_dput:
    dput(dentry); 
    path_release(&nd);
out:
    return filp;
}

/* Based on sys_fchmod() from linux 2.6.21 (mostly unchanged since 2.4.0). */
int cr_filp_chmod(struct file *filp, mode_t mode) {
    struct iattr newattrs;
    struct dentry *dentry = filp->f_dentry;
    struct inode *inode = dentry->d_inode;
    int retval;

    retval = -EROFS;
    if (IS_RDONLY(inode)) goto out;
    retval = -EPERM;
    if (IS_IMMUTABLE(inode) || IS_APPEND(inode)) goto out;

    cr_inode_lock(inode);
    newattrs.ia_mode = (mode == (mode_t)-1) ? inode->i_mode
					    : ((mode & S_IALLUGO)|(inode->i_mode & ~S_IALLUGO));
    newattrs.ia_valid = ATTR_MODE | ATTR_CTIME;
    retval = cr_notify_change(dentry, filp->f_path.mnt, &newattrs);
    cr_inode_unlock(inode);

out:
    return retval;
}

/* Construct an unlinked file and read its data from the context file */
struct file *cr_mkunlinked(struct file *cr_filp, const char *name, int mode, int flags, loff_t size, unsigned long unlinked_id) {
    struct file *filp;
    loff_t w, pos;

#if CRI_DEBUG
    filp = ERR_PTR(-EINVAL);
    if (!unlinked_id) {
	CR_ERR("Zero 'unlinked_id' passed to %s", __FUNCTION__);
	goto out;
    }
    if ((mode&S_IFMT) && ((mode&S_IFMT) != S_IFREG)) {
	CR_ERR("Bad mode %d passed to %s", mode, __FUNCTION__);
	goto out;
    }
#endif

    /* make a new unlinked file, suitable for writting in the saved data */
    filp = cr_filp_mknod(name, mode|S_IFREG|S_IWUSR, O_LARGEFILE|O_WRONLY, unlinked_id);
    if (IS_ERR(filp)) {
	CR_ERR("Failed to recreate unlinked file %s, err=%d.", name, (int)PTR_ERR(filp));
	goto out;
    }

    /* and populate it with the saved data */
    /* XXX: If/when we split the create from the populate, then we'll need to either
     * do an ftruncate() here, or else defer the llseek() until post-populate.
     */
    pos = 0;
    w = cr_copyfile(filp, &pos, cr_filp, NULL, size);
    if (w < size) {
	filp = ERR_PTR(w);
        if (w >= 0) filp = ERR_PTR(-EIO);
        goto out;
    }

    /* Now reopen with caller-requested flags, if different */
    if (filp->f_flags != flags) {
	struct file *tmp = cr_filp_reopen(filp, flags);
	(void)filp_close(filp, current->files);
	filp = tmp; /* Even on error. */
    }
    if (IS_ERR(filp)) goto out;

    /* If we forced S_IWUSR, then fix it so fstat() doesn't reveal any change. */
    if (!(mode & S_IWUSR)) { /* Setting S_IWUSR is our only possible flaw */
	int err = cr_filp_chmod(filp, mode);
	if (err < 0) {
	    filp_close(filp, current->files);
	    filp = ERR_PTR(err);
	}
    }

out:
    return filp;
}

/* Create a new link to an existing (but potentially unlinked) dentry.
 * If the target link already exists we return the dentry, but if the target
 * exists and is not a link to the desired object, we return -EEXIST.
 *
 * NOTE: We once tried to do this via vfs_link(), rather than sys_link().
 * That was "better" in the sense that it was able to link to unlinked
 * targets (which this cannot).  However, that was not working over NFS
 * for reasons I never did figure out. -PHH
 */
struct dentry *
cr_link(struct dentry *old_dentry, struct vfsmount *mnt, const char *name)
{
    struct nameidata nd;
    char *buf, *old_name;
    struct dentry *new_dentry = NULL;
    int retval;
    mm_segment_t oldfs;
    
    /* Lookup the path to the "old" file.
     * This is the part that prevents us from linking to an unlinked target.
     */
    retval = -ENOMEM;
    buf = __getname();
    if (!buf) goto out;
    old_name = cr_getpath(old_dentry, mnt, buf, PATH_MAX);
    if (IS_ERR(old_name)) {
	retval = PTR_ERR(old_name);
	goto out_free;
    }

    /* Now sys_link() */
    oldfs = get_fs();
    set_fs(KERNEL_DS);
    retval = sys_link(old_name, name);
    set_fs(oldfs);
    if (retval == -EEXIST) {
	/* Keep going, it may be the one we want */
    } else if (retval < 0) {
	CR_ERR("cr_link: sys_link(%s,%s) returned %d", old_name, name, retval);
	goto out_free;
    }

    /* Now get the dentry for the newly-created object.
     * YES, there is a potential race, but we check below that we have the right object.
     */
    retval = cr_path_lookup(name, LOOKUP_FOLLOW, &nd);
    if (retval < 0) {
	CR_ERR("cr_link: cr_path_lookup(%s) returned %d", name, retval);
	goto out_free;
    }
    new_dentry = dget(nd.dentry);
    path_release(&nd);

    /* Check that we have a link to the desired object.
     * Needed for sys_link() == -EEXIST and for the link-to-lookup race.
     */
    if (new_dentry->d_inode != old_dentry->d_inode) {
	dput(new_dentry);
	retval = -EEXIST;
	goto out_free;
    }

out_free:
    __putname(buf);
out:
    return (retval < 0) ? ERR_PTR(retval) : new_dentry;
}

/* Reopen the same filesystem object as an exsiting filp.
 * Result is a distinct filp, with potentially different f_flags and f_mode.
 */
struct file *
cr_filp_reopen(const struct file *orig_filp, int new_flags)
{
    struct dentry *dentry;
    struct vfsmount *mnt;
    struct file *new_filp;

    new_filp = ERR_PTR(-EBADF);
    if (!orig_filp) goto out;

    dentry = dget(orig_filp->f_dentry);
    mnt = mntget(orig_filp->f_vfsmnt);
    new_filp = cr_dentry_open(dentry, mnt, new_flags);
    if (IS_ERR(new_filp)) {
	dput(dentry);
	mntput(mnt);
    }
out:
    return new_filp;
}

/*
 * cr_fd_claim
 *
 * Atomically checks and claims an fd
 *
 * returns < 0 on conflict w/ an existing fd
 */
int
cr_fd_claim(int fd)
{
    cr_fdtable_t *fdt;
    int retval;

    /* Mark the fd in use */
    cr_file_lock_write(&current->files->file_lock);
    fdt = cr_fdtable(current->files);
    if (FD_ISSET(fd, fdt->open_fds)) {
	retval = -EBUSY;
    } else {
	retval = 0;
	FD_SET(fd, fdt->open_fds);
	FD_CLR(fd, fdt->close_on_exec);
    }
    cr_file_lock_unwrite(&current->files->file_lock);

    return retval;
}
