/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_module.c,v 1.29 2007/08/31 22:07:45 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/init.h>

/* Find some always-exported symbols needed to validate System.map */
#include <linux/unistd.h>		/* for __NR_close */

MODULE_AUTHOR("Lawrence Berkeley National Lab " PACKAGE_BUGREPORT);
MODULE_DESCRIPTION("Berkeley Lab Checkpoint/Restart (BLCR) kernel module");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#if CR_KERNEL_TRACING
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  MODULE_PARM(cr_ktrace_mask, "i");
#else
  #include <linux/moduleparam.h>
  module_param(cr_ktrace_mask, uint, 0644);
#endif
MODULE_PARM_DESC(cr_ktrace_mask, 
		 "Mask of ORed tracing flags: see cr_ktrace.h for list");
#endif

static int __init cr_init_module(void)
{
	int err;

#if CR_KERNEL_TRACING 
	CR_INFO("blcr: Berkeley Lab Checkpoint/Restart (BLCR) module version %s.  "
		"Tracing enabled (trace_mask=0x%x)", PACKAGE_VERSION, cr_ktrace_mask);
#else
	CR_INFO("blcr: Berkeley Lab Checkpoint/Restart (BLCR) module version %s.", PACKAGE_VERSION);
#endif
	CR_INFO("blcr: Supports BLCR kernel interface version %s.", CR_MODULE_VERSION);
	CR_INFO("blcr: %s", PACKAGE_BUGREPORT);

	/* Check current kernel against System.map used at configure time */
	{
	    unsigned long offset1 = CR_EXPORTED_KCODE_register_chrdev - (unsigned long)&register_chrdev;
	    unsigned long offset2 = CR_EXPORTED_KCODE_register_blkdev - (unsigned long)&register_blkdev;
	    if (
#if defined(CONFIG_RELOCATABLE) && defined(CONFIG_PHYSICAL_ALIGN)
	    /* Check that relocation offset is valid and constant */
	        ((offset1 & (CONFIG_PHYSICAL_ALIGN - 1)) || (offset1 != offset2))
#else
	    /* No relocation */
	        (offset1 || offset2)
#endif
	        ) {
		CR_ERR("Running kernel does not match the System.map used to build blcr.o");
		return -EINVAL;
	    }
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	/*
	 * Check System.map used at configure time vs. sys_call_table
	 * This is only needed for 2.4.x kernels, due to vmadump's use of sys_call_table.
	 */

    #if !CR_EXPORTED_KCODE_sys_ptrace && !CR_KCODE_sys_ptrace && \
        !CR_EXPORTED_KCODE_sys_close  && !CR_KCODE_sys_close
	#error "No symbols available to validate sys_call_table at load time"
    #endif

	if (0
    #if CR_EXPORTED_KDATA_sys_call_table
	    || ((void *)(CR_EXPORTED_KDATA_sys_call_table) != &sys_call_table[0])
    #endif
    #if CR_EXPORTED_KCODE_sys_ptrace
	    || ((void *)(CR_EXPORTED_KCODE_sys_ptrace) != sys_call_table[__NR_ptrace])
	    || ((void *)(CR_EXPORTED_KCODE_sys_ptrace) != (void *)&sys_ptrace)
    #elif CR_KCODE_sys_ptrace
	    || ((void *)(CR_KCODE_sys_ptrace) != sys_call_table[__NR_ptrace])
    #endif
    #if CR_EXPORTED_KCODE_sys_close
	    || ((void *)(CR_EXPORTED_KCODE_sys_close) != sys_call_table[__NR_close])
	    || ((void *)(CR_EXPORTED_KCODE_sys_close) != (void *)&sys_close)
    #elif CR_KCODE_sys_close
	    || ((void *)(CR_KCODE_sys_close) != sys_call_table[__NR_close])
    #endif
	   ) {
		CR_ERR("Running kernel's sys_call_table does not match probed information");
		return -EINVAL;
	}
#endif

	err = cr_proc_init();
	if (err) {
		return err;
	}

#ifdef CR_ENABLE_BOGUS_FOPS
	err = cr_fs_init();
	if (err) {
		return err;
	}
#endif
#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
	err = cr_init_ioctl32();
	if (err) {
		return err;
	}
#endif

	return 0;
}

static void __exit cr_cleanup_module(void)
{
#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
	cr_cleanup_ioctl32();
#endif
#ifdef CR_ENABLE_BOGUS_FOPS
	cr_fs_cleanup();
#endif
	cr_proc_cleanup();
	CR_INFO("Checkpoint/Restart module removed");
}

module_init(cr_init_module);
module_exit(cr_cleanup_module);
