/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_objects.c,v 1.4 2005/07/19 23:07:47 phargrov Exp $
 */

#include "cr_module.h"

#define MAP_SIZE 64
#define MAP_SHIFT 6

/* doubly linked list of key=val pairs */
struct cr_objectmap_entry_s {
    struct list_head	list;
    void *key;
    void *val; 
};

/* Map is an array of lock+list pairs */
struct cr_objectmap_s {
    rwlock_t		lock;
    struct list_head	list;
};

static
int hash_it(void *x)
{
    unsigned long tmp = (unsigned long)x;
    tmp = tmp ^ (tmp >> MAP_SHIFT) ^ (tmp >> (MAP_SHIFT*2));
    return (int)(tmp % MAP_SIZE);
}

cr_objectmap_t
cr_alloc_objectmap(void)
{
    struct cr_objectmap_s *map = kmalloc(MAP_SIZE*sizeof(struct cr_objectmap_s), GFP_KERNEL);

    if (map) {
	int i;
	for (i=0; i<MAP_SIZE; ++i) {
	    map[i].lock = RW_LOCK_UNLOCKED;
	    INIT_LIST_HEAD(&(map[i].list));
	} 
    }

    return map;
}

void
cr_release_objectmap(cr_objectmap_t map)
{
    int i;

    for (i=0; i<MAP_SIZE; ++i) {
	struct list_head *elem, *next;
	list_for_each_safe(elem, next, &(map[i].list)) {
	    kfree(list_entry(elem, struct cr_objectmap_entry_s, list));
	}
    }
    kfree(map);
}

/*
 * returns int rather than void * to allow NULL key's to be placed into table
 *
 * 1 - found
 * 0 - not found
 */
int
cr_find_object(cr_objectmap_t map, void *key, void **val_p)
{
    int retval = 0;

    /* result is NULL by default, use return value to distinguish good NULL from evil NULL */
    if (val_p != NULL) {
	*val_p = NULL;
    }

    if (key == NULL) {
    	/* special cased to avoid confusion with not present */
        // CR_KTRACE_LOW_LVL("map %p: Asked for NULL, returning NULL.", map);
        retval = 1;
    } else {
	struct list_head *elem;
        int h = hash_it(key);

        read_lock(&(map[h].lock));
	list_for_each(elem, &(map[h].list)) {
	    struct cr_objectmap_entry_s *entry = list_entry(elem, struct cr_objectmap_entry_s, list);
	    if (entry->key == key) {
		// CR_KTRACE_LOW_LVL("map %p: Found object %p in slot %d", map, key, h);
		if (val_p != NULL)
		    *val_p = entry->val;
		retval = 1; 
		break;
	    } else if (entry->key > key) {
		/* Sorted order would have placed entry here */
		break;
	    }
	}
        read_unlock(&(map[h].lock));

	// if (!retval) CR_KTRACE_LOW_LVL("map %p: Object %p not found", map, key);
    }

    return retval;
}

/*
 *  1 if it's in there already 
 *  0 if we insert it
 */
int
cr_insert_object(cr_objectmap_t map, void *key, void *val)
{
    struct cr_objectmap_entry_s *new_entry;
    struct list_head *elem;
    int retval = -1;
    int h = hash_it(key);

    /* special case NULL -> NULL */
    if (key == NULL) {
        retval = 1;
	return retval;
    }

    /* assume it's not there yet to move alloc outside of lock and keep a single pass */
    new_entry = kmalloc(sizeof(struct cr_objectmap_entry_s), GFP_KERNEL);
    new_entry->key = key;
    new_entry->val = val;

    write_lock(&(map[h].lock));
    list_for_each(elem, &(map[h].list)) {
	struct cr_objectmap_entry_s *entry = list_entry(elem, struct cr_objectmap_entry_s, list);
	if (entry->key == key) {
	    /* return 1 if it's in there already */
	    // CR_KTRACE_LOW_LVL("map %p: Object %p already inserted into slot %d", map, key, h);
	    retval = 1; 
	    break;
	} else if (entry->key > key) {
	    /* Sorted order places new entry here */
	    break;
	}
    }
    if (retval != 1) {
	// CR_KTRACE_LOW_LVL("map %p: Inserting object %p into slot %d", map, key, h);
	list_add_tail(&new_entry->list, elem);
	retval = 0;
    }
    write_unlock(&(map[h].lock));

    if (retval) {
	kfree(new_entry);
    }

    return retval;
}

/*
 *  0 if we remove it
 *  -1 if it's not in there
 */
int
cr_remove_object(cr_objectmap_t map, void *key)
{
    struct list_head *elem, *next;
    int h = hash_it(key);
    int retval=-1;

    write_lock(&(map[h].lock));
    list_for_each_safe(elem, next, &(map[h].list)) {
	struct cr_objectmap_entry_s *entry = list_entry(elem, struct cr_objectmap_entry_s, list);
	if (entry->key == key) {
	    list_del(&entry->list);
	    kfree(entry);
	    retval = 1; 
	    break;
	}
    }
    write_unlock(&(map[h].lock));

    return retval;
}
