/*-------------------------------------------------------------------------
 *  cr_shanon.c:  Code to save/restore shared-anonymous memory mappings
 *
 *  2007, The Regents of the University of California, through Lawrence
 *  Berkeley National Laboratory (subject to receipt of any required
 *  approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 *  Portions may be copyrighted by others, as may be noted in specific
 *  copyright notices within specific files.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_shanon.c,v 1.16 2007/07/17 05:55:48 phargrov Exp $
 *
 * This file handles shared-anonymous memory mappings
 *-----------------------------------------------------------------------*/

#include "cr_module.h"
#include <linux/mman.h>
#ifdef CONFIG_HUGETLBFS
  #include <linux/hugetlb.h>
#endif

// Descriptor for shared/anonymous mapping
// Used both in-memory and on-disk
struct cr_shanon_desc {
  /* What is mapped: */
    void *		shanon_id; /* The (struct inode *) at checkpoint time */
    loff_t		i_size;
    int			type;
  /* Where and how is it mapped: */
    unsigned long	start, end;
    unsigned long	flags;
    unsigned long	pgoff;  /* units of PAGE_SIZE */
};

// Vales for desc->type
enum {
  CR_SHANON_FILE,
  CR_SHANON_SHMEM,
#ifdef CONFIG_HUGETLBFS
  CR_SHANON_HUGETLB,
#endif
};

/* Read the shared-anonymous mapping metadata from the context file
 *
 * Returns 0 on success, or <0 on error
 */
long
cr_load_shanon_maps(cr_rstrt_proc_req_t *proc_req)
{
    struct cr_shanon_desc *desc_tbl = NULL;
    size_t size = 0;
    int count = 0;
    int retval;

    /* Read size (in bytes) of table */
    retval = cr_kread(proc_req->file, &size, sizeof(size));
    if (retval < sizeof(size)) {
        CR_ERR("short read: %d of %d bytes", retval, (int)sizeof(size));
        retval = -EIO;
        goto err;
    }
    if (!size) goto out_notbl;
    count = size / sizeof(struct cr_shanon_desc);
#if CRI_DEBUG
    if (size != count * sizeof(struct cr_shanon_desc)) {
        CR_ERR("read invalid shanon tbl size: %ld is not a multiple of %ld bytes", (long)size, (long)sizeof(struct cr_shanon_desc));
        retval = -EINVAL;
        goto err;
    }
#endif

    /* Allocate space */
    desc_tbl = vmalloc(size);
    if (!desc_tbl) {
	CR_ERR("Failed to allocate %ld bytes for shanon table", (long)size);
        retval = -ENOMEM;
        goto err;
    }

    /* Read the table */
    retval = cr_kread(proc_req->file, desc_tbl, size);
    if (retval < size) {
        CR_ERR("short read: %d of %d bytes", retval, (int)size);
        retval = -EIO;
        goto err;
    }

    /* Install */
out_notbl:
    proc_req->shanon_cnt = count;
    proc_req->shanon_tbl = desc_tbl;

err:
    return retval;
}

static struct file *cr_regenerate(struct file *cr_filp, const struct cr_shanon_desc *desc, loff_t size) {
    int mode, flags;
    struct file *filp;
    const char *name;

    mode = 0;
    if (desc->flags & VM_MAYREAD)  mode |= S_IRUSR;
    if (desc->flags & VM_MAYWRITE) mode |= S_IWUSR;
    if (desc->flags & VM_MAYEXEC)  mode |= S_IXUSR;

    switch (desc->flags & (VM_MAYREAD|VM_MAYWRITE)) {
    case VM_MAYREAD:	 flags = O_RDONLY; break;
    case VM_MAYWRITE:	 flags = O_WRONLY; break;
    default:		 flags = O_RDWR; break;
    }

    /* read original filename from context file */
    name = cr_getname(cr_filp, 0);
    if (IS_ERR(name)) {
	filp = (struct file *)name;
	goto out_nopage;
    }

    filp = cr_mkunlinked(cr_filp, name, mode, flags, size, (unsigned long)desc->shanon_id);
    if (IS_ERR(filp)) {
	CR_ERR("cr_mkunlinked returned error %d", (int)PTR_ERR(filp));
    }

    __putname(name);
out_nopage:
    return filp;
}

/* Read the shared-anonymous data from the context file
 *
 * Returns 0 on success, or <0 on error
 */
long
cr_load_shanon_data(cr_rstrt_proc_req_t *proc_req)
{
    const int count = proc_req->shanon_cnt;
    const struct cr_shanon_desc *desc = proc_req->shanon_tbl;
    struct mm_struct *mm = current->mm;
    int retval, i;

    retval = 0;
    for (i=0; i<count; ++i, ++desc) {
        struct file *filp = NULL;
	unsigned long flags = MAP_FIXED | MAP_SHARED;
	unsigned long len = desc->end - desc->start;
	unsigned long map_addr;
	int found, prot = 0;
	    
	prot = 0;
        if (desc->flags & VM_READ)  prot |= PROT_READ;
        if (desc->flags & VM_WRITE) prot |= PROT_WRITE;
        if (desc->flags & VM_EXEC)  prot |= PROT_EXEC;
        if (desc->flags & VM_DENYWRITE) flags |= MAP_DENYWRITE;

	/* "shanon_id" was checkpoint-time inode, but we map to the first-restored filp */
        found = cr_find_object(proc_req->req->map, desc->shanon_id, (void **)&filp);

	if (found) {
	    /* Nothing to prepare */
	} else if (desc->type != CR_SHANON_SHMEM) {
	    /* create (and populate unless hugetlb) before mmap() */
#ifdef CONFIG_HUGETLBFS
	    loff_t tmp = (desc->type == CR_SHANON_FILE) ? desc->i_size : 0;
#else
	    const loff_t tmp = desc->i_size;
#endif
	    filp = cr_regenerate(proc_req->file, desc, tmp);
	    if (IS_ERR(filp)) {
		retval = PTR_ERR(filp);
		CR_ERR("cr_regenerate returned %d", retval);
		goto err;
	    }

#ifdef CONFIG_HUGETLBFS
	    if (desc->type == CR_SHANON_HUGETLB) {
	        /* mmap() will extend the file */
	        DOWN_WRITE_MMAP_SEM(mm);
	        map_addr = do_mmap(filp, desc->start, len, prot|PROT_WRITE, flags, 0);
	        UP_WRITE_MMAP_SEM(mm);
	    
	        /* read in the data */
	        tmp = cr_uread(proc_req->file, (void *)desc->start, len);
	        if (tmp < len) {
		    CR_ERR("short read %ld while regenerating hugetlbfs file", (long)tmp);
		    retval = EIO;
		    goto err;
	        }
	    }
#endif
	} else {
	    const int mmap_prot = PROT_READ|PROT_WRITE|PROT_EXEC;
            struct vm_area_struct *map;
	    loff_t dst_pos = 0;
	    loff_t r;

	    /* mmap() at offset 0, even if it needs to move later */
	    DOWN_WRITE_MMAP_SEM(mm);
	    map_addr = do_mmap(NULL, desc->start, len, mmap_prot, flags, 0);
            map = find_vma(mm, desc->start);
	    filp = (map && (map->vm_start == desc->start)) ? map->vm_file : NULL;
	    UP_WRITE_MMAP_SEM(mm);
	    if (!filp) {
		goto fail;
	    }

	    /* Populate *after* mmap() */
	    r = cr_copyfile(filp, &dst_pos, proc_req->file, NULL, desc->i_size);
	    if (r < desc->i_size) {
		CR_ERR("short read in copy-in of anonymous mapping");
		retval = -EIO;
		goto err;
	    }

	    /* Fall through to 2nd mmap to ensure correct offset and protection */
	    get_file(filp); /* So (re)mmap doesn't drop it */
	}

        DOWN_WRITE_MMAP_SEM(mm);
	map_addr = do_mmap(filp, desc->start, len, prot, flags, (desc->pgoff << PAGE_SHIFT));
	UP_WRITE_MMAP_SEM(mm);

        if (!found) {
            (void)cr_insert_object(proc_req->req->map, desc->shanon_id, (void *)filp);
	    fput(filp);
	}

	if (!filp || (map_addr != desc->start)) {
fail:
	    CR_ERR("Failed to locate newborn shared anonymous space");
	    retval = -EINVAL;
	    goto err;
	}
    }

err:
    return retval;
}

/* Save the shared-anonymous mappings metadata for the current mm.
 * They are both written to the context file, and saved in the proc_req
 * for later use.
 *
 * Returns bytes written (>=0), or <0 on error
 */
long
cr_save_shanon_maps(cr_chkpt_proc_req_t *proc_req)
{
    struct mm_struct *mm = current->mm;
    struct vm_area_struct *map;
    struct cr_shanon_desc *desc_tbl = NULL;
    struct file **file_tbl = NULL;
    size_t size = 0;
    int i, count = 0;
    int retval;

    DOWN_READ_MMAP_SEM(mm);
    /* First pass just counts maps ... */
    for (map = mm->mmap; map ; map = map->vm_next) {
	count += ((map->vm_flags & VM_SHARED) && map->vm_file &&
		  vmad_dentry_unlinked(map->vm_file->f_dentry));
    }
    /* ... then we allocate the table ... */
    if (!count) goto out_notbl;
    size = count * sizeof(struct cr_shanon_desc);
    desc_tbl = vmalloc(size);
    file_tbl = vmalloc(count*sizeof(*file_tbl));
    retval = -ENOMEM;
    if (!desc_tbl || !file_tbl) {
        UP_READ_MMAP_SEM(mm);
	vfree(desc_tbl);
	goto err;
    }
    /* ... and the second pass fills the table. */
    for (i = 0, map = mm->mmap; (i < count) && map ; map = map->vm_next) {
	struct file *filp = map->vm_file;
	if ((map->vm_flags & VM_SHARED) && filp && vmad_dentry_unlinked(filp->f_dentry)) {
	    if (filp->f_op == &shmem_file_operations) {
		desc_tbl[i].type = CR_SHANON_SHMEM;
#ifdef CONFIG_HUGETLBFS
	    } else if (is_file_hugepages(filp)) {
		desc_tbl[i].type = CR_SHANON_HUGETLB;
#endif
	    } else {
		desc_tbl[i].type = CR_SHANON_FILE;
	    }
	    desc_tbl[i].shanon_id = filp->f_dentry->d_inode;
	    desc_tbl[i].i_size    = cr_i_size_read(filp->f_dentry->d_inode);
	    desc_tbl[i].start     = map->vm_start;
	    desc_tbl[i].end       = map->vm_end;
	    desc_tbl[i].flags     = map->vm_flags;
	    desc_tbl[i].pgoff     = map->vm_pgoff;
	    file_tbl[i] = filp;
	    ++i;
	}
    }
out_notbl:
    proc_req->shanon_cnt = count;
    proc_req->shanon_tbl = desc_tbl;
    UP_READ_MMAP_SEM(mm);

    /* write the size in bytes */
    retval = cr_kwrite(proc_req->file, &size, sizeof(size));
    if (retval < sizeof(size)) {
        CR_ERR("short write: %d of %d bytes", retval, (int)sizeof(size));
        retval = -EIO;
        goto err;
    }

    /* write the table (if any) */
    if (size) {
	long w = cr_kwrite(proc_req->file, desc_tbl, size);
        if (w < size) {
            CR_ERR("short write: %ld of %ld bytes", w, (long)size);
            retval = -EIO;
            goto err;
        }
        retval += w;
    }

    /* We've saved the (struct inode *) as the "shanon_id", because only it
     * is unique in the case of multiple open()s of the same file.
     * However, we need the (struct file *) later to help us save the data.
     * So, we replace the in-memory shanon_id field w/ the filp.
     */
    for (i = 0; i < count; ++i) {
	desc_tbl[i].shanon_id = file_tbl[i];
    }

err:
    vfree(file_tbl);
    return retval;
}

/* Save the shared-anonymous mapped data for the current mm.
 *
 * Returns bytes written (>=0), or <0 on error
 */
loff_t
cr_save_shanon_data(cr_chkpt_proc_req_t *proc_req)
{
    const int count = proc_req->shanon_cnt;
    const struct cr_shanon_desc *desc = proc_req->shanon_tbl;
    loff_t w, retval;
    int i;

    retval = 0;
    for (i = 0; i < count; ++i, ++desc) {
	struct file *filp = (struct file *)desc->shanon_id;
	struct inode *inode = filp->f_dentry->d_inode;
	loff_t size = desc->i_size;
	loff_t src_pos = 0;

	/* NOTE: we currently rely on the restore order matching the save order */
        if (cr_insert_object(proc_req->req->map, inode, (void *)1UL)) {
	    /* Somebody dumped this one already */
	    continue;
	}

	if (desc->type != CR_SHANON_SHMEM) {
	    w = cr_save_filename(proc_req->file, filp, NULL, 0);
	    if (w < 0) {
	        retval = -EIO;
	        goto err;
	    }
	    retval += w;
	}

#ifdef CONFIG_HUGETLBFS
	if (desc->type == CR_SHANON_HUGETLB) {
	    unsigned long len = (desc->end - desc->start);
	    if (len < size) {
		CR_WARN("[%d] May fail to restore incomplete ANONYMOUS hugetlbfs mapping "
			"(size=%lu, mapped=%lu)", current->pid, (unsigned long)size, len);
	    }

	    w = cr_uwrite(proc_req->file, (void *)desc->start, size);
	    if (w != size) {
	        retval = -EIO;
	        goto err;
	    }
	} else
#endif
	{
	    w = cr_copyfile(proc_req->file, NULL, filp, &src_pos, size);
	    if (w != size) {
	        retval = -EIO;
	        goto err;
	    }
	}
        retval += w;
    }

err:
    return retval;
}
