/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_core.c,v 1.168.2.6 2007/12/14 21:24:40 phargrov Exp $
 */

#define _LARGEFILE64_SOURCE 1   /* For O_LARGEFILE */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ucontext.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <dlfcn.h>
#include <features.h>

// XXX: needed as long as we do some save/restore work in user space:
#include <termios.h>	// XXX: for termios
#include <sys/param.h>	// XXX: for getcwd() call
#include <sys/time.h>	// XXX: for set/getitimer

#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#include "cr_private.h"

//
// Private variables
//

//
// Public variables
//

// How many threads are actively trying to checkpoint (PENDING + ACTIVE)
cri_atomic_t cri_live_count = {0,};

// Red-black lock for critical sections
cri_rb_lock_t cri_cs_lock = CRI_RB_LOCK_INITIALIZER;

// Table of hooks
volatile cr_hook_fn_t cri_hook_tbl[CR_NUM_HOOKS];

//
// Private functions
//

// Tell the kernel we are ready to be saved
// This version just invokes cri_syscall_chkpt(flags)
static int do_checkpoint(int flags)
{
    static cri_atomic_t counter = {1,};
    int first_thread = cri_atomic_dec_and_test(&counter);
    int rc;
#if CRI_DO_TERMIOS
    struct termios ti;
#endif
    struct itimerval    real_timer;
    struct itimerval    virt_timer;
    struct itimerval    prof_timer;
    const struct itimerval zero_timer = {
	    it_interval:	{ tv_sec: 0, tv_usec: 0},
	    it_value:		{ tv_sec: 0, tv_usec: 0},
    };

    // "GLOBAL" CHECKPOINT STUFF HERE:

    // "Pause" the itimers from exactly one thread
    if (first_thread) {
	setitimer(ITIMER_REAL,    &zero_timer, &real_timer);
	setitimer(ITIMER_VIRTUAL, &zero_timer, &virt_timer);
	setitimer(ITIMER_PROF,    &zero_timer, &prof_timer);
    }

#if CRI_DO_TERMIOS
    // Save the termios structure
    do {
	int tty = __cri_open("/dev/tty", O_WRONLY | O_NOCTTY, 0, NULL);
	
	if (tty >= 0) {
	    rc = tcgetattr(tty, &ti);
	    if (rc < 0) {
		CRI_ABORT("Trouble saving termios");
	    }
	    (void)__cri_close(tty, NULL);
	}
    } while (0);
#endif

    // This does the actual kernel call to take the checkpoint:
    rc = cri_syscall_chkpt(flags);
    if (rc < 0) {
	return -errno;
    }

    if (rc) {
	// "GLOBAL" RESTART STUFF HERE

#if CRI_DO_TERMIOS
	// Restore the termios structure
	do {
	    int tty = __cri_open("/dev/tty", O_WRONLY | O_NOCTTY, 0, NULL);
	
	    if (tty >= 0) {
		int err = tcsetattr(tty, TCSANOW, &ti);
		if (err < 0) {
		    CRI_ABORT("Trouble restoring termios");
		}
		(void)__cri_close(tty, NULL);
	    }
	} while (0);
#endif
    } else {
	// "GLOBAL" CONTINUE STUFF HERE
    }

    // Restore the itimers from the thread that saved them
    if (first_thread) {
	setitimer(ITIMER_REAL,    &real_timer, NULL); // XXX: should adjust via gettimeofday()?
	setitimer(ITIMER_VIRTUAL, &virt_timer, NULL);
	setitimer(ITIMER_PROF,    &prof_timer, NULL);
    }

    cri_atomic_inc(&counter);

    return rc;
}

static int abort_checkpoint(int flags)
{
    int rc;

    rc = cri_syscall(CR_OP_HAND_ABORT, flags);

    return rc;
}

// enter_idle_state(info)
//
// Transition to CR_STATE_IDLE, regardless of starting state
static inline void enter_idle_state(cri_info_t *info)
{
    cri_atomic_write(&info->cr_state, CR_STATE_IDLE);
    cri_atomic_write(&info->cr_cs_count, 1);
}

#ifdef CRI_SA_RESTORER
  CRI_SA_RESTORER
#endif

// Signal handler to dispatch the request
//
// XXX: This will eventually all be done by the kernel which will
// invoke the proper handler directly w/o using a signal number.
// Must NOT be static because we need to be able to dlsym() it
// when trying to mix static and shared libs.
//
// WARNING: Must only be called as a signal handler.
// Any other calls may not return correctly (see final lines).
/* static */ void cri_sig_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
#if LIBCR_TRACING
    int pid = (int)getpid();
#endif
    cri_info_t *info = cri_info_location();	// thread-specific

    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] signal %d received", pid, signr);

#ifdef SI_KERNEL
    if (siginfo->si_code != SI_KERNEL) {
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] Ignoring false signal", pid);
	goto out;
    }
#endif

    if (!info || !info->handler) {
	int rc, last;

	// No state and/or handler, so do this the "easy way"
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	cri_atomic_inc(&cri_live_count);
	rc = do_checkpoint(0);
	last = cri_atomic_dec_and_test(&cri_live_count);
	if (last && !cri_atomic_read(&cri_initialized)) {
	    // No statefull threads.  So, last one out closes the connection to the kernel.
	    cri_disconnect();
	}
	if (rc < 0) {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] do_checkpoint failed w/ rc=%d", pid, rc);
	} else {
	    CRI_RUN_HOOK(rc ? CR_HOOK_RSTRT_NO_CALLBACKS : CR_HOOK_CONT_NO_CALLBACKS);
	}
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else {
	// Run the handler appropriate to this thread
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] running handler %p", pid, info->handler);
	info->handler(signr, siginfo, context);
    }

// XXX: This mechanism for telling the userspace to re-block the
// signal has been disabled.  Because of the delay between generation
// and delivery, it was possible that the kernel cleared the bit in a
// signal mask that is now no longer current.
// The result (observed rarely in the replace_cb test on hive8 with
// a vanilla 2.6.21 kernel) was that the signal could end up blocked
// and pending while libcr is spinning in do_enter(), expecting the
// signal handler to run.  That yields deadlock.
#if 0
    // If kernel had to unblock the signal, we arrange to re-block it.
    // The otherwise unused si_pid of the siginfo contains a boolean.
    if (siginfo->si_pid) {
	ucontext_t *uc = (ucontext_t *)context;
	sigaddset(&uc->uc_sigmask, CR_SIGNUM);
    }
#endif

out:
    errno = saved_errno;

    // Ugh!! Bug 2003:
    // If an application calls sigaction, saves the sa_handler and sa_flags
    // and later restores them both (w/ same or different mask), one would
    // think things were OK, right?  No, since we have set the SA_RESTORER
    // bit in sa_flags, they would need to save/restore the sa_restorer
    // field as well.  If they copy the flag but leave the sa_restorer
    // uninitialized, then who knows where this signal handler might return
    // to!?!  This is not theoretical; this actually happens w/ guile.
#if defined(cri_restorer_jump)
    cri_restorer_jump(signr, siginfo, context, cri_sa_restorer);
#elif defined(CRI_SA_RESTORER)
    // No fix (yet) for Bug 2003 on this platform
#else
    // Bug 2003 does not affect this platform
#endif
}

static void cri_request(int fd, const char *filename, int scope)
{
    /* Static data: */
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    static char prev_name[PATH_MAX];
    static cr_checkpoint_handle_t cr_handle = (cr_checkpoint_handle_t)(-1);

    int rc;
    cr_checkpoint_args_t cr_args;

    /* Serialize */
    pthread_mutex_lock(&lock);

    cr_initialize_checkpoint_args_t(&cr_args);
    cr_args.cr_scope = scope;
    cr_args.cr_fd    = fd;

    // If there is a previous request un-reaped then reap it now.
    // XXX: redo in terms of poll?
    if ((int)cr_handle >= 0) {
      do {
        rc = cri_syscall_token((int)cr_handle, CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG);
        if (rc < 0) {
            if (errno == EINVAL) {
                // No request was pending - not an error.
                break;
            } else if (errno == EAGAIN) {
                // Wait for the request to complete
                // If interrupted by a signal the loop will try again
                cri_syscall_token((int)cr_handle, CR_OP_CHKPT_DONE, (uintptr_t)NULL);
            } else if (prev_name[0]) {
                // Failed - cleanup file from previous failure
                LIBCR_TRACE(LIBCR_TRACE_INFO, "Removing supposedly bogus context file "
                            "(%s), because rc=%d and errno=%d (%s)",
                            prev_name, rc, errno, cr_strerror(errno));
		if (!strcmp(prev_name, filename)) {
		    /* Don't remove the NEW file */
                } else if (unlink(prev_name) < 0) {
                    CRI_ABORT("unable to remove aborted/failed checkpoint file %s: %s\n",
                              prev_name, strerror(errno));
                }
            }
	}
      } while (rc < 0);
      (void)cri_disconnect_token((int)cr_handle);
      cr_handle = (cr_checkpoint_handle_t)(-1);
    }

    rc = cr_request_checkpoint(&cr_args, &cr_handle);
    if (rc) {
        CRI_ABORT("cr_request_checkpoint returned %d w/ errno=%d (%s)\n", rc, errno, cr_strerror(errno));
    }

    // Don't return until the request is either pending or completed.
    // This is important; otherwise we would have races with critical
    // sections which follow the cr_request() call.
    //
    // We can detect pending state because the live count is non-zero.
    // We can detect completed because a reap will succeed (or fail
    // with an errno other than EAGAIN).
    // When restarting we expect the reap to fail w/ errno = EINVAL,
    // (if the fd is restored, since there won't be any checkpoint
    // request anymore) or with EBADF if the fd is not restored.
    rc = 0;
    CRI_WHILE_COND_YIELD(
	!cri_atomic_read(&cri_live_count) &&
	((rc = cri_syscall_token((int)cr_handle, CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG)) < 0) &&
	errno == EAGAIN
    );
    LIBCR_TRACE(LIBCR_TRACE_INFO,
		"Reap loop finished with rc=%d and errno=%d (%s)",
		rc, errno, cr_strerror(errno));

    if (filename) {
        /* Remove checkpoint file if an unexpected error occurred (see above) */
        if (rc && (errno != EAGAIN) && (errno != EINVAL) && (errno != EBADF)) {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "Removing supposedly bogus context file "
		        "(%s), because rc=%d and errno=%d (%s)",
		        filename, rc, errno, cr_strerror(errno));
	    if (unlink(filename) < 0) {
	        CRI_ABORT("unable to remove aborted/failed checkpoint file %s: %s\n", 
		          filename, strerror(errno));
	    }
	    prev_name[0] = '\0';
	} else {
	    /* Save name for possible later clean up */
	    strncpy(prev_name, filename, PATH_MAX);
	}
    } else {
	prev_name[0] = '\0';
    }

    pthread_mutex_unlock(&lock);
}

void cri_initialize_checkpoint_args_t(cr_checkpoint_args_t *cr_args)
{
    cr_args->cr_scope   = CR_SCOPE_PROC;
    cr_args->cr_target  = 0;
    cr_args->cr_fd      = -1;
    cr_args->cr_signal  = 0;
    cr_args->cr_timeout = 0;    /* 0 == unbounded */
    cr_args->cr_flags   = 0;
}

//
// Public functions
//

cr_hook_fn_t cr_register_hook(cr_hook_event_t event, cr_hook_fn_t fn)
{
    cr_hook_fn_t old_fn = NULL;
    /* Static data: */
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

    if ((event >= 0) && (event < CR_NUM_HOOKS)) {
	pthread_mutex_lock(&lock);
	old_fn = cri_hook_tbl[event];
	cri_hook_tbl[event] = fn;
	pthread_mutex_unlock(&lock);
    }                          
    
    return old_fn;
}

int cr_request_checkpoint(cr_checkpoint_args_t *args,
                          cr_checkpoint_handle_t *handle)
{
    /* can probably ditch the lock */
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    struct cr_chkpt_req req;
    int rc;
    int pid;

    pid = getpid();

    /* Serialize */
    pthread_mutex_lock(&lock);

    rc = -3;
    if (args->cr_version != LIBCR_VERSION_ID) {
        goto out;
    }

    // initialize the handle
    rc = cri_connect_token();
    if (rc < 0) {
        goto out;
    }
    *handle = rc;

    // initialize the request
    req.cr_target = args->cr_target;
    req.cr_scope  = args->cr_scope;
    req.cr_fd     = args->cr_fd;
    req.cr_secs   = args->cr_timeout;
    req.cr_opts   = NULL;
    req.dump_format = cr_format_vmadump;
    req.signal    = args->cr_signal;
    req.flags     = args->cr_flags;

    // issue the request
    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] requesting the checkpoint.", pid);
    rc = cri_syscall_token(*handle, CR_OP_CHKPT_REQ, (uintptr_t)&req);
    if (rc < 0) {
        goto out;
    }

out:
    pthread_mutex_unlock(&lock);
    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] leaving with rc=%d.", pid, rc);
    return rc;
}

int cr_forward_checkpoint(cr_scope_t cr_scope, pid_t cr_target)
{
    int rc;
    int pid;
    struct cr_chkpt_req req;

    pid = getpid();

    req.cr_scope  = cr_scope;
    req.cr_target = cr_target;

    // forward the checkpoint
    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] forwarding checkpoint.", pid);
    rc = cri_syscall(CR_OP_CHKPT_FWD, (uintptr_t)&req);
    if (rc < 0) {
        goto out;
    }

out:
    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] leaving with rc=%d.", pid, rc);
    return rc;
}

// Block until the checkpoint request is complete or we exceed the timeout
//
// XXX: Should we loop if we EINTR?  Or just make the caller do that?
int cr_poll_checkpoint(cr_checkpoint_handle_t *handle, struct timeval *timeout)
{
    int token = *handle;
    int err;

    /* wait for the request to complete */
    err = cri_syscall_token(token, CR_OP_CHKPT_DONE, (uintptr_t) timeout);
    if (err > 0) {
	/* fall through to REAP call */
    } else if (err == 0) {
        /* (err == 0) is "not done" */
        goto out_not_done;
    } else {
        /* (err < 0) is an actual error condition */
        err = CR_POLL_CHKPT_ERR_PRE;
        goto out;
    }

    /* reap the completed request */
    err = cri_syscall_token(token, CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG);
    if (err < 0) {
	err = CR_POLL_CHKPT_ERR_POST;
        goto out;
    }

    err = 1; /* >0 indicates done */

out:
    if ((err < 0) && ((errno == ENOTTY)	/* "Bad ioctl()" - this is NOT our handle */
		   || (errno == EINTR)	/* "Interrupted" - keep handle for caller to retry */
		     )) {
	/* Don't disconnect */
    } else {
        (void)cri_disconnect_token(token); /* Might silently fail? */
	*handle = (cr_checkpoint_handle_t)(-1);
    }

out_not_done:
    return err;
}


// Request an application-initiated checkpoint
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
void cr_request_fd(int fd)
{
    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    cri_request(fd, NULL, CR_SCOPE_PROC);
}

// Request an application-initiated checkpoint of own process
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
// XXX: error handling needs work (ABORT is not the best policy)
void cr_request_file(const char *filename)
{
    int fd;

    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    CRI_EINTR_LOOP((fd = __cri_open(filename,
				    O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE,
				    0600, &errno)));
    if (fd < 0) {
        CRI_ABORT("open(%s)", filename);
    }
    cri_request(fd, filename, CR_SCOPE_PROC);
    CRI_EINTR_LOOP(__cri_close(fd, &errno));        /* silent failure possible */
}

// Request an application-initiated checkpoint
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
void cr_request(void)
{
#if 1 // Until VMAdump code deals w/ directory targets properly
#define CONTEXT_NAMELEN 16 /* context.XXXXX + \0 + extra */
    char filename[CONTEXT_NAMELEN];
    snprintf(filename, CONTEXT_NAMELEN, "context.%d", (int)getpid());
    cr_request_file(filename);
#else
    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    cri_request(CR_DEST_CWD, NULL, CR_SCOPE_PROC);
#endif
}

int cr_checkpoint(int flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int my_id;
    int outer_most;
    int index;
    int retval;
    int rc;
#if LIBCR_TRACING
	int pid = (int)getpid();
#endif

    // Check for improper calls
    if (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE) {
	CRI_ABORT("STATE ERROR.  Called while not in the ACTIVE state.");
    }

    // Save callers id
    my_id = info->run.index;
    outer_most = (my_id == info->cr_cb_count);

    // Loop calling callbacks until they have all been called and then
    // invoke do_checkpoint() to do the real work.
    //
    // This strange loop ensures that after the callback with (index == 0)
    // we next call do_checkpoint() even if callback0 is NULL or returns
    // without calling back into cr_checkpoint().
    do {
	if (flags & CR_CHECKPOINT_ABORT_MASK) {
	    // We've detected failure of the invoking callback
	    // We we must now call the kernel and stop stacking of callbacks.
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CANCELLING CHECKPOINT (flags=%d)",
			pid, flags);
	    if (abort_checkpoint(flags) == -1)
		info->run.rc = -errno;
	    index = info->run.index = -1;
	    break;
	}
    
	index = --info->run.index;
	if (index >= 0) {
	    // Call the next callback, skipping NULLs

	    if (info->cr_cb[index].func != NULL) {
		info->run.id = index;
		rc = (*info->cr_cb[index].func)(info->cr_cb[index].arg);
		if (rc) {
		    CRI_ABORT("Callback %d returned %d - ABORTING\n",
			      index, rc);
		}
	    }
	} else {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CHECKPOINTING (flags=%d)",
			pid, flags);
	    info->run.rc = do_checkpoint(flags);
	    if (info->run.rc < 0) {
		LIBCR_TRACE(LIBCR_TRACE_INFO,
			    "[%d] do_checkpoint() FAILED", pid);
	    } else if (info->run.rc) {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] RESTORING", pid);
	    } else {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CONTINUING", pid);
	    }
	}
    } while (info->run.index >= 0);

    // Once the outer-most invocation calls CR_OP_HAND_{CONT,RSTRT}
    // the kernel might checkpoint us again.  Save the return value now
    // because a second checkpoint could modify info->run.rc.
    retval = info->run.rc;

    /* Restore saved id for possible use by the caller (in replace_self). */
    info->run.id = my_id;

    // Perform cleanup only on the final/outer-most return
    if (outer_most) {
	enter_idle_state(info);
	cri_atomic_dec_and_test(&cri_live_count);
	if (retval > 0) {
	    CRI_RUN_HOOK(info->is_thread ? CR_HOOK_RSTRT_THREAD_CONTEXT
                                         : CR_HOOK_RSTRT_SIGNAL_CONTEXT);
	    rc = cri_syscall(CR_OP_HAND_RSTRT, CRI_SYSCALL_NOARG);
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_RSTRT) failed w/ rc=%d errno=%d", rc, errno);
	    }
	} else if (retval == 0) {
	    CRI_RUN_HOOK(info->is_thread ? CR_HOOK_CONT_THREAD_CONTEXT
                                         : CR_HOOK_CONT_SIGNAL_CONTEXT);
	    rc = cri_syscall(CR_OP_HAND_CONT, CRI_SYSCALL_NOARG);
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_CONT) failed w/ rc=%d errno=%d", rc, errno);
	    }
	}
	// XXX: We spin here for live count to reach zero because there is currently a small
	// race against the possibility of a NO-PHASE thread which doesn't decrement the
	// live count until after the do_checkpoint "postdump_barrier" is complete.  That
	// can lead to cr_request,cr_enter_cs,cr_leave_cr,rc=cr_status producing rc=PENDING.
	// XXX: DISABLED!  There is another race created by this fix!  The next checkpoint
	// request can arrive at any point after the CR_OP_HAND_{CONT,RSTRT} ioctl call is
	// entered.  Therefore we could end up spining here forever as a result of the
	// "next" rather than "prev" checkpoint request.
	// CRI_WHILE_COND_YIELD(cri_atomic_read(&cri_live_count));
    }

    return retval;
}

//
// Internal functions
//

// cri_checkpoint_info_init()
//
// Initialize a thread-local struct cr_checkpoint_info
void cri_checkpoint_info_init(cri_info_t *info)
{
    info->cr_checkpoint_info.dest = NULL;	// filled-in on demand
    info->cr_restart_info.src = NULL;		// filled-in on demand
}

// cri_info_free()
//
// dtor for thread-specific data
void
cri_info_free(void *info)
{
    if (info) {
	free(info);
	pthread_setspecific(cri_info_key, NULL);
    }
}


// cri_info_init()
//
// Routine to initialize the list of registered callbacks and other data
// On first call per thread, allocates the thread-specific cri_info structure.
cri_info_t* cri_info_init(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int rc;

    if (!info) {
	info = malloc(sizeof(cri_info_t));
	if (!info) {
	    CRI_ABORT("Failed to allocate cri_info structure");
	}

	memset(info, 0, sizeof(cri_info_t));
	enter_idle_state(info);
	info->handler = NULL;
	info->next_id = 0;
	info->is_thread = 0;

	rc = pthread_setspecific(cri_info_key, info);
	if (rc != 0) {
	    CRI_ABORT("pthread_setspecific() returned %d", rc);
	}
    }

    return info;
}

// cri_start_checkpoint()
//
// Routine to process the list of registered callbacks
void cri_start_checkpoint(cri_info_t *info)
{
    if (cri_cmp_swap(&info->cr_state, CR_STATE_PENDING, CR_STATE_ACTIVE)) {
	// State is valid, so start running callbacks
#if LIBCR_TRACING
	int pid = (int)getpid();
#endif
	int rc;
	info->run.index = info->cr_cb_count;
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	rc = cr_checkpoint(0);
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else if (!info->is_thread) {
	CRI_ABORT("STATE ERROR.  Probably an excess call to cr_leave_cs().");
    } else {
	// False wake up of a checkpoint thread - not an error.
    }
}

/* Initialization logic.
 *
 * XXX: need a balancing _fini()?
 */
static int raninit = 0;

static int cri_sigaction(int signum, const struct sigaction *act, struct sigaction *oact) {
#if CR_USE_SIGACTION
    return sigaction(signum, act, oact);
#else
    struct k_sigaction ksa, oksa;
    int rc;

    if (act) {
      ksa.ksa_sigaction = act->sa_sigaction;
      memcpy(&ksa.ksa_mask, &act->sa_mask, sizeof (sigset_t));
      ksa.ksa_flags = act->sa_flags;
      #if defined(CRI_SA_RESTORER)
        ksa.ksa_flags |= SA_RESTORER;
        ksa.ksa_restorer = &cri_sa_restorer;
      #endif
    }
    rc = __cri_ksigaction(signum, (act ? &ksa : NULL), (oact ? &oksa : NULL), (_NSIG/8), &errno);
    if (oact) {
      oact->sa_sigaction = oksa.ksa_sigaction;
      memcpy(&oact->sa_mask, &oksa.ksa_mask, sizeof (sigset_t));
      oact->sa_flags = oksa.ksa_flags;
      oact->sa_restorer = oksa.ksa_restorer;
    }
    return rc;
#endif
}


static void _do_init(void *replace)
{
    int rc;
    struct sigaction sa;
    int signum;

    //
    // Initialize tracing
    //
    LIBCR_TRACE_INIT();

    //
    // Initialize pthread-dependent parts if appropriate
    //
    cri_pthread_init();

    //
    // Setup signal handler
    //
    if (replace) {
	/* Check that we're registered already and then reregister */
	rc = cri_sigaction(CR_SIGNUM, NULL, &sa);
	if (sa.sa_sigaction != replace) {
	    CRI_ABORT("Failed to reregister signal %d in process %d from static constructor",
		      CR_SIGNUM, (int)getpid());
	}
    } else if (CR_SIGNUM != (signum = __libc_allocate_rtsig(0))) {
	CRI_ABORT("Failed to allocate signal %d in process %d: got signal %d instead",
		  CR_SIGNUM, (int)getpid(), signum);
    }
    sa.sa_sigaction = cri_sig_handler;
    sa.sa_flags = SA_RESTART | SA_SIGINFO;
    rc = sigfillset(&sa.sa_mask);
    if (rc != 0) {
	CRI_ABORT("sigfillset() failed: %s", strerror(errno));
    }
    rc = cri_sigaction(CR_SIGNUM, &sa, NULL);
    if (rc != 0) {
	CRI_ABORT("sigaction() failed: %s", strerror(errno));
    }
}

/* Initialization entry point.
 *
 * If shared lib, compiled with -DPIC, and use magic '_init' function, which
 * is always loaded just once.
 *
 * If compiled statically, use the 'constructor' attribute to run at
 * startup.
 *
 * TODO:  This doesn't completely solve the problem:  
 *  1)  When statically linked, if no other functions from this library used,
 *  linker won't link the library into the app, and so constructor not run.  
 *	If linking libcr and not calling any functions, then you get what you
 *	deserve.  However, since every client of libcr must call cr_init(),
 *	we need only ensure cr_init() will reference something in cr_core.o.
 *	In fact, cr_init() calls cri_info_init() which is in cr_core.o.
 *	Thus we are certain that any client that actually uses the library
 *	will include the constructor.  This issue is solved.
 *  2) If statically linked AND started with cr_run, init function will be run
 *  twice (and with dynamic lib having its own copy of all data structures).
 *	This is now (almost) solved by ensuring that only the statically linked
 *	version is ever run.  This works because the shared version will
 *	run its initializer first and the static one will overwrite the
 *	signal handler with the one in the static library.  Thus nobody
 *	will ever invoke code in the shared library once the static lib's
 *	constructor has run.  Since this is certain to happen before main(),
 *	we are certain that in the unlikely case that a checkpoint is
 *	requested before the static lib's constructor runs, no callbacks
 *	will be missed. XXX: untested
 *  3) We are unable to interpose on pthread_create, fork or vfork.
 *
 *  For now we still build only shared libs.
 */
#ifdef PIC
void _init(void)
{
    if (!raninit++)
	_do_init(NULL);
}
#else
static void __attribute__((constructor)) _cr_static_init(void)
{
    if (!raninit++) {
	// dlsym(NULL, ) segfaults for a static executable.
	// Use dlopen first to test that first.
	void *handle = dlopen(NULL, RTLD_LAZY);
	if (handle) {
	    _do_init(dlsym(handle, "cri_sig_handler"));
	    dlclose(handle);
	} else {
	    _do_init(NULL);
	}
    }
}
#endif
