/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2007, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cb_exit.c,v 1.1.2.2 2008/01/04 01:19:14 phargrov Exp $
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>

#include "libcr.h"
#include "crut_util.h"

static char filename[40];

enum {
	MSG_CHILD_READY = 42,
	MSG_CHILD_DONE,
	MSG_PARENT_DONE,
	TEST_EXIT_VAL1,
	TEST_EXIT_VAL2,
	TEST_EXIT_VAL3,
	TEST_EXIT_VAL4,
	TEST_EXIT_VAL5
};

// Callback that exits from CHECKPOINT portion of callback
static int cb_chkpt(void* arg)
{
    int count = (uintptr_t)arg;

    printf("%03d enter cb_chkpt\n", count++);
    printf("%03d child exit()\n", count++);
    fflush(NULL);
    exit(TEST_EXIT_VAL1);
    printf("XXX exit() returned unexpectedly\n");
    exit(-1);

    return 1;
}

// Callback that exits from CONTINUE portion of callback
static int cb_cont(void* arg)
{
    int count = (uintptr_t)arg;
    int rc;

    printf("%03d enter cb_cont\n", count++);
    rc = cr_checkpoint(CR_CHECKPOINT_READY);
    if (!rc) {
      printf("%03d child exit()\n", count++);
      fflush(NULL);
      exit(TEST_EXIT_VAL2);
      printf("XXX exit() returned unexpectedly\n");
    } else if (rc > 0) {
      ++count; // match the CONT
      printf("%03d child restarted\n", count++);
      exit(TEST_EXIT_VAL3);
    } else {
      printf("XXX cr_checkpoint() failed\n");
    }

    return 0;
}
//
// Callback that exits from RESTART portion of callback
static int cb_rstrt(void* arg)
{
    int count = (uintptr_t)arg;
    int rc;

    printf("%03d enter cb_rstrt\n", count++);
    rc = cr_checkpoint(CR_CHECKPOINT_READY);
    if (!rc) {
      printf("%03d child continued\n", count++);
      exit(TEST_EXIT_VAL4);
    } else if (rc > 0) {
      ++count; // match the CONT
      printf("%03d child restart\n", count++);
      printf("%03d child exit()\n", count++);
      fflush(NULL);
      exit(TEST_EXIT_VAL5);
      printf("XXX exit() returned unexpectedly\n");
    } else {
      printf("XXX cr_checkpoint() failed\n");
    }

    return 0;
}

void child_main(struct crut_pipes *pipes, int count, int flags, int (*cb)(void*))
{
    cr_callback_id_t cb_id;
    cr_client_id_t my_id;
    int mypid = getpid();

    printf("#ST_ALARM:60\n");
    snprintf(filename, sizeof(filename), "context.%d", mypid);

    my_id = cr_init();
    if (my_id < 0) {
	printf("XXX child %d cr_init() failed, returning %d\n", mypid, my_id);
	exit(-1);
    } else {
	printf("%03d child %d cr_init() succeeded\n", count++, mypid);
    }

    cb_id = cr_register_callback(cb, (void*)(uintptr_t)(count+1), flags);
    if (cb_id < 0) {
	printf("XXX cr_register_callback() unexpectedly returned %d\n", cb_id);
	exit(-1);
    } else {
	printf("%03d child %d cr_register_callback() succeeded\n", count++, mypid);
    }

    crut_pipes_putchar(pipes, MSG_CHILD_READY);
    while (1) pause();

    exit(1);
}

int run_tests(int base, int flags) {
    char cmd[1024];
    struct crut_pipes pipes;
    const char *chkpt_cmd = crut_find_cmd("cr_checkpoint");
    const char *rstrt_cmd = crut_find_cmd("cr_restart");
    int rc;

    // First test: exit in CHECKPOINT portion of a callback
    if (!crut_pipes_fork(&pipes)) {
	child_main(&pipes, base, flags, cb_chkpt);
	exit(1);
    } else {
	snprintf(cmd, sizeof(cmd), "%s --pid %d --kill --file %s 2>/dev/null", chkpt_cmd, pipes.child, filename);
	crut_pipes_expect(&pipes, MSG_CHILD_READY);
	rc = system(cmd);
	if (!WIFEXITED(rc) || (WEXITSTATUS(rc) != ESRCH)) {
	    printf("XXX system(%s) unexpectedly returned %d\n", cmd, rc);
	    exit(1);
	}
	crut_waitpid_expect(pipes.child, TEST_EXIT_VAL1);
    }
    crut_pipes_close(&pipes);
    (void)unlink(filename);

    // Second test: exit in CONTINUE portion of a callback
    if (!crut_pipes_fork(&pipes)) {
	child_main(&pipes, base+4, flags, cb_cont);
	exit(1);
    } else {
	snprintf(cmd, sizeof(cmd), "%s --pid %d --kill --file %s 2>/dev/null", chkpt_cmd, pipes.child, filename);
	crut_pipes_expect(&pipes, MSG_CHILD_READY);
	rc = system(cmd);
	if (!WIFEXITED(rc) || (WEXITSTATUS(rc) != 0)) {
	    printf("XXX system(%s) unexpectedly returned %d\n", cmd, rc);
	    exit(1);
	}
	crut_waitpid_expect(pipes.child, TEST_EXIT_VAL2);
    }
    crut_pipes_close(&pipes);
    snprintf(cmd, sizeof(cmd), "%s %s", rstrt_cmd, filename);
    rc = system(cmd);
    if (!WIFEXITED(rc) || (WEXITSTATUS(rc) != TEST_EXIT_VAL3)) {
	printf("XXX system(%s) unexpectedly returned %d\n", cmd, rc);
	exit(1);
    }
    (void)unlink(filename);

    // Third test: exit in RESTART portion of a callback
    if (!crut_pipes_fork(&pipes)) {
	child_main(&pipes, base+9, flags, cb_rstrt);
	exit(1);
    } else {
	snprintf(cmd, sizeof(cmd), "%s --pid %d --kill --file %s 2>/dev/null", chkpt_cmd, pipes.child, filename);
	crut_pipes_expect(&pipes, MSG_CHILD_READY);
	rc = system(cmd);
	if (!WIFEXITED(rc) || (WEXITSTATUS(rc) != 0)) {
	    printf("XXX system(%s) unexpectedly returned %d\n", cmd, rc);
	    exit(1);
	}
	crut_waitpid_expect(pipes.child, TEST_EXIT_VAL4);
    }
    crut_pipes_close(&pipes);
    snprintf(cmd, sizeof(cmd), "%s %s", rstrt_cmd, filename);
    rc = system(cmd);
    if (!WIFEXITED(rc) || (WEXITSTATUS(rc) != TEST_EXIT_VAL5)) {
	printf("XXX system(%s) unexpectedly returned %d\n", cmd, rc);
	exit(1);
    }
    (void)unlink(filename);

    return 15;
}

int main(void)
{
    int count = 0;
    int mypid = getpid();
	
    setlinebuf(stdout);
    printf("%03d Process started with pid %d\n", count++, mypid);

    snprintf(filename, sizeof(filename), "context_%d", mypid);

    count += run_tests(count, CR_SIGNAL_CONTEXT);
    if (crut_is_linuxthreads()) {
	fprintf(stderr, "Skipping threaded portion of 'cb_exit' tests due to LinuxThreads issues\n");
    } else {
	count += run_tests(count, CR_THREAD_CONTEXT);
    }

    printf("%03d DONE\n", count);

    return 0;
}
