/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: critical_sections.c,v 1.3 2007/07/21 03:03:31 phargrov Exp $
 */


const char description[] = 
"Description of tests/CriticalSections/cr_test:\n"
"\n"
"This test verifies the behavior of critical sections in libcr:\n"
" + cr_enter_cs()\n"
"	+ Enters a critical section which excludes checkpoints, inclusive\n"
"	  of thread-context callbacks.\n"
" + cr_leave_cs()\n"
"	Leaves a critical section, invoking a checkpoint if one is pending.\n"
" + cr_status()\n"
"	Must reflect the status of the thread-context callback as well.\n"
"\n"
"In a successful run, expect lines of output in SORTED order.\n"
;


#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "libcr.h"

static int my_callback(void* arg)
{
    int rc;

    printf("004 enter callback\n");
    rc = cr_checkpoint(CR_CHECKPOINT_READY);
    if (rc) {
	printf("XXX resuming unexpectedly in callback\n");
	exit(-1);
    } else {
	printf("005 continuing correctly in callback\n");
    }

    return 0;
}

int main(void)
{
    pid_t my_pid;
    cr_callback_id_t cb_id;
    cr_client_id_t my_id;
    char filename[40];
    struct stat s;
    int rc;
	
    setlinebuf(stdout);

    my_pid = getpid();
    snprintf(filename, sizeof(filename), "context.%d", my_pid);
    printf("000 Process started with pid %d\n", my_pid);
    printf("#ST_ALARM:120\n");

    my_id = cr_init();
    if (my_id < 0) {
	printf("XXX cr_init() failed\n");
	exit(-1);
    } else {
	printf("001 cr_init() succeeded\n");
    }

    cb_id = cr_register_callback(my_callback, NULL, CR_THREAD_CONTEXT);
    if (cb_id < 0) {
	printf("XXX cr_register_callback() unexpectedly returned %d\n", cb_id);
	exit(-1);
    }

    cr_enter_cs(my_id);
    rc = cr_status();
    if (rc != CR_STATE_IDLE) {
	printf("XXX cr_status() unexpectedly returned %d\n", rc);
	exit(-1);
    } else {
	printf("002 cr_status() correctly returned IDLE\n");
    }

    cr_request();
    // must now be pending in at least one thread

    rc = cr_status();
    if (rc != CR_STATE_PENDING) {
	printf("XXX cr_status() unexpectedly returned 0x%x\n", rc);
	exit(-1);
    } else {
	printf("003 cr_status() correctly returned PENDING\n");
    }

    cr_leave_cs(my_id);	// should cause checkpoint to start

    cr_enter_cs(my_id);	// should wait for checkpoint to finish
    cr_leave_cs(my_id);

    rc = stat(filename, &s);
    if (rc) {
	printf("XXX stat() unexpectedly returned %d\n", rc);
	exit(-1);
    } else {
	printf("006 stat(context.%d) correctly returned 0\n", my_pid);
    }

    if (s.st_size == 0) {
	printf("XXX context file unexpectedly empty\n");
	exit(-1);
    } else {
	printf("007 context.%d is non-empty\n", my_pid);
    }

    (void)unlink(filename); // might fail silently

    printf("008 DONE\n");

    return 0;
}
