/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: pid_in_use.c,v 1.6.26.1 2007/12/17 22:02:11 phargrov Exp $
 */


const char description[] = 
"Description of tests/pid_in_use:\n"
"\n"
"This test verifies the correct behavior when restarting from a context\n"
"corresponding to a still running task (the requester itself in this case).\n"
"In a successful run, expect lines of output in SORTED order.\n"
;


#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#include "libcr.h"
#include "crut_util.h"

enum {
    MSG_CHILD_READY,
    MSG_CHILD_CHKPT,
    MSG_CHILD_DONE,
    MSG_PARENT_DONE
};

static int child_cb(void *arg)
{
    struct crut_pipes *pipes = arg;

    crut_pipes_putchar(pipes, MSG_CHILD_CHKPT);
    return 0;
}

static int parent_cb(void *arg)
{
    struct crut_pipes *pipes = arg;
    crut_pipes_expect(pipes, MSG_CHILD_CHKPT);
    sleep(2); /* Cause a delay to encourage placing our child first in the checkpoint */
    return 0;
}

static void *thread_main(void *arg) 
{
    (void)pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    while(1) { pthread_testcancel(); sleep(5); }
    return NULL;
}

int main(void)
{
    const char *chkpt_cmd = crut_find_cmd("cr_checkpoint");
    const char *rstrt_cmd = crut_find_cmd("cr_restart");
    struct crut_pipes pipes;
    pthread_t th;
    pid_t my_pid;
    cr_client_id_t my_id;
    char filename[40];
    char cmd[1024];
    void *th_result;
    int rc;

    setlinebuf(stdout);

    my_pid = getpid();
    snprintf(filename, sizeof(filename), "context.%d", my_pid);
    printf("000 Process started with pid %d\n", my_pid);
    printf("#ST_ALARM:120\n");

    my_id = cr_init();
    if (my_id < 0) {
	printf("XXX cr_init() failed, returning %d\n", my_id);
	exit(-1);
    } else {
	printf("001 cr_init() succeeded\n");
    }

    /*
     * TEST 1: Just a single-threaded process
     */

    snprintf(cmd, sizeof(cmd), "%s --file %s --pid %d", chkpt_cmd, filename, my_pid);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc) || WEXITSTATUS(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    } else {
	printf("002 cr_checkpoint --pid OK\n");
    }

    snprintf(cmd, sizeof(cmd), "%s %s 2>/dev/null", rstrt_cmd, filename);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    }
    rc = WEXITSTATUS(rc);
    if (rc != EBUSY) {
	printf("XXX restart returned unexpected status %d (%s)\n", rc, strerror(rc));
	exit(-1);
    } else {
	printf("003 restart 1 proc failed in the expected way\n");
    }

    (void)unlink(filename); // might fail silently

    /*
     * TEST 2: A pthreaded process
     */

    if (pthread_create(&th, NULL, &thread_main, NULL) != 0) {
	printf("XXX pthread_create() failed\n");
	exit(-1);
    }

    snprintf(cmd, sizeof(cmd), "%s --file %s --pid %d", chkpt_cmd, filename, my_pid);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc) || WEXITSTATUS(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    } else {
	printf("004 cr_checkpoint --pid/pthread OK\n");
    }

    snprintf(cmd, sizeof(cmd), "%s %s 2>/dev/null", rstrt_cmd, filename);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    }
    rc = WEXITSTATUS(rc);
    if (rc != EBUSY) {
	printf("XXX restart #2 returned unexpected status %d (%s)\n", rc, strerror(rc));
	exit(-1);
    } else {
	printf("005 restart pthreaded proc failed in the expected way\n");
    }

    rc = pthread_cancel(th);
    if (rc) {
	printf("XXX pthread_cancel() returned %d\n", rc);
	exit(-1);
    }
    rc = pthread_join(th, &th_result);
    if (rc) {
	printf("XXX pthread_join() returned %d\n", rc);
	exit(-1);
    }
    if (th_result != PTHREAD_CANCELED) {
	printf("XXX pthread_join() returned unexpected value\n");
	exit(-1);
    }

    /* Now restart again w/o the thread running so we get only a PARTIAL pid conflict */
    snprintf(cmd, sizeof(cmd), "%s %s 2>/dev/null", rstrt_cmd, filename);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    }
    rc = WEXITSTATUS(rc);
    if (rc != EBUSY) {
	printf("XXX restart #3 returned unexpected status %d (%s)\n", rc, strerror(rc));
	exit(-1);
    } else {
	printf("006 restart pthread/partial failed in the expected way\n");
    }

    (void)unlink(filename); // might fail silently

    /*
     * TEST 3: A process tree
     */

    rc = crut_pipes_fork(&pipes);
    if (!rc) {
	/* In the child */
        my_id = cr_init();
	(void)cr_register_callback(child_cb, &pipes, CR_SIGNAL_CONTEXT);
	crut_pipes_putchar(&pipes, MSG_CHILD_READY);
	crut_pipes_expect(&pipes, MSG_PARENT_DONE);
	crut_pipes_putchar(&pipes, MSG_CHILD_DONE);
	exit(0);
    } else if (rc < 0) {
	printf("XXX fork() failed\n");
	exit(-1);
    }

    (void)cr_register_callback(parent_cb, &pipes, CR_SIGNAL_CONTEXT);

    crut_pipes_expect(&pipes, MSG_CHILD_READY);
    printf("007 child started with pid %d\n", pipes.child);

    snprintf(cmd, sizeof(cmd), "%s --file %s --tree %d", chkpt_cmd, filename, my_pid);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc) || WEXITSTATUS(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    } else {
	printf("008 cr_checkpoint --tree OK\n");
    }

    snprintf(cmd, sizeof(cmd), "%s %s 2>/dev/null", rstrt_cmd, filename);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    }
    rc = WEXITSTATUS(rc);
    if (rc != EBUSY) {
	printf("XXX restart #4 returned unexpected status %d (%s)\n", rc, strerror(rc));
	exit(-1);
    } else {
	printf("009 restart tree failed in the expected way\n");
    }

    alarm(30);
    crut_pipes_putchar(&pipes, MSG_PARENT_DONE);
    crut_pipes_expect(&pipes, MSG_CHILD_DONE);
    crut_waitpid_expect(pipes.child, 0);
    alarm(0);

    /* Now restart again w/o the child running so we get only a PARTIAL pid conflict */
    snprintf(cmd, sizeof(cmd), "%s %s 2>/dev/null", rstrt_cmd, filename);
    rc = system(cmd);
    if ((rc < 0) || !WIFEXITED(rc)) {
	printf("XXX system(%s) failed (%d)\n", cmd, rc);
	exit(-1);
    }
    rc = WEXITSTATUS(rc);
    if (rc != EBUSY) {
	printf("XXX restart #5 returned unexpected status %d (%s)\n", rc, strerror(rc));
	exit(-1);
    } else {
	printf("010 restart tree/partial failed in the expected way\n");
    }

    (void)unlink(filename); // might fail silently
    printf("011 DONE\n");

    return 0;
}
