/*-------------------------------------------------------------------------
 *  vmadcheck.c:  Dump file sanity checker (for debugging only)
 *
 *  Copyright (C) 1999-2001 by Erik Hendriks <erik@hendriks.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: vmadcheck.c,v 1.6 2004/06/22 18:28:25 eroman Exp $
 *-----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

/* Including these files is fairly perilous. */
#include <asm/page.h>
#include <asm/ptrace.h>
struct task_struct;
#include <asm/types.h>
#include <asm/processor.h>

#include "vmadump.h"

/* This stuff is here because the kernel and glibc disagree about signal datatypes */
#define _NSIG           64
#define _NSIG_BPW       (sizeof(long)*8)
#define _NSIG_WORDS     (_NSIG / _NSIG_BPW)
typedef struct {
        unsigned long sig[_NSIG_WORDS];
} k_sigset_t;

struct k_sigaction {
    void *sa_handler;
    unsigned long sa_flags;
    void *sa_restorer;
    k_sigset_t sa_mask;
};



int read_check(int fd, void *p, size_t bytes) {
    int r;
    r = read(fd, p, bytes);
    if (r < 0) {
	fprintf(stderr, "Read error: %s\n", strerror(errno));
	exit(1);
    }
    if (r != bytes) {
	printf("Short file.\n");
	exit(1);
    }
    return r;
}

#define PTRWIDTH ((int)(sizeof(void *)*2))
char *arch_names[] = {"","i386","sparc", "alpha", "ppc"};
#define MAX_ARCH ((sizeof(arch_names)/sizeof(char *))-1)

static char vmad_magic[4] = VMAD_MAGIC;

void show_regs(int fd) {
    struct pt_regs regs;

    read_check(fd, &regs, sizeof(regs));
#if defined(__i386__)
    {
	struct thread_struct threadtmp;
	union i387_union fpu;
	long i;
	char flag;

	read_check(fd, &flag, sizeof(flag));
	if (flag) {
	    read_check(fd, &fpu, sizeof(fpu));
	}
	read_check(fd, &threadtmp, sizeof(threadtmp));

	printf("regs:   eax=%08lx  ebx=%08lx  ecx=%08lx  edx=%08lx\n"
	       "regs:   esi=%08lx  edi=%08lx  esp=%08lx  ebp=%08lx\n"
	       "regs:   eip=%08lx flgs=%08lx   cs=%04x      ss=%04x\n"
	       "regs:    ds=%04x       es=%04x       fs=%04lx      gs=%04lx\n",
	       regs.eax, regs.ebx, regs.ecx, regs.edx,
	       regs.esi, regs.edi, regs.esp, regs.ebp,
	       regs.eip, regs.eflags, regs.xcs, regs.xss,
	       regs.xds, regs.xes, threadtmp.fs, threadtmp.gs);
	if (flag)
	    printf("fpu:    FPU state present.\n");
	else
	    printf("fpu:    No FPU state.\n");
	printf("db:    ");
	for (i=0; i < 8; i++) printf(" %08lx", threadtmp.debugreg[i]);
	printf("\n");
    }
#elif defined(__sparc__)
    {
	int i;
	printf("regs:   pc=%08lx  npc=%08lx  y=%08lx\n",
	       regs.pc, regs.npc, regs.y);
	printf("regs: G");
	for (i=UREG_G0; i <=UREG_G7; i++) printf(" %08lx", regs.u_regs[i]);
	printf("\n");
	printf("regs: I");
	for (i=UREG_I0; i <=UREG_I7; i++) printf(" %08lx", regs.u_regs[i]);
	printf("\n");
    }
#elif defined(__alpha__)
    {
	struct switch_stack ss;
	unsigned long usp;
	read_check(fd, &ss,  sizeof(ss));
	read_check(fd, &usp, sizeof(usp));
	printf("regs: pc = %016lx  ra = %016lx  ps = %04lx\n"
	       "regs: r0 = %016lx  r1 = %016lx  r2 = %016lx\n"
	       "regs: r3 = %016lx  r4 = %016lx  r5 = %016lx\n"
	       "regs: r6 = %016lx  r7 = %016lx  r8 = %016lx\n"
	       "regs: r9 = %016lx  r10= %016lx  r11= %016lx\n"
	       "regs: r12= %016lx  r13= %016lx  r14= %016lx\n"
	       "regs: r15= %016lx\n"
	       "regs: r16= %016lx  r17= %016lx  r18= %016lx\n"
	       "regs: r19= %016lx  r20= %016lx  r21= %016lx\n"
	       "regs: r22= %016lx  r23= %016lx  r24= %016lx\n"
	       "regs: r25= %016lx  r27= %016lx  r28= %016lx\n"
	       "regs: gp = %016lx  sp = %016lx\n",
               regs.pc,  regs.r26, regs.ps,
               regs.r0,  regs.r1,  regs.r2,
               regs.r3,  regs.r4,  regs.r5,
               regs.r6,  regs.r7,  regs.r8,
               ss.r9,    ss.r10,   ss.r11,
               ss.r12,   ss.r13,   ss.r14,
               ss.r15,
               regs.r16, regs.r17, regs.r18,
	       regs.r19, regs.r20, regs.r21,
               regs.r22, regs.r23, regs.r24,
               regs.r25, regs.r27, regs.r28,
	       regs.gp, usp);
    }
#elif defined(powerpc)
    {
	double fpr[32];
	unsigned long fpscr;
	typedef struct {
	    __u32 u[4];
	} vector128;
	vector128 vr[32];
	vector128 vscr;
	read_check(fd, fpr, sizeof(fpr));
	read_check(fd, &fpscr, sizeof(fpscr));
	read_check(fd, vr, sizeof(vr));
	read_check(fd, &vscr, sizeof(vscr));

	for (i=0; i < 32; i++) {
	    if (i%4 == 0) printf("regs:");
	    printf(" r%02d = %08lx", i, regs.gpr[i]);
	    if (i%4 == 3) printf("\n");
	}

	for (i=0; i < 32; i++) {
	    if (i%4 == 0) printf("regs:");
	    printf(" fp%02d = %f", i, fpr[i]);
	    if (i%4 == 3) printf("\n");
	}
	printf("regs: fpscr = %08lx\n", fpscr);

	/* XXX We're presuming altivec is present here */
	for (i=0; i < 32; i++) {
	    if (i%2 == 0) printf("regs:");
	    printf(" av%02d = %08x%08x%08x%08x", i,
		   vr[i].u[0], vr[i].u[1], vr[i].u[2], vr[i].u[3]);
	    if (i%2 == 1) printf("\n");
	}
	printf("regs: vscr = %08x%08x%08x%08x\n",
	       vscr.u[0], vscr.u[1], vscr.u[2], vscr.u[3]);
    }
#else
    printf("regs: no register display for this arch\n");
#endif
}

void
show_blocked(int fd) {
    k_sigset_t blocked;
    int i;

    read_check(fd, &blocked, sizeof(blocked));
    printf("sigblk: ");
    for (i=0; i < _NSIG; i++) {
	if (blocked.sig[i / (sizeof(long)*8)] & (1 << (i % (sizeof(long)*8))))
	    printf(" %d", i+1);
    }
    printf("\n");
}

int main(int argc, char *argv[]) {
    int fd, i;
    int num_threads;
    char filename[PAGE_SIZE];
    struct vmadump_header      head;
    struct vmadump_vma_header  map;
    struct vmadump_page_header page;
    struct vmadump_mm_info     mm;
    struct k_sigaction sa[_NSIG];

    if (argc != 2) {
	fprintf(stderr, "Usage: %s dumpfile\n", argv[0]);
	exit(0);
    }
    
    fd = open(argv[1], O_RDONLY);
    if (fd == -1) {
	perror(argv[1]);
	exit(1);
    }

    read_check(fd, &head, sizeof(head));
    if (memcmp(head.magic, vmad_magic, sizeof(head.magic))) {
	printf("Invalid magic number.\n");
	exit(1);
    }
    printf("dump version: %d\n", (int) head.fmt_vers);
    if (head.fmt_vers != VMAD_FMT_VERS) {
	fprintf(stderr, "I only know about dump format version %d.  (Sorry)\n", VMAD_FMT_VERS);
	exit(1);
    }
    printf("kernel: %d.%d.%d\n",
	   (int) head.major, (int) head.minor, (int) head.patch);
    if (head.arch <= MAX_ARCH)
	printf("arch:   %s\n", arch_names[head.arch]);
    else
	printf("arch:   %d (unknown arch)\n", (int) head.arch);

    if (head.arch != VMAD_ARCH) {
	fprintf(stderr, "I can only inspect dumps from my own architecture.  (Sorry)\n");
	exit(1);
    }

    read_check(fd, filename, 16);
    printf("comm:   %s\n", filename);

    {
        int pid;
        read_check(fd, &pid, sizeof(int));
        printf("PID:   %d\n", pid);
    }

    /*--- register display --------------------------------------------*/
    show_regs(fd);

    read_check(fd, &mm, sizeof(mm));
    printf("code:   %0*lx-%0*lx\n", PTRWIDTH, mm.start_code, PTRWIDTH, mm.end_code);
    printf("data:   %0*lx-%0*lx\n", PTRWIDTH, mm.start_data, PTRWIDTH, mm.end_data);
    printf("brk:    %0*lx-%0*lx\n", PTRWIDTH, mm.start_brk,  PTRWIDTH, mm.brk);
    printf("stack:  %*s-%0*lx\n",   PTRWIDTH, "",            PTRWIDTH, mm.start_stack);
    printf("arg:    %0*lx-%0*lx\n", PTRWIDTH, mm.arg_start,  PTRWIDTH, mm.arg_end);
    printf("env:    %0*lx-%0*lx\n", PTRWIDTH, mm.env_start,  PTRWIDTH, mm.env_end);

    show_blocked(fd);

    read_check(fd, &sa,      sizeof(sa));
    for (i=0; i < _NSIG; i++) {
	if (sa[i].sa_handler == 0) continue;
	printf("sigact: %d", i+1);
	if (sa[i].sa_handler == (void *)1)
	    printf(" IGNORED");
	else {
	    printf(" handler=%p", sa[i].sa_handler);
#if 0
	if (blocked[i / (sizeof(long)*8)] & (1 << (i % (sizeof(long)*8))))
	    printf(" %d", i+1);
#endif
	}
	printf("\n");
    }

    /*--- Memory data ----------------------------------------------*/
    read_check(fd, &map, sizeof(map));
    while (map.start != ~0 || map.end != ~0) {
	if (map.namelen) {
	    if (map.namelen > PAGE_SIZE) {
		printf("Map has invalid name length: %d\n", (int) map.namelen);
		exit(1);
	    }
	    read_check(fd, filename, map.namelen);
	    filename[map.namelen]=0;
	    printf("map:    %0*lx-%0*lx %04x file %s:%0*lx\n",
		   PTRWIDTH, map.start, PTRWIDTH, map.end, (int)map.flags,
		   filename, PTRWIDTH, map.offset);
	} else {
	    printf("map:    %0*lx-%0*lx %04x (data provided)\n",
		   PTRWIDTH, map.start, PTRWIDTH, map.end, (int)map.flags);
	}
	read_check(fd, &page, sizeof(page));
	while (page.start != ~0) {
	    printf("page:   %0*lx (data provided)\n", PTRWIDTH, page.start);
	    lseek(fd, PAGE_SIZE, SEEK_CUR);
	    read_check(fd, &page, sizeof(page));
	}
	read_check(fd, &map, sizeof(map));
    }

#if defined(__i386__)
    {
#define LDT_ENTRY_SIZE 8

	long ldt_size;
        struct {
            unsigned int   lo32;
            unsigned int   hi32;
        } tmp_ldt;
        long i;

	read_check(fd, &ldt_size, sizeof(ldt_size));

	if (ldt_size) {
	    printf("ldt:    LDT data present.\n");

            for(i=0; i<ldt_size; ++i) {
                read_check(fd, &tmp_ldt, sizeof(tmp_ldt));
                if (tmp_ldt.hi32 != 0 || tmp_ldt.lo32 != 0) {
                    printf("LDT[%ld]: HI=%8x LO=%8x\n", i, tmp_ldt.hi32, tmp_ldt.lo32);
                }
            }
	} else
	    printf("ldt:    No LDT data.\n");
    }
#endif

    read_check(fd, &num_threads, sizeof(int));
    printf("num_threads:   %d\n", num_threads);

    /* dump out per-thread regs */
    for (i=1; i<num_threads; ++i) {
        int pid;
        read_check(fd, &pid, sizeof(int));
        printf("PID:   %d\n", pid);
        show_regs(fd);
	show_blocked(fd);
    }

    {
        int r = 0;
	int t = 0;
	char buf[128];
	r = read(fd, buf, sizeof(buf));
	t += r;
	while (r > 0) {
	    if (r < 0) {
	        perror("cruft:");
	    } else if (r > 0 && r < sizeof(buf)) {
                t += r;
	        break;
	    } else {
                t+= r;
	    }
	    r = read(fd, buf, sizeof(buf));
	}
	printf("junk: %d bytes at end\n", t);
    }


    exit(0);
}
  
/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
