/*-------------------------------------------------------------------------
 *  vmadump.c:  Virtual Memory Area dump/restore routines
 *
 *  Copyright (C) 1999-2001 by Erik Hendriks <erik@hendriks.cx>
 *      Multithreaded support written by Eric Roman <ERoman@lbl.gov>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Changelog:
 *  * November 14, 2003
 *     Extensive changes to this version (distributed with BLCR), by the
 *     BLCR development team checkpoint@lbl.gov
 *
 * $Id: vmadump.c,v 1.80 2007/08/24 22:23:51 phargrov Exp $
 *-----------------------------------------------------------------------*/
#ifndef EXPORT_SYMTAB
  #define EXPORT_SYMTAB
#endif

#include <linux/autoconf.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
  #define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
  #define MODVERSIONS
#endif
#if defined(MODVERSIONS)
  #include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/errno.h>
#include <linux/binfmts.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/personality.h>
#include <linux/highmem.h>
#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/vmalloc.h>

#ifdef __i386__
#include <asm/i387.h>
#include <asm/ldt.h>
#include <asm/desc.h>
#endif
#ifdef __sparc__
#include <asm/cprefix.h>
#endif

#include "vmadump.h"

/* Get results of autoconf probes */
#undef PACKAGE_VERSION
#include "blcr_config.h"

static char vmad_magic[3] = VMAD_MAGIC;
extern void * sys_call_table[];
static int syscall_num = 0;

MODULE_AUTHOR("Erik Hendriks <erik@hendriks.cx> and the BLCR Team http://ftg.lbl.gov/checkpoint");
MODULE_DESCRIPTION("VMADump - Virtual Memory Area Dumper (modified for " PACKAGE_STRING ")");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
MODULE_PARM(syscall_num, "i");
MODULE_PARM_DESC(syscall_num,
"Syscall number to allow calling VMADump directly.  The default (0) means "
"no VMADump syscall.  There is no bounds check on the syscall number so "
"be careful with this option.");

/*--------------------------------------------------------------------
 *  Some utliity stuff for reading and writing and misc kernel syscalls.
 *------------------------------------------------------------------*/
static
ssize_t default_read_file(struct vmadump_map_ctx *ctx, struct file *file,
			  void *buf, size_t count) {
    ssize_t ret = -EBADF;
    if (file->f_mode & FMODE_READ) {
#if HAVE_LOCKS_VERIFY_AREA
	ret = locks_verify_area(FLOCK_VERIFY_READ, file->f_dentry->d_inode,
				file, file->f_pos, count);
#elif HAVE_RW_VERIFY_AREA
	ret = rw_verify_area(READ, file, &file->f_pos, count);
#else
	#error "No verify_area"
#endif
	if (!ret) {
	    ssize_t (*read)(struct file *, char *, size_t, loff_t *);
	    ret = -EINVAL;
	    if (file->f_op && (read = file->f_op->read) != NULL)
		ret = read(file, buf, count, &file->f_pos);
	}
    }
    return ret;
}

static
ssize_t read_user(struct vmadump_map_ctx *ctx,
		  struct file *file, void *buf, size_t count) {
    ssize_t r, bytes = count;
    ssize_t (*rfunc)(struct vmadump_map_ctx *ctx, struct file *file,
		     void *buf, size_t count);
    rfunc = (ctx && ctx->read_file) ? ctx->read_file : default_read_file;
    while (bytes) {
#if SIZEOF_VOID_P == 4
	/* Avoid accidentally negative return values */
	r = rfunc(ctx, file, buf, ((bytes > 0x40000000) ? 0x40000000 : bytes));
#else
	r = rfunc(ctx, file, buf, bytes);
#endif
	if (r < 0)  return r;
	if (r == 0) return count - bytes;
	bytes -= r; buf += r;
    }
    return count;
}

static
ssize_t read_kern(struct vmadump_map_ctx *ctx, 
		  struct file *file, void *buf, size_t count) {
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = read_user(ctx, file, buf, count);
    set_fs(oldfs);
    return err;
}

static
ssize_t default_write_file(struct vmadump_map_ctx *ctx, struct file *file,
			   const void *buf, size_t count) {
    ssize_t ret = -EBADF;
    if (file->f_mode & FMODE_WRITE) {
#if HAVE_LOCKS_VERIFY_AREA
	struct inode *inode = file->f_dentry->d_inode;
	ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
				file->f_pos, count);
#elif HAVE_RW_VERIFY_AREA
	ret = rw_verify_area(WRITE, file, &file->f_pos, count);
#else
	#error "No verify_area"
#endif
	if (!ret) {
	    ssize_t (*write)(struct file *, const char *, size_t, loff_t *);
	    ret = -EINVAL;
	    if (file->f_op && (write = file->f_op->write) != NULL)
		ret = write(file, buf, count, &file->f_pos);
	}
    }
    return ret;
}

static
ssize_t write_user(struct vmadump_map_ctx *ctx, struct file *file,
		   const void *buf, size_t count) {
    ssize_t w, bytes = count;
    ssize_t (*wfunc)(struct vmadump_map_ctx *ctx, struct file *file,
		     const void *buf, size_t count);
    wfunc = (ctx && ctx->write_file) ? ctx->write_file : default_write_file;
    while (bytes) {
#if SIZEOF_VOID_P == 4
	/* Avoid accidentally negative return values */
	w = wfunc(ctx, file, buf, ((bytes > 0x40000000) ? 0x40000000 : bytes));
#else
	w = wfunc(ctx, file, buf, bytes);
#endif
	if (w < 0)  return w;
	if (w == 0) return count - bytes;
	bytes -= w; buf += w;
    }
    return count;
}

static
ssize_t write_kern(struct vmadump_map_ctx *ctx, struct file *file,
		   const void *buf, size_t count) {
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = write_user(ctx, file, buf, count);
    set_fs(oldfs);
    return err;
}

#if BPROC /* Not needed/maintained for BLCR */
static
long k_mprotect(unsigned long start, size_t len, unsigned long prot) {
    return ((long (*)(unsigned long,size_t,unsigned long))
	    sys_call_table[__NR_mprotect])(start,len,prot);
}

static
long k_exit(int code) {
    return ((long(*)(int))sys_call_table[__NR_exit])(code);
}

static
long k_access(const char *path, int mode) {
    return ((long(*)(const char *, int))
	    sys_call_table[__NR_access])(path,mode);
}
#endif


#if defined(__i386__)
/* IA32-specific stuff for the LDT */
  #if HAVE_MM_CONTEXT_SEGMENTS
    #define MM_LDT(mm)			((mm)->context.segments)
    #define MM_LDT_SIZE(mm)		LDT_ENTRIES
  #elif HAVE_MM_CONTEXT_LDT
    #define MM_LDT(mm)			((mm)->context.ldt)
    #define MM_LDT_SIZE(mm)		((mm)->context.size)
  #elif HAVE_MM_CONTEXT_LDT_PAGES
    #define MM_LDT(mm)			error
    #define MM_LDT_SIZE(mm)		((mm)->context.size)
  #else
    #error
  #endif

  #if HAVE_MM_CONTEXT_SEM
    #define MM_LDT_DOWN_READ(mm)	down(&(mm)->context.sem)
    #define MM_LDT_DOWN_WRITE(mm)	down(&(mm)->context.sem)
    #define MM_LDT_UP_READ(mm)		up(&(mm)->context.sem)
    #define MM_LDT_UP_WRITE(mm)		up(&(mm)->context.sem)
  #else
    #define MM_LDT_DOWN_READ		DOWN_READ_MMAP_SEM
    #define MM_LDT_DOWN_WRITE		DOWN_WRITE_MMAP_SEM
    #define MM_LDT_UP_READ		UP_READ_MMAP_SEM
    #define MM_LDT_UP_WRITE		UP_WRITE_MMAP_SEM
  #endif

/* Force creation of an LDT large enough to hold the given size
 * We create a not-present segment, but we'll overwrite it anyway */
static
int create_ldt(int size) {
    #if HAVE_STRUCT_MODIFY_LDT_LDT_S
        typedef struct modify_ldt_ldt_s	my_ldt_entry_t;
    #elif HAVE_STRUCT_USER_DESC
        typedef struct user_desc	my_ldt_entry_t;
    #else
        #error
    #endif

    int err;
    mm_segment_t oldfs;
    my_ldt_entry_t entry = { size-1, 0, 0, 1, 0, 0, 0, 1, 0 }; 
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = ((int(*)(int, my_ldt_entry_t *, size_t))
	    sys_call_table[__NR_modify_ldt])(1,&entry,sizeof(entry));
    set_fs(oldfs);
    return err;
}

/* Secure the segment selectors, zeroing invalid ones */
/* We require RPL = 3 (user) */
/* We only allow for fs and gs to select the LDT or the defined TLS segments */
unsigned short secure_selector(unsigned short s)
{
    int index = (int)(s >> 3);

    if (s == 0) {
	/* Common? case */
    } else if (((s&3) != 3)
	#if HAVE_TLS_IN_GDT
		|| (!(s&4) /* in GDT */ && (index < GDT_ENTRY_TLS_MIN || index > GDT_ENTRY_TLS_MAX))
	#endif
		|| ((s&4) /* in LDT */ && (index >= MM_LDT_SIZE(current->mm)))) {
	s = 0;
    }

    return s;
}
#endif

/* Work the same w/ varying page table walk code */
#if HAVE_PTE_OFFSET_MAP
  #define PTE_OFFSET_MAP(p,a)	pte_offset_map(p,a)
  #define PTE_UNMAP(p)		pte_unmap(p)
#elif HAVE_PTE_OFFSET
  #define PTE_OFFSET_MAP(p,a)	pte_offset(p,a)
  #define PTE_UNMAP(p)		do {} while (0)
#else
  #error
#endif

/* Work the same w/ varying signal code */
#if HAVE_0_ARG_RECALC_SIGPENDING
  #define RECALC_SIGPENDING_CURRENT()	recalc_sigpending()
#elif HAVE_1_ARG_RECALC_SIGPENDING
  #define RECALC_SIGPENDING_CURRENT()	recalc_sigpending(current)
#else
  #error
#endif
#if HAVE_2_4_SIGNAL_STRUCT
  #define LOCK_SIGMASK_CURRENT()	spin_lock_irq(&current->sigmask_lock)
  #define UNLOCK_SIGMASK_CURRENT()	spin_unlock_irq(&current->sigmask_lock)
  #define SIGACTION_CURRENT(i)		(current->sig->action[i])
#elif HAVE_2_6_SIGNAL_STRUCT
  #define LOCK_SIGMASK_CURRENT()	spin_lock_irq(&current->sighand->siglock);
  #define UNLOCK_SIGMASK_CURRENT()	spin_unlock_irq(&current->sighand->siglock);
  #define SIGACTION_CURRENT(i)		(current->sighand->action[i])
#else
  #error
#endif

#if BPROC /* Not needed/maintained for BLCR */
/*--------------------------------------------------------------------
 *  Library list stuff.
 *------------------------------------------------------------------*/
/* XXX This should be more configurable than this - this is idea of
 *     figuring out what's a library  */
struct liblist_entry {
    struct list_head list;
    char libname[0];
};

static LIST_HEAD(liblist);

static
void liblist_clear(void) {
    struct liblist_entry *entry;
    while (!list_empty(&liblist)) {
	entry = list_entry(liblist.next, struct liblist_entry, list);
	list_del(liblist.next);
	kfree(entry);
    }
}

static
struct liblist_entry *
liblist_find(const char *filename) {
    struct list_head *p;
    struct liblist_entry *entry;
    p = liblist.next;
    while (p != &liblist) {
	entry = list_entry(p, struct liblist_entry, list);
	if (strcmp(entry->libname, filename) == 0)
	    return entry;
	p = p->next;
    }
    return 0;
}

static
int liblist_add(char *filename) {
    struct liblist_entry *entry;

    /* Don't add things twice */
    if (liblist_find(filename)) return 0;

    entry = kmalloc(sizeof(*entry)+strlen(filename)+1, GFP_KERNEL);
    if (!entry) return -ENOMEM;
    strcpy(entry->libname, filename);
    list_add(&entry->list, &liblist);
    return 0;
}

static
int liblist_del(char *filename) {
    struct liblist_entry *entry;
    entry = liblist_find(filename);
    if (entry) {
	list_del(&entry->list);
	kfree(entry);
	return 0;
    }
    return -ENOENT;
}

/* Returns the size of our library list converted to text */
static
int liblist_size(void) {
    int len;
    struct list_head *p;
    struct liblist_entry *entry;

    len = 0;
    p = liblist.next;
    while (p != &liblist) {
	entry = list_entry(p, struct liblist_entry, list);
	len += strlen(entry->libname)+1;
	p = p->next;
    }
    len++;			/* for trailing null. */
    return len;
}

static
int do_lib_op(int request, char *buf, int size) {
    int err, len;
    char *filename;
    struct list_head *p;
    struct liblist_entry *entry;

    switch(request) {
    case VMAD_LIB_CLEAR:
	if (!suser()) return -EPERM;
	liblist_clear();
	return 0;

    case VMAD_LIB_ADD:
	if (!suser()) return -EPERM;
	filename = getname(buf);
	if (IS_ERR(filename)) return PTR_ERR(filename);
	err = liblist_add(filename);
	putname(filename);
	return err;

    case VMAD_LIB_DEL:
	if (!suser()) return -EPERM;
	filename = getname(buf);
	if (IS_ERR(filename)) return PTR_ERR(filename);
	err = liblist_del(filename);
	putname(filename);
	return err;

    case VMAD_LIB_SIZE:
	return liblist_size();

    case VMAD_LIB_LIST:
	len = liblist_size();
	if (len > size) return -ENOSPC;
	size = len;
	/* Copy all those strings out to user space. */
	p = liblist.next;
	while (p != &liblist) {
	    entry = list_entry(p, struct liblist_entry, list);
	    len = strlen(entry->libname);
	    if (copy_to_user(buf, entry->libname, len)) return EFAULT;
	    buf += len;
	    put_user('\0', buf++);
	    p = p->next;
	}
	put_user('\0', buf);
	return size;

    default:
	return -EINVAL;
    }
}
#endif /* Not needed/maintained for BLCR */

static
char *default_map_name(struct file *f, char *buffer, int size) {
    return d_path(f->f_dentry, f->f_vfsmnt, buffer, PAGE_SIZE);
}


/* this is gonna be handled with contexts too */
#if BPROC /* Not needed/maintained for BLCR */
static
int is_library(const char *filename) {
    return (liblist_find(filename) != 0);
}
#else
  #define is_library(_name) 0
#endif

static
struct file *default_map_open(const char *filename, int flags) {
    return filp_open(filename,flags,0);
}

/*--------------------------------------------------------------------
 *  Process "thawing" routines.
 *------------------------------------------------------------------*/
static
int mmap_file(struct vmadump_map_ctx *ctx,
	      const struct vmadump_vma_header *head, char *filename,
	      unsigned long flags) {
    struct file *file;
    long retval = -1;
    int open_flags;
    unsigned long prot;

    if (head->flags & VM_MAYSHARE) {
	if (head->flags & VM_MAYWRITE) {
	    open_flags = (head->flags & (VM_MAYREAD|VM_MAYEXEC)) ? O_RDWR : O_WRONLY;
	    prot = PROT_WRITE;
	} else {
	    open_flags = O_RDONLY;
	    prot = 0;
	}
	if (head->flags & VM_MAYREAD)  prot |= PROT_READ;
	if (head->flags & VM_MAYEXEC)  prot |= PROT_EXEC;
    } else {
	open_flags = O_RDONLY;
	prot = PROT_READ|PROT_WRITE|PROT_EXEC;
    }
  
    /* This is a lot like open w/o a file descriptor */
    if (ctx && ctx->map_open)
	file = ctx->map_open(ctx, filename, open_flags);
    else
	file = default_map_open(filename, open_flags);
    if (IS_ERR(file)) {
	printk("vmadump: open('%s', 0x%x) failed: %d\n", filename, open_flags, (int)PTR_ERR(file));
	return PTR_ERR(file);
    }

    DOWN_WRITE_MMAP_SEM(current->mm);
    retval = do_mmap(file, head->start, head->end - head->start,
		     prot, flags, head->offset);
    UP_WRITE_MMAP_SEM(current->mm);
    retval = (retval != head->start);
    if (retval) printk("do_mmap(<file>, %p, %p, ...) failed: %p\n",
		       (void *) head->start, (void *) head->end-head->start,
		       (void *) retval);
    fput(file);
    return retval;
}

int vmadump_load_page_list(struct vmadump_map_ctx *ctx,
			   struct file *file, int is_exec)
{
    struct vmadump_page_header page;
    long r;

    r = read_kern(ctx, file, &page, sizeof(page));
    while (r == sizeof(page) && page.start != ~0UL) {
	size_t len = page.num_pages << PAGE_SHIFT;
	r = read_user(ctx, file, (void *) page.start, len);
	if (r != len) goto err;
	if (is_exec)
	    flush_icache_range(page.start, page.start + len);
	r = read_kern(ctx, file, &page, sizeof(page));
    }
    if (r != sizeof(page)) goto err;
    return 0;

 err:
    if (r >= 0) r = -EIO;	/* map short reads to EIO */
    return r;
}
    

static
int load_map(struct vmadump_map_ctx *ctx,
	     struct file *file, struct vmadump_vma_header *head) {
    long r;
    unsigned long mmap_prot, mmap_flags, addr;

    mmap_prot  = 0;
    mmap_flags = MAP_FIXED | ((head->flags & VM_MAYSHARE) ? MAP_SHARED : MAP_PRIVATE);
    if (head->flags & VM_READ)  mmap_prot |= PROT_READ;
    if (head->flags & VM_WRITE) mmap_prot |= PROT_WRITE;
    if (head->flags & VM_EXEC)  mmap_prot |= PROT_EXEC;
    if (head->flags & VM_GROWSDOWN) mmap_flags |= MAP_GROWSDOWN;
    if (head->flags & VM_EXECUTABLE) mmap_flags |= MAP_EXECUTABLE;
    if (head->flags & VM_DENYWRITE) mmap_flags |= MAP_DENYWRITE;

    if (head->namelen) {
	char *filename;
	if (head->namelen > PAGE_SIZE) {
	    printk("vmadump: thaw: bogus namelen %d\n", (int) head->namelen);
	    return -EINVAL;
	}
	filename = kmalloc(head->namelen+1,GFP_KERNEL);
	r = read_kern(ctx, file, filename, head->namelen);
	if (r != head->namelen) goto err;
	filename[head->namelen] = 0;

	if (mmap_file(ctx, head, filename, mmap_flags)) {
	    kfree(filename);
	    printk("vmadump: mmap failed: %s\n", filename);
	    return -1;
	}
	kfree(filename);
    } else {
	/* Load the data from the dump file */
	DOWN_WRITE_MMAP_SEM(current->mm);
	addr = do_mmap(0, head->start, head->end - head->start,
		       mmap_prot|PROT_WRITE, mmap_flags, 0);
	UP_WRITE_MMAP_SEM(current->mm);
	if (addr != head->start) {
	    printk("do_mmap(0, %08lx, %08lx, ...) = 0x%08lx (failed)\n",
		   head->start, head->end - head->start, addr);
	    return -EINVAL;
	}
    }

    /* Read in patched pages */
    r = vmadump_load_page_list(ctx, file, (mmap_prot & PROT_EXEC));
    if (r) goto err;

    if (sys_mprotect(head->start,head->end - head->start, mmap_prot))
	printk("vmadump: thaw: mprotect failed. (ignoring)\n");
    return 0;

 err:
    if (r >= 0) r = -EIO;	/* map short reads to EIO */
    return r;
}

static
int vmadump_load_sigpending(struct vmadump_map_ctx *ctx, struct file *file,
                            struct sigpending *sigpending, int shared) {
    sigset_t pending_set;
    int flags;
    long r;

    r = read_kern(ctx, file, &flags, sizeof(flags));
    if (r != sizeof(flags)) goto bad_read;

    while (flags != -1) {
	struct siginfo info;

	r = read_kern(ctx, file, &info, sizeof(info));
	if (r != sizeof(info)) goto bad_read;

	read_lock_irq(&tasklist_lock);
#if HAVE_2_6_SIGNAL_STRUCT
	if (shared) {
	    r = group_send_sig_info(info.si_signo, &info, current);
	} else {
            LOCK_SIGMASK_CURRENT();
	    r = specific_send_sig_info(info.si_signo, &info, current);
            UNLOCK_SIGMASK_CURRENT();
	}
#else
	r = send_sig_info(info.si_signo, &info, current);
#endif
	read_unlock_irq(&tasklist_lock);

	r = read_kern(ctx, file, &flags, sizeof(flags));
	if (r != sizeof(flags)) goto bad_read;
    }

    /* Corresponding signal set comes last */
    r = read_kern(ctx, file, &pending_set, sizeof(pending_set));
    if (r != sizeof(pending_set)) goto bad_read;
    LOCK_SIGMASK_CURRENT();
    sigorsets(&sigpending->signal, &sigpending->signal, &pending_set);
    RECALC_SIGPENDING_CURRENT();
    UNLOCK_SIGMASK_CURRENT();

    return 0;

 bad_read:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    return r;
}

long vmadump_thaw_proc(struct vmadump_map_ctx *ctx,
		       struct file *file, struct pt_regs *regs,
		       int flags) {
    long r;
    int pid;
#if defined(__i386__)
    struct thread_struct threadtmp;	/* For the segment registers fs and gs  */
#endif

    struct vmadump_header header;

    /*--- First some sanity checking ---*/
    r = read_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) {
	printk("vmadump: failed to read header: %ld\n", r);
	return -EINVAL;
    }
    if (memcmp(header.magic, vmad_magic, sizeof(header.magic))) {
	printk("vmadump: invalid signature\n");
	return -EINVAL;
    }
    if (header.fmt_vers != VMAD_FMT_VERS) {
	printk(KERN_INFO "vmadump: dump version mistmatch. dump=%d; "
	       "kernel=%d\n", (int)header.fmt_vers, (int)VMAD_FMT_VERS);
	return -EINVAL;
    }
    if (header.arch != VMAD_ARCH) {
	printk("vmadump: architecture mismatch.\n");
	return -EINVAL;
    }

    if (header.major != ((LINUX_VERSION_CODE >> 16) & 0xFF) ||
	header.minor != ((LINUX_VERSION_CODE >> 8) & 0xFF)
#if STRICT_VERSION_CHECK
	||header.patch != (LINUX_VERSION_CODE & 0xFF)
#endif
	) {
	printk("vmadump: kernel version mismatch.\n");
	return -EINVAL;
    }

    /* Ummm... Point of no-return is here.... maybe try to move this
     * down a bit? */

    if (!(flags & VMAD_DUMP_REGSONLY)) {
        /* Read our new comm */
        r = read_kern(ctx, file, current->comm, sizeof(current->comm));
        if (r != sizeof(current->comm)) goto bad_read;
    } 

    /* 
     * PID 
     *
     * NOTE:  We don't actually restore the PID here, rather we just return
     * it to the caller.  This is intended to allow whoever is taking care
     * of restoring the thread library state to match up old PIDs with new 
     * ones later.
     */
    r = read_kern(ctx, file, &pid, sizeof(int));
    if (r != sizeof(int)) goto bad_read;

    /*
     * CPU-specific register restore stuff
     *
     * Note that we're not presuming that our current regs pointer
     * points to anything even vaguely reasonable.  This is done to
     * support bproc type kernel threads that have never been user
     * processes before.
     */
    {
    struct pt_regs regtmp;

    r = read_kern(ctx, file, &regtmp, sizeof(regtmp));
    if (r != sizeof(regtmp)) goto bad_read;
#if defined(__i386__)
    {
    union i387_union     i387tmp;
    char flag;

    /* Don't let the user pick bogo-segregs.  Restoring other values
     * will either lead us to fault while restoring or worse it might
     * allow users to do bad(tm) things in kernel space. */
    regtmp.xcs = __USER_CS;
    regtmp.xds = __USER_DS;
    regtmp.xes = __USER_DS;
    regtmp.xss = __USER_DS;
    /* fs and gs aren't represented in struct pt_regs... *hrm* */

    /* Only restore bottom 9 bits of eflags.  Restoring anything else
     * is bad bad mojo for security. (0x200 = interrupt enable) */
    regtmp.eflags = 0x200 | (regtmp.eflags & 0x000000FF);
    memcpy(regs, &regtmp, sizeof(regtmp));

    r = read_kern(ctx, file, &flag, sizeof(flag));
    if (r != sizeof(flag)) goto bad_read;

    if (flag) {
	/* Restore i386 FPU.
	 * - Clear TS so that we can do FPU ops.
	 * - Set PF_USEDFPU so the kernel knows that what's in the fpu right
	 *   now is of interest to us.
	 * - Read our dump data into the tss.
	 * - Manually restore from tss.
	 * - If the restore causes a GPF:
	 *     - Set TS so that FPU ops will trap.
	 *     - Clear used_math and PF_USEDFPU so the kernel will
	 *       give us a clean FPU when we trap.
	 */
	r = read_kern(ctx, file, &i387tmp, sizeof(i387tmp));
	if (r != sizeof(current->thread.i387)) goto bad_read;
	memcpy(&current->thread.i387, &i387tmp, sizeof(i387tmp));
	clts();
	current->used_math  = 1;
	current->flags     |= PF_USEDFPU;

	/* Invalid FPU states can blow us out of the water so we will do
	 * the restore here in such a way that we trap the fault if the
	 * restore fails.  This modeled after get_user and put_user. */
	if (cpu_has_fxsr) {
	    asm volatile
		("1: fxrstor %1               \n"
		 "2:                          \n"
		 ".section .fixup,\"ax\"      \n"
		 "3:  movl %2, %0             \n"
		 "    jmp 2b                  \n"
		 ".previous                   \n"
		 ".section __ex_table,\"a\"   \n"
		 "    .align 4                \n"
		 "    .long 1b, 3b            \n"
		 ".previous                   \n"
		 : "=r"(r)
		 : "m" (current->thread.i387.fxsave), "i"(-EFAULT));
	} else {
	    asm volatile
		("1: frstor %1                \n"
		 "2:                          \n"
		 ".section .fixup,\"ax\"      \n"
		 "3:  movl %2, %0             \n"
		 "    jmp 2b                  \n"
		 ".previous                   \n"
		 ".section __ex_table,\"a\"   \n"
		 "    .align 4                \n"
		 "    .long 1b, 3b            \n"
		 ".previous                   \n"
		 : "=r"(r)
		 : "m" (current->thread.i387.fsave), "i"(-EFAULT));
	}

	if (r == -EFAULT) {
	    /* Check for restore failure and bitch about it */
	    printk("vmadump: %d: FPU restore failure.\n", current->pid);
	    current->used_math  = 0;
	    current->flags     &= ~PF_USEDFPU;
	    stts();
	}
    } else {
	/* No FPU state for this process - setup for a clean FPU */
	current->used_math  = 0;
	current->flags     &= ~PF_USEDFPU;
	stts();
    }

    r = read_kern(ctx, file, &threadtmp, sizeof(threadtmp));
    if (r != sizeof(threadtmp)) goto bad_read;

    /* XXX FIX ME: RESTORE DEBUG INFORMATION ?? */
    }
#elif defined(__sparc__)
    regtmp.psr = regs->psr;	/* Don't restore PSR */
    memcpy(regs, &regtmp, sizeof(regtmp));
    /* no freely available docs = no fpu support */
#elif defined(__alpha__)
    {
	struct switch_stack *ss, sstmp;
	unsigned long usp;
	ss = ((struct switch_stack *)regs)-1;

	/* Read extra reg info for alpha */
	r = read_kern(ctx, file, &sstmp, sizeof(sstmp));
	if (r != sizeof(sstmp)) goto bad_read;
	r = read_kern(ctx, file, &usp, sizeof(usp));
	if (r != sizeof(usp)) goto bad_read;

	/* These are the tid-bits we don't want to restore */
	regtmp.ps  = 8;		/* Processor status 8 = user space */
	regtmp.hae = regs->hae;	/* XXX What is this reg? */
	sstmp.r26  = ss->r26;	/* Not actually the user r26(ra). That
				 * one is in struct pt_regs. */
	/* Ok, do restore */
	memcpy(regs, &regtmp, sizeof(regtmp));
	memcpy(ss, &sstmp, sizeof(sstmp)); /* Restores FP regs + saved regs */
	wrusp(usp);
    }
#elif defined(powerpc)
    {
 	/* Don't restore machine state register but set the PR bit
	 * since this is unpriviledged user space stuff we're
	 * restoring. */
	regtmp.msr = regs->msr | MSR_PR;
	memcpy(regs, &regtmp, sizeof(regtmp));

	r = read_kern(ctx, file, &current->thread.fpr,
		      sizeof(current->thread.fpr));
	if (r != sizeof(current->thread.fpr)) goto bad_read;

	r = read_kern(ctx, file, &current->thread.fpscr,
		      sizeof(current->thread.fpscr));
	if (r != sizeof(current->thread.fpscr)) goto bad_read;

#ifdef CONFIG_ALTIVEC
	/* Restore Altivec */
	if (regs->msr & MSR_VEC)
	    giveup_altivec(current);
	r = read_kern(ctx, file, &current->thread.vr,
		      sizeof(current->thread.vr));
	if (r != sizeof(current->thread.vr)) goto bad_read;

	r = read_kern(ctx, file, &current->thread.vscr,
		      sizeof(current->thread.vscr));
	if (r != sizeof(current->thread.vscr)) goto bad_read;
#endif
    }
#else
#error No processor specific state restore defined.
#endif
    }

    /*--- Signal information ---------------------------------------*/
    {
        int i;
        sigset_t             sig_blocked;
        struct k_sigaction   sig_action;
	    
        /* Install sets of blocked and pending signals */
        r = read_kern(ctx, file, &sig_blocked, sizeof(sig_blocked));
        if (r != sizeof(sig_blocked)) goto bad_read;

        sigdelsetmask(&sig_blocked, sigmask(SIGKILL)|sigmask(SIGSTOP));
        LOCK_SIGMASK_CURRENT();
        flush_signals(current);	/* Necessary? */
        memcpy(&current->blocked, &sig_blocked, sizeof(sig_blocked));
        RECALC_SIGPENDING_CURRENT();
        UNLOCK_SIGMASK_CURRENT();

        r = vmadump_load_sigpending(ctx, file, &current->pending, 0);
        if (r < 0) goto bad_read;
#if HAVE_2_6_SIGNAL_STRUCT
        if ((atomic_read(&current->signal->count) == 1) || !(flags & VMAD_DUMP_REGSONLY)) {
	    /* Restore shared queue if not actually shared, or if we are the "leader" */
	    r = vmadump_load_sigpending(ctx, file, &current->signal->shared_pending, 1);
	    if (r < 0) goto bad_read;
        }
#endif

        if (!(flags & VMAD_DUMP_REGSONLY)) {
            for (i=0; i < _NSIG; i++) {
	        r = read_kern(ctx, file, &sig_action, sizeof(sig_action));
	        if (r != sizeof(sig_action)) goto bad_read;
	        /* Security: Make sure the user doesn't block or handle the
	         * unblockable signals. */
	        if (i == SIGSTOP-1 || i == SIGKILL-1) {
	            sig_action.sa.sa_handler = SIG_DFL;
	            sig_action.sa.sa_flags   = 0;
	            sigemptyset(&sig_action.sa.sa_mask);
	        }
	        sigdelsetmask(&sig_action.sa.sa_mask,
		              sigmask(SIGKILL)|sigmask(SIGSTOP));

	        /* Install one sigaction */
	        LOCK_SIGMASK_CURRENT();
	        memcpy(&SIGACTION_CURRENT(i), &sig_action, sizeof(sig_action));
        	RECALC_SIGPENDING_CURRENT();
	        UNLOCK_SIGMASK_CURRENT();
	    }
        } 
	/* Nothing to do otherwise, since reference counts have been
	 * taken care of already */
    }

    /*--- Misc -------------------------------------*/
    #if HAVE_2_6_TASK_LINKAGE	/* RedHat 9 back ported this */
    r = read_kern(ctx, file, &current->clear_child_tid, sizeof(current->clear_child_tid));
    if (r != sizeof(current->clear_child_tid)) goto bad_read;
    #endif

    /*--- Memory map meta data -------------------------------------*/
    if (!(flags & VMAD_DUMP_REGSONLY)) {
        struct mm_struct *mm;
        struct vm_area_struct *map;
        struct vmadump_mm_info mm_info;
        struct vmadump_vma_header mapheader;

#if defined(__i386__)
  long ldt_size;
  #if HAVE_MM_CONTEXT_LDT_PAGES
	int page_nr;
  #endif
#endif

        mm = current->mm;
    
        r = read_kern(ctx, file, &mm_info, sizeof(mm_info));
        if (r != sizeof(mm_info)) { goto bad_read; }

        /* Purge current maps - I'm sure there's a way to keep theses around
         * incase creation of the new ones fails in some unfortunate way... */
        while(mm->mmap) {
	    map = mm->mmap;
#ifdef VM_ACCOUNT
	    r = do_munmap(mm, map->vm_start, map->vm_end - map->vm_start, 0);
#else
	    r = do_munmap(mm, map->vm_start, map->vm_end - map->vm_start);
#endif
	    if (r) {
	        printk("do_munmap(%lu, %lu) = %d\n",
		       map->vm_start, map->vm_end-map->vm_start, (int)r);
	    }
        }

        /* Load new map data */
        r = read_kern(ctx, file, &mapheader, sizeof(mapheader));
        while (r == sizeof(mapheader) &&
	       (mapheader.start != ~0 || mapheader.end != ~0)) {
	    if ((r = load_map(ctx, file, &mapheader))) goto bad_read;
	    r = read_kern(ctx, file, &mapheader, sizeof(mapheader));
        }
        if (r != sizeof(mapheader)) goto bad_read;    


        DOWN_WRITE_MMAP_SEM(current->mm);
        mm->start_code = mm_info.start_code;
        mm->end_code   = mm_info.end_code;
        mm->start_data = mm_info.start_data;
        mm->end_data   = mm_info.end_data;
        mm->start_brk  = mm_info.start_brk;
        mm->brk        = mm_info.brk;
        mm->start_stack= mm_info.start_stack;
        mm->arg_start  = mm_info.arg_start;
        mm->arg_end    = mm_info.arg_end;
        mm->env_start  = mm_info.env_start;
        mm->env_end    = mm_info.env_end;
        UP_WRITE_MMAP_SEM(current->mm);

#if defined(__i386__)
	/* Now try to restore the LDT
         * Note that 'ldt_size' is the number of entries, not size in bytes */
	r = read_kern(ctx, file, &ldt_size, sizeof(long));
	if (r != sizeof(long)) goto bad_read;

	if (ldt_size < MM_LDT_SIZE(mm)) {
	    /* Zero the existing LDT, if any.  It's too hard to free it. */
	    MM_LDT_DOWN_WRITE(mm);
#if HAVE_MM_CONTEXT_LDT_PAGES
	    __flush_tlb_global();
	    for (page_nr = 0; page_nr < MAX_LDT_PAGES; ++page_nr) {
		struct page *page = mm->context.ldt_pages[page_nr];
		if (page) {
		    memset(kmap(page), 0, PAGE_SIZE);
		    kunmap(page);
		}
	    }
#else
	    if (MM_LDT(mm) != NULL) {
		memset(MM_LDT(mm), 0, MM_LDT_SIZE(mm)*LDT_ENTRY_SIZE);
	    }
#endif
	    MM_LDT_UP_WRITE(mm);
	}
	if (ldt_size == 0) {
	    /* Nothing to do */
	} else if (ldt_size > LDT_ENTRIES) {
	    printk("vmadump:  Bad LDT size read\n");
	    r = -EINVAL;
	    goto bad_read;
	} else {
	    struct {
		__u32	lo32;
		__u32	hi32;
	    } * tmp_ldt;
	    size_t alloc_size;
	    int i;

	    alloc_size = ldt_size * LDT_ENTRY_SIZE;
	    tmp_ldt = vmalloc(alloc_size);

	    r = -ENOMEM;
	    if (tmp_ldt == NULL) {
		goto bad_read;
	    }

	    /* read the new LDT */
	    r = read_kern(ctx, file, tmp_ldt, alloc_size);
	    if (r != alloc_size) {
		printk("FAILED TO READ THE LDT\n");
		vfree(tmp_ldt);
	        goto bad_read;
            }

	    /* Secure the LDT by zeroing invalid/insecure entries */
	    for (i = 0; i < ldt_size; ++i) {
		if ((tmp_ldt[i].hi32 & 0x00207000) != 0x00007000) {
		    tmp_ldt[i].lo32 = tmp_ldt[i].hi32 = 0;
		}
	    }
	    
	    /* Force creation of an LDT */
	    if (create_ldt(ldt_size) != 0) {
		printk("FAILED TO CREATE AN LDT\n");
	        vfree(tmp_ldt);
		r = -ENOMEM;
	        goto bad_read;
	    }

	    /* Now install the new LDT */
	    MM_LDT_DOWN_WRITE(mm);
#if HAVE_MM_CONTEXT_LDT_PAGES	    
	    for (page_nr = 0; page_nr < MAX_LDT_PAGES; ++page_nr) {
		struct page *page = mm->context.ldt_pages[page_nr];
		if (page) {
		    int first = page_nr * PAGE_SIZE / LDT_ENTRY_SIZE;
		    int count = ldt_size - first;

		    if (count <= 0) {
		    	/* done */
		        break;
		    } else if (count > PAGE_SIZE / LDT_ENTRY_SIZE) {
			count = PAGE_SIZE / LDT_ENTRY_SIZE;
		    }
		    memcpy(kmap(page), &(tmp_ldt[first]), count);
		    kunmap(page);
		} else {
		    MM_LDT_UP_WRITE(mm);
		    vfree(tmp_ldt);
		    printk("LDT missing/truncated?\n");
		    r = -ENOMEM;
		    goto bad_read;
		}
	    }
#else
            if (MM_LDT(mm)) {
                memcpy(MM_LDT(mm), tmp_ldt, alloc_size);
	    } else {
		MM_LDT_UP_WRITE(mm);
		vfree(tmp_ldt);
		printk("LDT missing?\n");
		r = -ENOMEM;
		goto bad_read;
	    }
#endif
	    MM_LDT_UP_WRITE(mm);
	    vfree(tmp_ldt);
	}

#endif /* __i386__ */
    }

#if defined(__i386__)
    #if HAVE_TLS_IN_GDT
    {
	int cpu = smp_processor_id();
	struct desc_struct *local_gdt_table = cpu_gdt_table[cpu];
	int i;

	for (i = 0; i < GDT_ENTRY_TLS_ENTRIES; i++) {
	    struct desc_struct *entry = threadtmp.tls_array + i;

	    if (!desc_empty(entry) && ((entry->b & 0x00207000) != 0x00007000)) {
		entry->a = entry->b = 0;
	    }

	    local_gdt_table[GDT_ENTRY_TLS_MIN + i] = 
				current->thread.tls_array[i] = *entry;
	}
    }
    #endif /* HAVE_TLS_IN_GDT */

    /* Restore fs & gs segment registers (after LDT & GDT are restored). */
    threadtmp.fs = secure_selector(threadtmp.fs);
    threadtmp.gs = secure_selector(threadtmp.gs);
    current->thread.fs = threadtmp.fs;
    current->thread.gs = threadtmp.gs;
    loadsegment(fs, threadtmp.fs);
    loadsegment(gs, threadtmp.gs);
#endif

    return pid;

 bad_read:
    if (r >= 0) r = -EIO;
    return r;
}

/*--------------------------------------------------------------------
 *  Process "freezing" routines
 *------------------------------------------------------------------*/
/* This routine checks if a page from a filemap has been copied via
 * copy on write.  Basically, this is just checking to see if the page
 * is still a member of the map or not.  Note this this should not end
 * up copying random things from VM_IO regions. */
static
int addr_copied(struct mm_struct *mm, unsigned long addr) {
    pgd_t *pgd;
    pmd_t *pmd;
    pte_t *pte;
    struct page *pg;
    int ret;

    spin_lock(&mm->page_table_lock);
    pgd = pgd_offset(mm, addr);
    if (pgd_none(*pgd)) { ret = 0; goto out; }
    pmd = pmd_offset(pgd, addr);
    if (pmd_none(*pmd)) { ret = 0; goto out; }
    pte = PTE_OFFSET_MAP(pmd, addr);
    if (pte_present(*pte)) {
	pg = pte_page(*pte);
	ret = !pg->mapping;
	/* I think this was paranoia:
	   ret = !pg->mapping || pg->mapping !=
           map->vm_file->f_dentry->d_inode->i_mapping; */
    } else {
	/* pte_none is false for a swapped (written) page */
	ret = !pte_none(*pte);
    }
    PTE_UNMAP(pte);
 out:
    spin_unlock(&mm->page_table_lock);
    return ret;
}

/* This is the version for working on a region that is a file map.  In
 * this case we need to fault the page in to check for zero.  This
 * isn't a big deal since we'll be faulting in for sending anyway if
 * it's not.  */
static
int addr_nonzero_file(struct mm_struct *mm, unsigned long addr) {
    int i;
    unsigned long val;

    /* Simple zero check */
    for (i=0; i < (PAGE_SIZE/sizeof(long)); i++) {
	get_user(val, (((long*)addr)+i));
	if (val) return 1;
    }
    return 0;
}

/* This version is for use on regions which are *NOT* file maps.  Here
 * we look at the page tables to see if a page is zero.  If it's never
 * been faulted in, we know it's zero - and we don't fault it in while
 * checking for this. */
static
int addr_nonzero(struct mm_struct *mm, unsigned long addr) {
    int i, none;
    pgd_t *pgd;
    pmd_t *pmd;
    pte_t *pte;
    unsigned long val;

    spin_lock(&mm->page_table_lock);
    pgd = pgd_offset(mm, addr);
    if (pgd_none(*pgd)) goto out_zero;
    pmd = pmd_offset(pgd, addr);
    if (pmd_none(*pmd)) goto out_zero;
    pte = PTE_OFFSET_MAP(pmd, addr);
    none = pte_none(*pte);
    PTE_UNMAP(pte);
    if (none) goto out_zero;
    spin_unlock(&mm->page_table_lock);

    /* Ok, the page could be non-zero - check it... */
    for (i=0; i < (PAGE_SIZE/sizeof(long)); i++) {
	get_user(val, (((long*)addr)+i));
	if (val) return 1;
    }
    return 0;

 out_zero:
    spin_unlock(&mm->page_table_lock);
    return 0;
}

static
long store_page_list_eof(struct vmadump_map_ctx *ctx, struct file *file) {
    long r;
    struct vmadump_page_header header;
    header.start = ~0UL;
    r = write_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) r = -EIO;
    return r;
}

static /* Returns 0 on failure, so caller should also check num_contig_pages */
long store_page_range(struct vmadump_map_ctx *ctx, struct file *file,
		      unsigned long chunk_start, unsigned long num_contig_pages) {
    struct vmadump_page_header header;
    long r, bytes = 0;
    if (num_contig_pages) {
	header.start = chunk_start;
	header.num_pages = num_contig_pages;
	r = write_kern(ctx, file, &header, sizeof(header));
	if (r != sizeof(header)) return 0;
	bytes += r;
	r = write_user(ctx, file, (void *)chunk_start, num_contig_pages << PAGE_SHIFT);
	if (r != num_contig_pages << PAGE_SHIFT) return 0;
	bytes += r;
    }
    return bytes;
}

static
loff_t store_page_list(struct vmadump_map_ctx *ctx, struct file *file,
		       unsigned long start, unsigned long end,
		       int (*pfunc)(struct mm_struct *mm, unsigned long)) {
    long r;
    loff_t bytes = 0;
    unsigned long addr;
    unsigned long chunk_start, chunk_end, num_contig_pages;
    chunk_start = chunk_end = start;
    num_contig_pages = 0;
    for (addr = start; addr < end; addr += PAGE_SIZE) {
	if (pfunc(current->mm, addr)) {
	    if (chunk_end == addr) {
		num_contig_pages++;
	    } else {
		/* Write previous chunk, if any */
		r = store_page_range(ctx, file, chunk_start, num_contig_pages);
		if (num_contig_pages && !r) return -EIO;
		bytes += r;
		/* Start a new chunk */
		chunk_start = addr;
		num_contig_pages = 1;
	    }        
	    chunk_end = addr + PAGE_SIZE;
	}
    }
    /* Write final chunk, if any */
    r = store_page_range(ctx, file, chunk_start, num_contig_pages);
    if (num_contig_pages && !r) return -EIO;
    bytes += r;
    r = store_page_list_eof(ctx, file);
    if (r < 0) return r;
    bytes += r;
    return bytes;
}

loff_t vmadump_store_page_list(struct vmadump_map_ctx *ctx, struct file *file,
			       unsigned long start, unsigned long end) {
    return store_page_list(ctx, file, start, end, addr_nonzero_file);
}

static
loff_t store_map(struct vmadump_map_ctx *ctx, struct file *file,
	         struct vm_area_struct *map, int flags) {
    loff_t bytes;
    struct vmadump_vma_header head;
    char *filename=0;
    char *buffer = 0;
    loff_t r;
    unsigned long start, end;
    int isfilemap = 0;

    head.start   = map->vm_start;
    head.end     = map->vm_end;
    head.flags   = map->vm_flags;
    head.namelen = 0;
    head.offset  = map->vm_pgoff << PAGE_SHIFT; /* XXX LFS! */

    /* Decide Whether or not we're gonna store the map's contents or
     * a reference to the file they came from */
    if (map->vm_file) {
	struct dentry *dentry = map->vm_file->f_dentry;
	buffer = (char *) get_free_page(GFP_KERNEL);
	if (!buffer) { return -ENOMEM; }
	if (ctx && ctx->map_name)
	    filename = ctx->map_name(ctx, map->vm_file, buffer, PAGE_SIZE);
	else
	    filename = default_map_name(map->vm_file, buffer, PAGE_SIZE);
	head.namelen = strlen(filename);

	if (vmad_dentry_unlinked(dentry)) {
	    /* Region is an unlinked file or anonymous mapping */
            if ((flags & VMAD_DUMP_NOSHANON) && (map->vm_flags & VM_SHARED)) {
		free_page((long)buffer);
		return 0;
	    }
	    head.namelen = 0;
	} else if (map->vm_flags & VM_IO) {
	    /* Region is an IO map. */
	    
	    /* Never store the contents of a VM_IO region */
	} else if (map->vm_flags & VM_EXECUTABLE) {
	    /* Region is an executable */
	    if (flags & VMAD_DUMP_EXEC)
		head.namelen = 0;
	} else if (is_library(filename)) {
	    /* Region is a library */
	    if (flags & VMAD_DUMP_LIBS)
		head.namelen = 0;
	} else {
	    /* Region is something else */
	    if (flags & VMAD_DUMP_OTHER)
		head.namelen=0;
	}
	isfilemap = 1;
    }

    start     = map->vm_start;
    end       = map->vm_end;
    /* Release the mm_sem here to avoid deadlocks with page faults and
     * write locks that may happen during the writes.  (We can't use
     * the "map" pointer beyond this point. */
    UP_READ_MMAP_SEM(current->mm);

    /* Spit out the section header */
    r = write_kern(ctx, file, &head, sizeof(head));
    if (r != sizeof(head)) goto err;
    bytes = r;

    if (head.namelen) {
	/* Store the filename */
	r = write_kern(ctx, file, filename, head.namelen);
	if (r != head.namelen) goto err;
	bytes += r;
	r = store_page_list(ctx, file, start, end, addr_copied);
	if (r < 0) goto err;
	bytes += r;
    } else {
	/* Store the contents of the VMA as defined by start, end */
	r = store_page_list(ctx, file, start, end,
			    isfilemap ? addr_nonzero_file : addr_nonzero);
	if (r < 0) goto err;
	bytes += r;
    }
    if (buffer)   free_page((long)buffer);
    DOWN_READ_MMAP_SEM(current->mm);
    return bytes;

 err:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    if (buffer)   free_page((long)buffer);
    DOWN_READ_MMAP_SEM(current->mm);
    return r;
}

static
long vmadump_store_sigpending(struct vmadump_map_ctx *ctx, struct file *file,
                              struct sigpending *sigpending) {
    sigset_t pending_set;
    long bytes = 0, r;
    struct sigqueue *q;
    int flags = 0;

    /* We know the list can't shrink (as all threads are stopped),
     * So, we are playing a bit loose w/ the locking.
     */
    LOCK_SIGMASK_CURRENT();
    for (q = sigpending->head; q; q = q->next) {
	struct siginfo tmp_info, info;
	mm_segment_t oldfs;

	tmp_info = q->info;
	UNLOCK_SIGMASK_CURRENT();

	/* copy_siginfo_to_user() copies only non-padding fields */
	memset(&info, 0, sizeof(info));
	oldfs = get_fs(); set_fs(KERNEL_DS);
	r = copy_siginfo_to_user(&info, &tmp_info);
	set_fs(oldfs);

	if (r) goto err;

	r = write_kern(ctx, file, &flags, sizeof(flags));
	if (r != sizeof(flags)) goto err;
	bytes += r;
	r = write_kern(ctx, file, &info, sizeof(info));
	if (r != sizeof(info)) goto err;
	bytes += r;

	LOCK_SIGMASK_CURRENT();
    }
    memcpy(&pending_set, &sigpending->signal, sizeof(pending_set));
    UNLOCK_SIGMASK_CURRENT();

    /* Flags == -1 marks end of the list */
    flags = -1;
    r = write_kern(ctx, file, &flags, sizeof(flags));
    if (r != sizeof(flags)) goto err;
    bytes += r;

    /* Corresponding signal set comes last */
    r = write_kern(ctx, file, &pending_set, sizeof(pending_set));
    if (r != sizeof(pending_set)) goto err;
    bytes += r;

    return bytes;

 err:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    return r;
}

loff_t vmadump_freeze_proc(struct vmadump_map_ctx *ctx, struct file *file,
			   struct pt_regs *regs, int flags) {
    loff_t r, bytes=0;
    static struct vmadump_header header ={VMAD_MAGIC, VMAD_FMT_VERS, VMAD_ARCH,
					  (LINUX_VERSION_CODE >> 16) & 0xFF,
					  (LINUX_VERSION_CODE >> 8) & 0xFF,
					  LINUX_VERSION_CODE & 0xFF };

    /*--- Write out the file header ---*/
    r = write_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) goto err;
    bytes += r;

    if (!(flags & VMAD_DUMP_REGSONLY)) {
        r = write_kern(ctx, file, current->comm, sizeof(current->comm));
        if (r != sizeof(current->comm)) goto err;
        bytes += r;
    }

    /*--- PID ------------------------------------------------------*/
    r = write_kern(ctx, file, &current->pid, sizeof(int));
    if (r != sizeof(int)) goto err;
    bytes += r;

    /*--- CPU State Information ------------------------------------*/
    r = write_kern(ctx, file, regs, sizeof(*regs));
    if (r != sizeof(*regs)) goto err;
    bytes += r;

#if defined(__i386__)
    {
	struct user_i387_struct dummy;
	union i387_union i387tmp;
	char flag;

	/* The i386's FPU state isn't included in struct pt_regs, so
	 * save it seperately here. */

	/* Explanation of flags:
	 * used_math  =  1 iff process has used FPU
	 * PF_USEDFPU =  1 iff process has used math since last context switch
	 * TS         =  1 iff task switched, i.e. fpu ops will cause trap
	 */

	/* If we haven't touched the FPU, make like we have so that we
	 * have something to save.  First initialize it and then store
	 * it in our thread (unlazy does this.) */
	if (current->used_math) {
	    /* If the FPU has been used, try to save the state */
	    flag = 1;
	    r = write_kern(ctx, file, &flag, sizeof(flag));
	    if (r != sizeof(flag)) goto err;

	    if (current->flags & PF_USEDFPU) {
		/* hack here, unlazy in not available from to modules
		 * so call dump_fpu which has unlazy as a side effect.
		 * *sigh* */
		dump_fpu(0, &dummy);
	    }
	    memcpy(&i387tmp, &current->thread.i387, sizeof(i387tmp));
	    r = write_kern(ctx, file, &i387tmp, sizeof(i387tmp));
	    if (r != sizeof(i387tmp)) goto err;
	    bytes += r;
	} else {
	    /* No math used */
	    flag = 0;
	    r = write_kern(ctx, file, &flag, sizeof(flag));
	    if (r != sizeof(flag)) goto err;
	}

	/* Force lazyly saved segment registers to memory */
        asm volatile("movl %%fs,%0":"=m" (*(int *)&current->thread.fs));
        asm volatile("movl %%gs,%0":"=m" (*(int *)&current->thread.gs));

	/* Store debugging state */
	r = write_kern(ctx, file, &current->thread,
		      sizeof(current->thread));
	if (r != sizeof(current->thread)) goto err;
	bytes += r;
    }
#elif defined(__alpha__)
    /* Alpha has more registers than what's in struct ptregs to save
     * The 'switch_stack' structure contains the rest of the integer
     * registers and the floating point registers.  Then we still have
     * to go after the user's stack pointer via the PAL code. */
    {
	struct switch_stack *ss;
	unsigned long usp;
	/* Some pointer voodoo... the switch_stack structure got
	 * pushed on top of the struct pt_regs.. */
	ss = ((struct switch_stack *) regs)-1;
	usp = rdusp();

	r = write_kern(ctx, file, ss, sizeof(*ss));
	if (r != sizeof(*ss)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &usp, sizeof(usp));
	if (r != sizeof(usp)) goto err;
	bytes += r;
    }
#elif defined(powerpc)
    {
	/* Floating point regs */
	if (regs->msr & MSR_FP)
	    giveup_fpu(current);
	r = write_kern(ctx, file, &current->thread.fpr,
		       sizeof(current->thread.fpr));
	if (r != sizeof(current->thread.fpr)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &current->thread.fpscr,
		       sizeof(current->thread.fpscr));
	if (r != sizeof(current->thread.fpscr)) goto err;
	bytes += r;
#ifdef CONFIG_ALTIVEC
	/* XXX I really need to find out if this is right */
	if (regs->msr & MSR_VEC)
	    giveup_altivec(current);
	r = write_kern(ctx, file, &current->thread.vr,
		       sizeof(current->thread.vr));
	if (r != sizeof(current->thread.vr)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &current->thread.vscr,
		       sizeof(current->thread.vscr));
	if (r != sizeof(current->thread.vscr)) goto err;
	bytes += r;
#endif
    }
#else
#error No processor specific state save defined.
#endif

    {
        int i;
        sigset_t             sig_blocked;
        struct k_sigaction   sig_action;
    
        LOCK_SIGMASK_CURRENT();
        memcpy(&sig_blocked, &current->blocked, sizeof(sig_blocked));
        UNLOCK_SIGMASK_CURRENT();
    
        r = write_kern(ctx, file, &sig_blocked, sizeof(sig_blocked));
        if (r != sizeof(sig_blocked)) goto err;
        bytes += r;

        r = vmadump_store_sigpending(ctx, file, &current->pending);
        if (r < 0) goto err;
        bytes += r;
#if HAVE_2_6_SIGNAL_STRUCT
        if ((atomic_read(&current->signal->count) == 1) || !(flags & VMAD_DUMP_REGSONLY)) {
	    /* Dump shared queue if not actually shared, or if we are the "leader" */
	    r = vmadump_store_sigpending(ctx, file, &current->signal->shared_pending);
	    if (r < 0) goto err;
	    bytes += r;
        }
#endif

        if (!(flags & VMAD_DUMP_REGSONLY)) {
            for (i=0; i < _NSIG; i++) {
	        LOCK_SIGMASK_CURRENT();
	        memcpy(&sig_action, &SIGACTION_CURRENT(i), sizeof(sig_action));
	        UNLOCK_SIGMASK_CURRENT();
	    
	        r = write_kern(ctx, file, &sig_action, sizeof(sig_action));
	        if (r != sizeof(sig_action)) goto err;
	        bytes += r;
	    }
        }
    }
    
    /*--- Misc -------------------------------------*/
    #if HAVE_2_6_TASK_LINKAGE	/* RedHat 9 back ported this */
    r = write_kern(ctx, file, &current->clear_child_tid, sizeof(current->clear_child_tid));
    if (r != sizeof(current->clear_child_tid)) goto err;
    bytes += r;
    #endif

    /*--- Memory Information ---------------------------------------*/
    if (!(flags & VMAD_DUMP_REGSONLY)) {
        struct vm_area_struct     *map, *next_map;
        struct vmadump_mm_info     mm_info;
        struct mm_struct          *mm = current->mm;
        struct vmadump_vma_header  term;
        unsigned long              next_addr;

#if defined(__i386__)
	void *tmp_ldt = NULL;
	long ldt_size;
	size_t alloc_size;
  #if HAVE_MM_CONTEXT_LDT_PAGES	    
	int page_nr;
  #endif
#endif

        DOWN_READ_MMAP_SEM(mm);
        mm_info.start_code  = mm->start_code;
        mm_info.end_code    = mm->end_code;
        mm_info.start_data  = mm->start_data;
        mm_info.end_data    = mm->end_data;
        mm_info.start_brk   = mm->start_brk;
        mm_info.brk         = mm->brk;
        mm_info.start_stack = mm->start_stack;
        mm_info.arg_start   = mm->arg_start;
        mm_info.arg_end     = mm->arg_end;
        mm_info.env_start   = mm->env_start;
        mm_info.env_end     = mm->env_end;
        UP_READ_MMAP_SEM(mm);

        r = write_kern(ctx, file, &mm_info, sizeof(mm_info));
        if (r != sizeof(mm_info)) goto err;
        bytes += r;

        DOWN_READ_MMAP_SEM(mm);
        next_map = mm->mmap;
	next_addr = next_map ? next_map->vm_start : 0;
        while (next_map) {
	    /* Call find_vma() to find the map we're looking for.  We
	     * have to do it this way because store_map needs to release
	     * the mmap_sem. */
	    map = find_vma(mm, next_addr);
	    if (map != next_map) break;
	    next_map = map->vm_next;
	    next_addr = next_map ? next_map->vm_start : 0;
	    r = store_map(ctx, file, map, flags);
	    if (r < 0) {
	        UP_READ_MMAP_SEM(mm);
	        goto err;
	    }
	    bytes += r;
        }
        UP_READ_MMAP_SEM(mm);
    
        /* Terminate maps list */
        term.start = term.end = ~0L;
        r = write_kern(ctx, file, &term, sizeof(term));
        if (r != sizeof(term)) goto err;
        bytes += r;

#if defined(__i386__)
	/* Store the local descriptor table */
	/* allocate a temporary data structure to hold our LDT */
	MM_LDT_DOWN_READ(mm);
	ldt_size = MM_LDT_SIZE(mm);
	alloc_size = ldt_size*LDT_ENTRY_SIZE;
	if (ldt_size) {
#if HAVE_MM_CONTEXT_LDT_PAGES	    
	    tmp_ldt = vmalloc(alloc_size);
            if (tmp_ldt == NULL) {
	        MM_LDT_UP_READ(mm);
                r = -ENOMEM;
	        goto err;	    
	    }
	    __flush_tlb_global();
	    for (page_nr = 0; page_nr < MAX_LDT_PAGES; ++page_nr) {
		struct page *page = mm->context.ldt_pages[page_nr];
		if (page) {
		    int first = page_nr * PAGE_SIZE / LDT_ENTRY_SIZE;
		    int count = ldt_size - first;

		    if (count <= 0) {
		    	/* done */
		        break;
		    } else if (count > PAGE_SIZE / LDT_ENTRY_SIZE) {
			count = PAGE_SIZE / LDT_ENTRY_SIZE;
		    }
		    memcpy(((char *)tmp_ldt + count*LDT_ENTRY_SIZE), kmap(page), count);
		    kunmap(page);
		} else {
		    MM_LDT_UP_WRITE(mm);
		    vfree(tmp_ldt);
		    printk("LDT missing/truncated?\n");
		    r = -ENOMEM;
		    goto err;
		}
	    }
#else
	    if (MM_LDT(mm)) {
	        tmp_ldt = vmalloc(alloc_size);
                if (tmp_ldt == NULL) {
	            MM_LDT_UP_READ(mm);
                    r = -ENOMEM;
	            goto err;	    
	        }
                memcpy(tmp_ldt, MM_LDT(mm), alloc_size);
	    } else {
		/* OK to have MM_LDT==0 while MM_LDT_SIZE is non-zero */
		ldt_size = 0;
	    }
#endif
	}
	MM_LDT_UP_READ(mm);

	r = write_kern(ctx, file, &ldt_size, sizeof(long));
	if (r != sizeof(long)) goto err;
	bytes += r;

	if (ldt_size) {
	    r = write_kern(ctx, file, tmp_ldt, alloc_size);
	    if (r != alloc_size) goto err;
	    bytes += r;
	}

	vfree(tmp_ldt);
#endif /* __i386__ */
    }

    return bytes;

 err:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    return r;
}

/*--------------------------------------------------------------------
 * syscall interface
 *------------------------------------------------------------------*/
/* These tell us how to get syscall arguments out of a struct pt_regs */
#if defined(__i386__)
#define reg_arg0(r)      ((r)->ebx)
#define reg_arg1(r)      ((r)->ecx)
#define reg_arg2(r)      ((r)->edx)
#elif defined(__sparc__)
#define reg_arg0(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I0])
#define reg_arg1(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I1])
#define reg_arg2(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I2])
#elif defined(__alpha__)
#define reg_arg0(r)      ((r)->trap_a0)
#define reg_arg1(r)      ((r)->trap_a1)
#define reg_arg2(r)      ((r)->trap_a2)
#elif defined(powerpc)
#define reg_arg0(r)      ((r)->gpr[3])
#define reg_arg1(r)      ((r)->gpr[4])
#define reg_arg2(r)      ((r)->gpr[5])
#else
#error reg_arg[0-2] needs to be defined for this arch
#endif

#if BPROC /* Not needed/maintained for BLCR */
long do_vmadump(struct pt_regs *regs) {
    long retval;
    int op    = reg_arg0(regs);
    int fd    = reg_arg1(regs);
    int flags = reg_arg2(regs);
    struct file *file;

    MOD_INC_USE_COUNT;
    switch (op) {
    case VMAD_DO_DUMP:
	if ((file = fget(fd))) {
	    retval = vmadump_freeze_proc(0, file, regs, flags, 1);
	    fput(file);
	} else
	    retval = -EBADF;
	break;
    case VMAD_DO_UNDUMP:
	if ((file = fget(fd))) {
	    retval = vmadump_thaw_proc(0, file, regs, 0, NULL);
	    fput(file);
	} else
	    retval = -EBADF;

	/* Un-dump is a whole lot like exec() */
	if (retval == 0) {
	    if (current->euid == current->uid && current->egid == current->gid) {
#if HAVE_MM_DUMPABLE
		current->mm->dumpable = 1;
#endif
#if HAVE_TASK_DUMPABLE
		current->dumpable = 1;
#elif HAVE_TASK_TASK_DUMPABLE
		current->task_dumpable = 1;
#endif
	    }
	    current->did_exec = 1;
	    current->self_exec_id++;
	    if (current->ptrace & PT_PTRACED)
                send_sig(SIGTRAP, current, 0);
	}
	break;
    case VMAD_DO_EXECDUMP: {
	struct vmadump_execdump_args args;
	char * filename;
	retval = -EFAULT;
	if (copy_from_user(&args, (void *) reg_arg1(regs), sizeof(args)))
	    break;

	/* Check permissions on the binary image. */
	retval = k_access(args.arg0, 4);
	if (retval)
	    break;

	filename = getname(args.arg0);
	retval = PTR_ERR(filename);
	if (IS_ERR(filename)) break;

	file = fget(args.fd);
	if (!file) {
	    retval = -EBADF;
	    putname(filename);
	    break;
	}

	retval = do_execve(filename, (char **)args.argv,
			   (char **)args.envp, regs);
	putname(filename);
	if (retval) {
	    fput(file);
	    break;
	}

	retval = vmadump_freeze_proc(0, file, regs, args.flags, 1);
	fput(file);
	if (retval > 0) retval = 0;
	MOD_DEC_USE_COUNT;
	k_exit(retval);		/* Ok, we're done... */
	/* NOT REACHED */
    } break;

    case VMAD_DO_THREADDUMP: {
	struct vmadump_threaddump_args args;

        retval = -EFAULT;
        if (copy_from_user(&args, (void *) reg_arg1(regs), sizeof(args)))
            break;

        file = fget(args.fd);
        if (!file) {
            retval = -EBADF;
            break;
         }

	retval = vmadump_freeze_threads(file, regs, &args);
    } break;

    case VMAD_DO_THREADUNDUMP: {
	struct vmadump_threadundump_args args;

        retval = -EFAULT;
        if (copy_from_user(&args, (void *) reg_arg1(regs), sizeof(args)))
            break;

        file = fget(args.fd);
        if (!file) {
            retval = -EBADF;
            break;
         }

        retval = vmadump_thaw_threads(file, regs, &args);	
    } break;

    case VMAD_LIB_CLEAR:
    case VMAD_LIB_ADD:
    case VMAD_LIB_DEL:
    case VMAD_LIB_LIST:
    case VMAD_LIB_SIZE:
	lock_kernel();		/* very very lazy... */
	retval = do_lib_op(op, (char *) reg_arg1(regs), reg_arg2(regs));
	unlock_kernel();
	break;

    default:
	retval = -EINVAL;
    }
    MOD_DEC_USE_COUNT;
    return retval;
}

/*--------------------------------------------------------------------
 * Stubs for syscall entry on various architectures
 *------------------------------------------------------------------*/
#ifdef __i386__
static
int sys_vmadump(struct pt_regs regs) {
    return do_vmadump(&regs);
}
#endif

#ifdef __alpha__
#define SWITCH_STACK_SIZE "320"
int sys_vmadump(void);
/* do_switch_stack/undo_switch_stack stolen from arch/alpha/kernel/entry.S */
asm(
".align 3\n"
".ent	save_switch_stack\n"
"save_switch_stack:\n"
"	lda	$30,-"SWITCH_STACK_SIZE"($30)\n"
"	stq	$9,0($30)\n"
"	stq	$10,8($30)\n"
"	stq	$11,16($30)\n"
"	stq	$12,24($30)\n"
"	stq	$13,32($30)\n"
"	stq	$14,40($30)\n"
"	stq	$15,48($30)\n"
"	stq	$26,56($30)\n"
"	stt	$f0,64($30)\n"
"	stt	$f1,72($30)\n"
"	stt	$f2,80($30)\n"
"	stt	$f3,88($30)\n"
"	stt	$f4,96($30)\n"
"	stt	$f5,104($30)\n"
"	stt	$f6,112($30)\n"
"	stt	$f7,120($30)\n"
"	stt	$f8,128($30)\n"
"	stt	$f9,136($30)\n"
"	stt	$f10,144($30)\n"
"	stt	$f11,152($30)\n"
"	stt	$f12,160($30)\n"
"	stt	$f13,168($30)\n"
"	stt	$f14,176($30)\n"
"	stt	$f15,184($30)\n"
"	stt	$f16,192($30)\n"
"	stt	$f17,200($30)\n"
"	stt	$f18,208($30)\n"
"	stt	$f19,216($30)\n"
"	stt	$f20,224($30)\n"
"	stt	$f21,232($30)\n"
"	stt	$f22,240($30)\n"
"	stt	$f23,248($30)\n"
"	stt	$f24,256($30)\n"
"	stt	$f25,264($30)\n"
"	stt	$f26,272($30)\n"
"	stt	$f27,280($30)\n"
"	mf_fpcr	$f0\n"
"	stt	$f28,288($30)\n"
"	stt	$f29,296($30)\n"
"	stt	$f30,304($30)\n"
"	stt	$f0,312($30)\n"
"	ldt	$f0,64($30)\n"
"	ret	$31,($1),1\n"
".end save_switch_stack\n"
"\n"
".align 3\n"
".ent	restore_switch_stack\n"
"restore_switch_stack:\n"
"	ldq	$9,0($30)\n"
"	ldq	$10,8($30)\n"
"	ldq	$11,16($30)\n"
"	ldq	$12,24($30)\n"
"	ldq	$13,32($30)\n"
"	ldq	$14,40($30)\n"
"	ldq	$15,48($30)\n"
"	ldq	$26,56($30)\n"
"	ldt	$f30,312($30)	# get saved fpcr\n"
"	ldt	$f0,64($30)\n"
"	ldt	$f1,72($30)\n"
"	ldt	$f2,80($30)\n"
"	ldt	$f3,88($30)\n"
"	mt_fpcr	$f30		# install saved fpcr\n"
"	ldt	$f4,96($30)\n"
"	ldt	$f5,104($30)\n"
"	ldt	$f6,112($30)\n"
"	ldt	$f7,120($30)\n"
"	ldt	$f8,128($30)\n"
"	ldt	$f9,136($30)\n"
"	ldt	$f10,144($30)\n"
"	ldt	$f11,152($30)\n"
"	ldt	$f12,160($30)\n"
"	ldt	$f13,168($30)\n"
"	ldt	$f14,176($30)\n"
"	ldt	$f15,184($30)\n"
"	ldt	$f16,192($30)\n"
"	ldt	$f17,200($30)\n"
"	ldt	$f18,208($30)\n"
"	ldt	$f19,216($30)\n"
"	ldt	$f20,224($30)\n"
"	ldt	$f21,232($30)\n"
"	ldt	$f22,240($30)\n"
"	ldt	$f23,248($30)\n"
"	ldt	$f24,256($30)\n"
"	ldt	$f25,264($30)\n"
"	ldt	$f26,272($30)\n"
"	ldt	$f27,280($30)\n"
"	ldt	$f28,288($30)\n"
"	ldt	$f29,296($30)\n"
"	ldt	$f30,304($30)\n"
"	lda	$30,"SWITCH_STACK_SIZE"($30)\n"
"	ret	$31,($1),1\n"
".end restore_switch_stack\n"
"\n"
".align 3\n"
".globl  sys_vmadump\n"
".ent    sys_vmadump\n"
"sys_vmadump:\n"
"        ldgp    $29,0($27)\n"
"        bsr     $1,save_switch_stack\n"
"        lda     $16,"SWITCH_STACK_SIZE"($30)\n"
"        jsr     $26,do_vmadump\n"
"        bsr     $1,restore_switch_stack\n"
"        ret     $31,($26),1\n"
".end    sys_vmadump\n");
#endif

#ifdef __sparc__
int sys_vmadump(void);
#define STRING_1(x) #x
#define STRING(x) STRING_1(x)
#define LABEL(x) STRING(C_LABEL(x))
asm(
"	.align	4\n"
"	.globl	"LABEL(sys_vmadump)"\n"
LABEL(sys_vmadump)":\n"
"	mov	%o7, %l5\n"
"	add	%sp, 0x40, %o0\n"
"	call	"LABEL(do_vmadump)"\n"
"	 mov	%l5, %o7");
#endif

#ifdef powerpc
int sys_vmadump(int p1, int p2, int p3, int p4, int p5, int p6,
		struct pt_regs *regs) {
    return do_vmadump(regs);
}
#endif

#endif /* Not needed/maintained for BLCR */

/*--------------------------------------------------------------------
 *  New binary format code
 *------------------------------------------------------------------*/
#ifdef HAVE_BINFMT_VMADUMP

static
int load_vmadump(struct linux_binprm *bprm, struct pt_regs *regs) {
    int retval;
    struct vmadump_header *header;

    header = (struct vmadump_header *) bprm->buf;
    if (memcmp(header->magic, vmad_magic, sizeof(header->magic)) != 0  || 
	header->fmt_vers != VMAD_FMT_VERS ||
	header->major != ((LINUX_VERSION_CODE >> 16) & 0xFF) ||
	header->minor != ((LINUX_VERSION_CODE >> 8) & 0xFF)
#if STRICT_VERSION_CHECK
	|| header->patch != (LINUX_VERSION_CODE & 0xFF)
#endif
	)
	return -ENOEXEC;

    if (!bprm->file->f_op || !bprm->file->f_op->mmap)
	return -ENOEXEC;

    retval = vmadump_thaw_proc(0, bprm->file, regs, 0, NULL);
    if (retval == 0) {
	if (current->euid == current->uid && current->egid == current->gid) {
#if HAVE_MM_DUMPABLE
	    current->mm->dumpable = 1;
#endif
#if HAVE_TASK_DUMPABLE
	    current->dumpable = 1;
#elif HAVE_TASK_TASK_DUMPABLE
	    current->task_dumpable = 1;
#endif
	}
	current->did_exec = 1;
	current->self_exec_id++;
        if (current->exec_domain && current->exec_domain->module)
	    __MOD_DEC_USE_COUNT(current->exec_domain->module);
        if (current->binfmt && current->binfmt->module)
	    __MOD_DEC_USE_COUNT(current->binfmt->module);
	current->exec_domain = 0;
	current->binfmt = 0;
    }
    return retval;
}

struct linux_binfmt vmadump_fmt = {
    0,THIS_MODULE,load_vmadump,0,0,0
};
#endif

#if BPROC
static
char *get_rev(char *revstr) {
    char *ch; if ((ch=strchr(revstr+11,' ')))*ch = 0;
    return revstr+11;
}
#endif

#if BPROC /* Not needed/maintained for BLCR */
static void * old_sys_call;
#endif
int init_module(void) {
#if BPROC
    printk(KERN_INFO "vmadump: %s Erik Hendriks "
	   "<erik@hendriks.cx>\n", get_rev("$Revision: 1.80 $"));
#else
    printk(KERN_INFO "vmadump: from bproc-%s Erik Hendriks "
	   "<erik@hendriks.cx>\n", BPROC_VERSION);
    printk(KERN_INFO "vmadump: Modified for %s <%s>\n",
	   PACKAGE_STRING, PACKAGE_BUGREPORT);
#endif
    /* We can't get an upper bound on the number of syscalls so be
     * careful here! */
    if (syscall_num < 0) return -EINVAL;

    /* Check current kernel against System.map used af configure time */
    if ((CR_EXPORTED_KCODE_register_chrdev != (unsigned long)&register_chrdev) ||
	(CR_EXPORTED_KCODE_register_blkdev != (unsigned long)&register_blkdev)) {
	    printk(KERN_ERR "Running kernel does not match the System.map used to build vmadump.o\n");
	    return -EINVAL;
    }

#ifdef HAVE_BINFMT_VMADUMP
    register_binfmt(&vmadump_fmt);
#endif

#if BPROC /* Not needed/maintained for BLCR */
    if (syscall_num) {
	printk(KERN_INFO "vmadump: syscall number %d\n", syscall_num);
	old_sys_call = sys_call_table[syscall_num];
	sys_call_table[syscall_num] = sys_vmadump;
    }
#endif
    return 0;
}

void cleanup_module(void) {
#if BPROC /* Not needed/maintained for BLCR */
    liblist_clear();
    if (syscall_num) {
	sys_call_table[syscall_num] = old_sys_call;
    }
#endif
#ifdef HAVE_BINFMT_VMADUMP
    unregister_binfmt(&vmadump_fmt);
#endif
}

EXPORT_SYMBOL_GPL(vmadump_load_page_list);
EXPORT_SYMBOL_GPL(vmadump_store_page_list);
EXPORT_SYMBOL(vmadump_freeze_proc);
EXPORT_SYMBOL(vmadump_thaw_proc);
#if BPROC /* Not needed/maintained for BLCR */
EXPORT_SYMBOL(do_vmadump);
#endif

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
